subroutine SUB_NAME (ifile,name,arr,gzip)
  !
  ! Write an n-D array dataset to an open file. This assumes some
  ! preprocessor variables have been set - see hdf_wrapper.F90.
  !
  ! Compress data if 'gzip' parameter is set - value is compression level, 0-9,
  ! as used by the unix gzip command.
  !
  implicit none
  integer, parameter              :: LEN_STR = 256
  integer, parameter              :: MAX_NEST = 10
  character(len=*), intent(in)    :: name 
  integer,intent(in)              :: ifile
  integer,intent(in),optional     :: gzip
#ifndef SCALAR
  ARR_TYPE , dimension ARRAY_DIM  :: arr
#else
  ARR_TYPE                        :: arr
#endif
  integer(hid_t)                  :: dset_id,dtype_id,dspace_id,loc_id
  integer(hid_t)                  :: group_id
  integer(hid_t)                  :: id_chain(MAX_NEST)
  integer                         :: n_nest
  integer                         :: hdf_err
  integer(hsize_t), dimension(10) :: dims,maxdims
  integer(size_t)                 :: size, nativesize
#ifndef SCALAR
  integer(hsize_t)                :: dimensions(NDIMS)
#endif
  integer                         :: class, rank
  integer                         :: i, starti, endi
  character, parameter            :: slash='/'
  character(len=LEN_STR)          :: thestring,currentgroup
  character(len=LEN_STR)          :: temp,currentdataset
  integer(hid_t)                  :: prp_id

  if(ifile.lt.1.or.ifile.gt.nfilemax)call hdf5_abort( &
       'Invalid file handle in write_data',name=name)

  if(file_id(ifile).lt.0) call hdf5_abort( &
       'File is not open in write_data()!', &
       name=name)

  if(read_only(ifile))call hdf5_abort( &
       "Attempt to write dataset to read only file!", &
       name=name,fname=fname_table(ifile))

  ! The location to open the dataset:
  loc_id = file_id(ifile)

  ! Generate array of dimensions:
#ifndef SCALAR
  do i=1,NDIMS
      dimensions(i) = ubound(arr,i) - lbound(arr,i) + 1
  enddo
#endif

  call h5pcreate_f(H5P_DATASET_CREATE_F, prp_id, hdf_err) 
  ! Check if dataset should be zipped.
  if(present(gzip))then
     if( NDIMS .gt.0 )then
        if(HDF_VERBOSITY .ge. 1)write(*,*) &
             '[hdf5_write_dataset] Dataset ',trim(name),' will be compressed'
        ! May need to choose a less silly chunking scheme - currently uses
        ! one big chunk!
#ifndef SCALAR
        call h5pset_chunk_f(prp_id, NDIMS , dimensions, hdf_err) 
        call h5pset_deflate_f(prp_id,gzip,hdf_err) 
        if(hdf_err.lt.0)write(*,*) &
             '[hdf5_write_dataset] Unable to gzip data. Will write uncompressed data.'
#endif
     else
        write(*,*) &
             '[hdf5_write_dataset] Cannot compress a scalar! Continuing...'
     end if
  end if

  ! Parse name string, create groups if necessary:
  thestring = name
  i = index(thestring,slash)
  if (i .eq. 1) then
     !strip leading slash
     thestring = thestring(2:LEN_STR)
     i = index(thestring,slash)
  endif

  starti = 1
  endi   = i

  n_nest = 0
  do while(i > 0)
     !Get the name of the current group:
     currentgroup = thestring(1:i-1)

     !Cut the current group from the start of the string:
     thestring = thestring(i+1:LEN_STR)

     !Update position of the first /
     i = index(thestring,slash)

     !Create groups:
     call h5eset_auto_f(0, hdf_err)
     call h5gcreate_f(loc_id, currentgroup, group_id, hdf_err, 0)
     if (hdf_err .eq. 0 .and. HDF_VERBOSITY .ge. 1) then
        write(*,*)	'[hdf5_write_dataset] Creating group ',trim(currentgroup)
     endif
     if(HDF_ERROR_SUPPRESS .eq. 0)call h5eset_auto_f(1, hdf_err)

     call h5gopen_f(loc_id, currentgroup, group_id, hdf_err)
     if(hdf_err.lt.0)call hdf5_abort( &
          'Unable to open group in write_data()!', &
          name=currentgroup,fname=fname_table(ifile))

     n_nest = n_nest + 1
     id_chain(n_nest) = group_id
     loc_id = group_id
  enddo

  !If we have got this far then the remainder of the string is the
  !dataset name:
  currentdataset = thestring

  if (HDF_VERBOSITY .ge. 1) then
     write(*,*)'[hdf5_write_dataset] Writing dataset ',trim(currentdataset)
  endif

  ! Now we can write the data:
#ifndef SCALAR
  call h5screate_simple_f(NDIMS,dimensions,dspace_id,hdf_err)
#else
  call h5screate_f(H5S_SCALAR_F, dspace_id, hdf_err)
#endif
  if(hdf_err.lt.0)call hdf5_abort( &
       'Unable to open dataspace in write_data()!', &
       name=name,fname=fname_table(ifile))

#ifndef STRING
  call h5tcopy_f( NATIVE_TYPE, dtype_id, hdf_err)
  if(hdf_err.lt.0)call hdf5_abort( &
       'Unable to open datatype in write_data()!', &
       name=name,fname=fname_table(ifile))
#else
    call h5tcopy_f(H5T_NATIVE_CHARACTER, dtype_id, hdf_err)
    call h5tset_size_f(dtype_id, INT(LEN(arr),KIND=SIZE_T), hdf_err) 
#endif

  call h5dcreate_f(loc_id ,currentdataset, dtype_id, &
       dspace_id, dset_id, hdf_err,prp_id)
  if(hdf_err.lt.0)call hdf5_abort( &
       'Unable to open dataset in write_data()!', &
       name=name,fname=fname_table(ifile))

  ! Fortran routine to write the dataset - can't do integer*8.
  ! call h5dwrite_f(dset_id, NATIVE_TYPE, arr, dimensions, hdf_err)
  ! C routine to write the dataset
  call write_hdf5_dataset(dset_id, dtype_id , arr, hdf_err)

  if(hdf_err.lt.0)call hdf5_abort( &
       'Unable to write dataset in write_data()!', &
       name=name,fname=fname_table(ifile))

  !Close the nest of groups:
  do i=n_nest,1,-1
     call h5gclose_f(id_chain(i),hdf_err)
     if(hdf_err .lt. 0) call hdf5_abort( &
          'Unable to close group in write_data()!', &
          name=name,fname=fname_table(ifile))
  enddo

  call h5dclose_f(dset_id,hdf_err)
  call h5tclose_f(dtype_id,hdf_err)
  call h5sclose_f(dspace_id,hdf_err)
  call h5pclose_f(prp_id,hdf_err)

end subroutine SUB_NAME


