;*****************************************************************
;*                       PINOCCHI0  V2.2                         *
;*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
;*****************************************************************
;
;This code was written by
;Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
;Copyright (C) 2008
;
;web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
;
;This program is free software; you can redistribute it and/or modify
;it under the terms of the GNU General Public License as published by
;the Free Software Foundation; either version 2 of the License, or
;(at your option) any later version.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



; These procedures load pinocchio merger trees from the
; frag.$redshift.$run.halo.out files and plot them

; The .halo. file gives information on the merger history of each halo
; which ever got as massive as 10 particles. At each halo merger, the
; less massive one disappears from the list of existent halos, while
; the more massive, which retains its ID, one acquires all the mass of
; the system. As a consequence, each merger is connected with a halo
; that disappears from the list. A merger tree is connected to each
; existent halo.

; EXAMPLE
; To read the trees of the lcdm01 example and plot the merger tree of
; halo itree, cd to the rundir/ directory and start idl:
; IDL> all_trees = read_tress('lcdm01','0.00')
; NB: the ascii_template procedure will open a widget window
; IDL> itree = [an integer number >1]
; IDL> single_tree = extract_one_tree(itree,0.00,all_trees)
; IDL> plot_merger_tree,single_tree

function read_trees,run,redshift

; This function reads the merger trees of a pinocchio run 
; from the frag.$redshift.$run.halo.out file
;
; input: run      - 6 character label with the run name
;        redshift - 4 character label with the redshift

; The structure all_trees is returned by the function, and contains
; scalar long integer:
; all_trees.nhalos           -> total number of halos
; long integer vectors of lenght all_trees.nhalos:
; all_trees.halo_belong      -> ID of the halo to which the i-th halo belongs at $redshift
; all_trees.halo_ll          -> pointer to the next halo of the tree
; all_trees.merged_with      -> ID of the (larger) halo with which the i-th halo has merged
; all_trees.mass_at_merger   -> halo mass of the i-th halo at the merger time
; all_trees.mass_merged_with -> halo mass of the halo it has merged with (at the merger time)
; real vectors of length all_trees.nhalos:
; all_trees.z_merging        -> redshift of merging
; all_trees.z_appear         -> redshift of appearance of the i-th halo (redshift at which the halo
;                               became as massive as 10 particles or more)


; opens the file and reads all the data
halofile = run+'/frag.'+redshift+'.'+run+'.halo.out'
template = ascii_template(halofile)
data_in_file = read_ascii(halofile,template=template)

; computes the total number of halos
ndataa = size(data_in_file.field01,/dimension)
ndata  = ndataa[0]
nhalos = max(data_in_file.field01)

print,'Number of lines found in the file: ',ndata
print,'Number of halos found in the run: ',nhalos

; defines the structure with all the trees
all_trees = {                         $
  nhalos           : 0L , $
  halo_ID          : lonarr(nhalos) , $
  halo_belong      : lonarr(nhalos) , $
  halo_ll          : lonarr(nhalos) , $
  merged_with      : lonarr(nhalos) , $
  mass_at_merger   : lonarr(nhalos) , $
  mass_merged_with : lonarr(nhalos) , $
  z_merging        : fltarr(nhalos) , $
  z_appear         : fltarr(nhalos)}

all_trees.nhalos = nhalos
for i=0l,ndata-1 do begin
    ID = data_in_file.field01[i]
    all_trees.halo_belong[ID-1]      = data_in_file.field02[i]
    all_trees.halo_ll[ID-1]          = data_in_file.field03[i]
    all_trees.merged_with[ID-1]      = data_in_file.field04[i] 
    all_trees.mass_at_merger[ID-1]   = data_in_file.field05[i] 
    all_trees.mass_merged_with[ID-1] = data_in_file.field06[i] 
    all_trees.z_merging[ID-1]        = data_in_file.field08[i] 
    all_trees.z_appear[ID-1]         = data_in_file.field10[i] 
endfor

data_in_file=0

return,all_trees

end

function extract_one_tree,itree,final_z,all_trees

; It extracts the merger tree n. itree from the list
; and puts it in the single_tree structure, which is analogous to all_trees.

; checks that the itree value makes sense
if itree GT all_trees.nhalos || itree LE 0 then begin
    print,'wrong itree value',all_trees.nhalos
    return,0l
endif

; checks that the tree exists in the list
if all_trees.mass_at_merger[itree-1] EQ 0l then begin
    print,'This halo is too small '
    return,0l
endif

; if itree is a satellite gives the merger tree of the parent halo
if all_trees.halo_belong[itree-1] NE itree then begin
    itree =  all_trees.halo_belong[itree-1]
    print,'This halo is a satellite, the parent is ',itree
endif

; first step: counts the number of branches in the tree and creates a
; vector that points from the 'global' halo ID to the position in the
; vectors of the single tree

index     = lonarr(all_trees.nhalos)

ipos      = itree
nbranches = 1
while (all_trees.halo_ll[ipos-1] NE itree) do begin
    ipos          = all_trees.halo_ll[ipos-1]
    index[ipos-1] = nbranches-1
    nbranches     = nbranches+1
endwhile
index[itree-1]=nbranches-1

print,'Number of branches of the tree: ',nbranches

; second step: defines the structure

single_tree = {                          $
  nbranches        : 0l                , $
  halo_ID          : lonarr(nbranches) , $
  halo_belong      : lonarr(nbranches) , $
  halo_ll          : lonarr(nbranches) , $
  merged_with      : lonarr(nbranches) , $
  mass_at_merger   : lonarr(nbranches) , $
  mass_merged_with : lonarr(nbranches) , $
  z_merging        : fltarr(nbranches) , $
  z_appear         : fltarr(nbranches)}

; third step: extracts the tree by scrolling the linking list; the
; single_tree.merged_with vector points to the position in the
; single_tree lists and not to the 'global' ID

single_tree.nbranches = nbranches
ipos    = all_trees.halo_ll[itree-1]

for ibranch=0,nbranches-1 do begin
    single_tree.halo_ID[ibranch]          = all_trees.halo_ID[ipos-1]
    single_tree.halo_belong[ibranch]      = all_trees.halo_belong[ipos-1]     
    single_tree.halo_ll[ibranch]          = all_trees.halo_ll[ipos-1]         
    if ibranch NE nbranches-1 then $
      single_tree.merged_with[ibranch]      = index[all_trees.merged_with[ipos-1]-1] $
    else $
      single_tree.merged_with[ibranch]      = -1l
    single_tree.mass_at_merger[ibranch]   = all_trees.mass_at_merger[ipos-1]  
    single_tree.mass_merged_with[ibranch] = all_trees.mass_merged_with[ipos-1] 
    single_tree.z_merging[ibranch]        = all_trees.z_merging[ipos-1]       
    single_tree.z_appear[ibranch]         = all_trees.z_appear[ipos-1]        
    ipos    = all_trees.halo_ll[ipos-1]
endfor

; the merging redshift of the parent is set to the final redshift.
single_tree.z_merging[nbranches-1]=final_z

index=0
return,single_tree

end


pro plot_merger_tree,single_tree

; This procedure makes a plot of the merger tree specified by
; the structure single_tree.  In the plot, the y-axis gives the
; redshift, the crosses mark the appearance of a halo, the circles
; mark a merger, whose dimension is scaled with its mass to the 1/3.
; The main progenitor is always to the left, and is marked by a thicker line.


; initialization
x=fltarr(2)
y=fltarr(2)
large = 100.

xcoord_a = fltarr(single_tree.nbranches)
position = lonarr(single_tree.nbranches)

delta    = 1./(single_tree.nbranches-1.)

fixed_merged_with = single_tree.merged_with
fixed_merged_with[single_tree.nbranches-1]=single_tree.nbranches-1

a = findgen(17) * (!PI*2/16.)  
usersym, cos(a), sin(a)

; the mergers are sorted in redshift
sorted_z = sort(single_tree.z_merging)

; the position vector gives the ordering of the halos, assigned as follows:
; starting from z=0 and scrolling the list of mergers in redshift order, 
; each merging halo is inserted to the right of the halo it merges with
position[single_tree.nbranches-1]=0l
for i_sorted = 1,single_tree.nbranches-1 do begin
    insert_halo,sorted_z[i_sorted],fixed_merged_with[sorted_z[i_sorted]],position
endfor

; the initial positions of the halos in the graph are scaled to the position vector
xcoord_a = position*delta
old_a    = xcoord_a

; plot of the axes
zup = max(single_tree.z_appear)*1.05
x[0]=-100 & x[1]=x[0] & y[0]=x[0] & y[1]=x[0]
plot,x,y,xrange=[-0.1,1.1],/xstyle, $
  yrange= [-0.1,zup],/ystyle,psym=7

; sorts the mergers in chronological order
sorted_z = sort(-single_tree.z_merging)
last_z   = large

; loop on the mergers
for i_sorted=0,single_tree.nbranches-1 do begin

    i_sat  = sorted_z[i_sorted]
    i_main = fixed_merged_with[i_sat]
    z      = single_tree.z_merging[i_sat]
    if z lt 0 then z=0

; plots the symbol for the merger, scaled to the halo size
    r=(float(single_tree.mass_at_merger[i_sat]+single_tree.mass_merged_with[i_sat])/ $
       float(single_tree.mass_at_merger[single_tree.nbranches-1]))^0.3333
    plots,xcoord_a[fixed_merged_with[i_sat]],single_tree.z_merging[i_sat],psym=8,symsize=3*r,thick=3*r ;,color=(i_main*156+99)/(single_tree.nbranches-1)

; loop on all the progenitors
    for i_loop=0,single_tree.nbranches-1 do begin

; cycle if the progenitor has disappeared or if in case of muliple mergers
        if z ge single_tree.z_merging[i_loop] && z lt last_z then begin

; at the first merger join all the halos that have appeared before to the current time
            if last_z eq large && z le single_tree.z_appear[i_loop] then begin
                if i_loop eq i_sat then x[1]=xcoord_a[fixed_merged_with[i_loop]] else x[1]= xcoord_a[i_loop]
                x[0]=old_a[i_loop]
                y[0]=single_tree.z_appear[i_loop] & y[1]=z

                if i_loop eq single_tree.nbranches-1 then lw=3 else lw=1
; plot the appearance times
                plots,x[0],y[0],psym=7 ;,color=(i_loop*156+99)/(single_tree.nbranches-1)
; join them to the current redshift
                oplot,x,y,psym=0,thick=lw ;,color=(i_loop*156+99)/(single_tree.nbranches-1)
            endif
 
; after the first merger consider all halos that still exist
            if last_z lt large && z le single_tree.z_appear[i_loop] then begin
; join their position to the next merger
                x[1]=old_a[i_loop]
                y[1]=last_z
                if single_tree.z_appear[i_loop] lt last_z then begin
; if the halo has just appeared plot the symbol
                    y[1]=single_tree.z_appear[i_loop]
                    plots,x[1],y[1],psym=7 ;,color=(i_loop*156+99)/(single_tree.nbranches-1)
                endif
                if i_loop eq i_sat then begin 
                    x[0]=xcoord_a[fixed_merged_with[i_loop]]
                    y[0]=single_tree.z_merging[i_loop]
                endif else begin
                    x[0]=xcoord_a[i_loop]
                    y[0]=z
                endelse
                if i_loop eq single_tree.nbranches-1 then lw=3 else lw=1
                oplot,x,y,psym=0,thick=lw ;,color=(i_loop*156+99)/(single_tree.nbranches-1)
            endif
        endif
; end of loop on halos
    endfor
    last_z=z

; shrink the position of halos so as to have an inverted pyramid-like plot
    old_a = xcoord_a
    shrink,i_sat,delta,xcoord_a

endfor

;nanna

end

pro insert_halo,to_insert,after_this,positions

value   = positions[after_this]
indices = where(positions gt value, count)
if count ne 0 then positions[indices]   = positions[indices]+1
positions[to_insert] = value+1

end

pro shrink,sat,delta,xcoord

add = where (xcoord lt xcoord[sat],count)
if count gt 0 then xcoord(add)=xcoord(add)+0.5*delta
sub = where (xcoord gt xcoord[sat],count)
if count gt 0 then xcoord(sub)=xcoord(sub)-0.5*delta
xcoord[sat]=-1.

end

;pro scroll,v,pos
;
;s=size(v)
;
;for i=pos,s[1]-2 do v[i]=v[i+1]
;v[s[1]-1]=-1
;
;end

