!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

!
! Cosmological tools - F90 version
!


! *******************
! LINEAR GROWING MODE
! *******************

module grow_module
  use numbers
  integer(kind=singleI),parameter::nbin=100
  real(kind=doubleR),dimension(nbin)::amat,bmat
  logical,save::firsttime=.true.
end module grow_module


function grow(z)

! Linear growing mode, normalized as grow=1/(1+z) at high redshift

! DIMENSIONLESS

  use pinocchio_common
  use grow_module
  implicit none
  real(kind=doubleR) :: grow
  real(kind=doubleR),intent(in)::z

! local variables
  real(kind=doubleR):: aa,bb

  if (firsttime) then
     call grow_start()
     firsttime=.false.
  endif

  if (Omega0 == 1.) then
     grow=1./(1.+z)
  else
     aa=log10(1./(1.+z))
     call interpola(amat,bmat,nbin,aa,bb)
     grow=10.**bb
  endif
  return
end function grow


function inverse_grow(b)

! Redshift corresponding to the growing mode b

! DIMENSIONLESS

  use pinocchio_common
  use grow_module
  implicit none
  real(kind=doubleR) :: inverse_grow
  real(kind=doubleR),intent(in)::b

! local variables
  real(kind=doubleR)::loga,logb
      

  if (firsttime) then
     call grow_start()
     firsttime=.false.
  endif

  if (Omega0 == 1.) then
     inverse_grow=1.d0/b-1.d0
  else 
     logb=log10(max(b,1.d-40))
     call interpola(bmat,amat,nbin,logb,loga)
     inverse_grow=10.**(-loga)-1.
  endif

  return
end function inverse_grow


subroutine grow_start()

! Calculates the integral for grow(z) in a Lambda Universe

  use pinocchio_common
  use grow_module
  implicit none

! local variables
  integer(kind=singleI)::i
  real(kind=doubleR):: tt,hh,bb,grow,z,tau
  real(kind=doubleR),parameter:: ttmin=-2.5,dtt=2.7
  real(kind=doubleR),external :: grow_integrand

  if (OmegaLambda > 0 .and. abs(Omega0+OmegaLambda-1.d0) > 1.d-6) then
     call abort_code('Wrong cosmology in grow_start')
  endif

  if (OmegaLambda == 0. .and. Omega0 < 0.) then

     do i=1,nbin
        z=float(nbin-i)/(nbin-1.)*60.
        tau=sqrt((1.+z)/(1./Omega0-1.)+1.)
        grow=2.5*(1.+3.*(tau*tau-1.)*(1.+.5*tau*log((tau-1.)/(tau+1.))))/ &
             (1./Omega0-1.)
        amat(i)=log10(1./(1.+z))
        bmat(i)=log10(grow)
     enddo

! normalized to b(z=0)=1
     call interpola(amat,bmat,nbin,0.,bb)
     bmat=bmat-bb

  else

     do i=1,nbin
        tt=10.**(ttmin+(i-1.)*dtt/(nbin-1.))
        hh=1./tanh(tt)
        call integrate(grow_integrand,hh,max(100.d0,hh*100.),bb)
        amat(i)=(1./Omega0-1.)**(-1./3.)*sinh(tt)**(2./3.)
        bmat(i)=hh*bb*5./3./(1./Omega0-1.)**(1./3.)
        amat(i)=log10(amat(i))
        bmat(i)=log10(bmat(i))
     enddo

! normalized to b(z=0)=1
     call interpola(amat,bmat,nbin,0.,bb)
     bmat=bmat-bb

  endif
  return
end subroutine grow_start


function grow_integrand(x)

! Integrand for grow_start

  use numbers
  implicit none
  real(kind=doubleR) :: grow_integrand
  real(kind=doubleR),intent(in)::x

  grow_integrand=1./(x*x*(x*x-1)**(.33333333333333d0))

  return
end function grow_integrand


! ***********************
! MASS VIARIANCE (LAMBDA)
! ***********************


module Lambda_module
! module for Lambda and Radius functions
  use numbers
  real(kind=doubleR)::r2
  integer(kind=singleI),parameter::nbin=200
  real(kind=doubleR),dimension(nbin):: rv,lamv,lst,rst,dldr
  logical::firsttime=.true.
end module Lambda_module


function aLambda_once(r)

! Linear variance on scale r, with smoothing as specified by FILTER;
! performs the integral each time
! r: Mpc

! DIMENSIONLESS

  use pinocchio_common
  use Lambda_module
  implicit none
  real(kind=doubleR) :: aLambda_once
  real(kind=doubleR),intent(in)::r

  real(kind=doubleR):: out,limit
  real(kind=doubleR),external::Lambda_integrand

  r2=r
  if (filter.eq.0) then
     limit=log(5./r)
  else if (filter.eq.1) then
     limit=log(1./r)
  else
     limit=log(10.9041/r)
  endif

  call integrate(Lambda_integrand,-6.9d0,limit,out)
  aLambda_once=out

  return
end function aLambda_once


function Lambda_integrand(logk)

! Integrand for evaluation of Lambda

  use Lambda_module
  implicit none
  real(kind=doubleR) :: Lambda_integrand
  real(kind=doubleR),intent(in)::logk
  real(kind=doubleR),external::power,window

  real(kind=doubleR):: k

  k=exp(logk)

  if (k.lt.1.d8) then         
     Lambda_integrand=power(k)*window(k*r2)**2*k**3
  else
     Lambda_integrand=0.
  endif

  return
end function Lambda_integrand


subroutine Lambda_start()

! Calculates the integrals for the mass variance

  use pinocchio_common
  use Lambda_module
  implicit none

  real(kind=doubleR),external::aLambda_once

  real(kind=doubleR),parameter:: rmin=-6.d0, dr=0.05d0
  integer(kind=singleI)::i
  real(kind=doubleR)::r
  logical,save::done=.false.

  if (done) return

!  open(34,file='Lambda_start.out',status='unknown')
  do i=1,nbin

     rv(i)=(rmin+(i-.5)*dr)
     r=10.**rv(i)
     lamv(i)=log10(max(aLambda_once(r),tiny(r)))

     rst(nbin-i+1)=rv(i)
     lst(nbin-i+1)=lamv(i)

!     write(34,'(1x,i5,9(1x,e12.6))')i,rv(i),lamv(i)

  enddo

  do i=1,nbin
     if (i.eq.1) then
        dldr(i)=(lamv(i+1)-lamv(i))/(rv(i+1)-rv(i))
     else if (i.eq.nbin) then
        dldr(i)=(lamv(i)-lamv(i-1))/(rv(i)-rv(i-1))
     else
        dldr(i)=(lamv(i+1)-lamv(i-1))/(rv(i+1)-rv(i-1))
     endif
  enddo

  done=.true.
  return
end subroutine Lambda_start


function aLambda(r)

! Linear variance on scale r, with smoothing as specified by Lambda_start
! r: Mpc 

! DIMENSIONLESS

  use pinocchio_common
  use Lambda_module

  implicit none
  real(kind=doubleR) :: aLambda
  real(kind=doubleR),intent(in)::r

! local variables
  real(kind=doubleR)::rlog
 
  if (firsttime) then

     call Lambda_start()
     firsttime=.false.
  endif

! smallest radius is assumed to be 1 pc
  rlog=log10(max(r,1.d-6))
  call interpola(rv,lamv,nbin,rlog,aLambda)
  aLambda=10.**aLambda

  return
end function aLambda



function Radius(v)

! Radius corresponding to variance v, with smoothing as specified 
! by Lambda_start

! DIMENSION: Mpc

  use pinocchio_common
  use Lambda_module
  implicit none
  real(kind=doubleR) :: Radius
  real(kind=doubleR),intent(in)::v
  
! local variables
  real(kind=doubleR)::vlog

  if (firsttime) then
     call Lambda_start()
     firsttime=.false.
  endif

  vlog=log10(max(v,tiny(v)))
  call interpola(lst,rst,nbin,vlog,Radius)
  Radius=10.**Radius

  return
end function Radius

! ************
! SCALE FACTOR
! ************


module scale_module
  use numbers
  integer(kind=singleI),parameter::nbin=100
  real(kind=doubleR),dimension(nbin)::time_v,scale_v
end module scale_module

subroutine scale_start()

! Calculates the vectors necessary to evaluate scale(t)

  use pinocchio_common
  use scale_module
  implicit none
  real(kind=doubleR),parameter::amin=-3.,dalog=3.5
  real(kind=doubleR)          ::alog,z
  real(kind=doubleR),external ::time
  integer(kind=singleI)::i

  do i=1,nbin
     alog=amin+(i-1.)*dalog/(nbin-1.)
     z=10.**(-alog)-1.
     time_v(i)=log10(time(z))
     scale_v(i)=alog
  enddo

  return
end subroutine scale_start


function scale(t)

! Scale factor as a function of time

! DIMENSIONLESS

  use pinocchio_common
  use scale_module
  implicit none
  real(kind=doubleR) :: scale
  real(kind=doubleR),intent(in)::t
  real(kind=doubleR)::tlog
  logical,save::firsttime=.true.

  if (firsttime) then
     call scale_start()
     firsttime=.false.
  endif

  tlog=log10(max(t,1.d-40))
  call interpola(time_v,scale_v,nbin,tlog,scale)
  scale=10.**scale

  return

end function scale


! ************************
! COSOMOLOGICAL QUANTITIES
! ************************


function Omega(z)

! Cosmological density parameter as a function of redshift

! DIMENSIONLESS

  use pinocchio_common
  implicit none
  real(kind=doubleR) :: Omega
  real(kind=doubleR),intent(in)::z

  if (Omega0 == 1..and.OmegaLambda == 0.) then
     Omega=1.d0
  else if (omega0 < 1..and.OmegaLambda == 0.) then
     Omega=1./(1.+(1./Omega0-1.)/(1.+z))
  else if (omega0 < 1.and.abs(Omega0+OmegaLambda-1.).lt.1.d-4) then
     Omega=1./(1+(1./omega0-1.)/(1.+z)**3)
  else
     print *,'Wrong cosmology in function OMEGA!'
     print *,'Omega_0 = ',Omega0
     print *,'Omega_Lambda = ',OmegaLambda
     stop
  endif

  return
end function omega

function Hubble(z)

! Hubble parameter as a function of redshift

! DIMENSION: km/s/Mpc

  use pinocchio_common
  implicit none
  real(kind=doubleR) :: Hubble
  real(kind=doubleR),intent(in)::z
  real(kind=doubleR),external::Omega

  Hubble=100.*Hubble100*(1.+z)**1.5*sqrt(Omega0/Omega(z))

  return
end function Hubble


function hubble_Gyr(z)

! Hubble parameter as a function of redshift

! DIMENSION: Gyr^-1

  use numbers
  implicit none
  real(kind=doubleR) :: hubble_Gyr
  real(kind=doubleR),intent(in)::z
  real(kind=doubleR),external::Hubble

  Hubble_gyr=Hubble(z)/977.76d0

  return
end function hubble_Gyr


function time(z)

! Cosmological time as a function of redshift

! DIMENSION: Gyr

  use pinocchio_common
  implicit none
  real(kind=doubleR) :: time
  real(kind=doubleR),intent(in)::z
  real(kind=doubleR)::time0
  real(kind=doubleR),external::asinh,acosh

  time0=9.7776/Hubble100
  if (Omega0 == 1..and.OmegaLambda == 0.) then
     time=time0/1.5/(1.+z)**1.5
  else if (Omega0 < 1..and.OmegaLambda == 0.) then
     time=time0*Omega0*.5*(1.-Omega0)**(-1.5)*(-acosh((z*Omega0-Omega0+2.) &
          /(z*Omega0+Omega0))+2.*sqrt((1.-Omega0)*(Omega0*z+1.))/Omega0/(1.+z))
  else if (Omega0 < 1.and.abs(Omega0+OmegaLambda-1).lt.1.d-4) then
     time=time0/1.5/sqrt(OmegaLambda)*asinh(sqrt(OmegaLambda/Omega0)/(1.+z)**1.5)
  else
     print *,'Wrong cosmology in function TIME!'
     print *,'Omega_0 = ',omega0
     print *,'Omega_Lambda = ',OmegaLambda
     stop
  endif

  return
end function time


function fomega(z)

! Peebles' f(Omega) function, dlogb/dloga

! DIMENSIONLESS

  use numbers
  implicit none
  real(kind=doubleR) :: fomega
  real(kind=doubleR),intent(in)::z
  real(kind=doubleR),external::grow

  real(kind=doubleR)::dz,db

  dz=.02
  db=.5*(grow(z-dz)-grow(z+dz))/dz
  fomega=db*(1.+z)/grow(z)

  return
end function fomega


function aMass(v)

! Mass at linear variance V

! DIMENSION: M_sun

  use pinocchio_common
  implicit none
  real(kind=doubleR) :: aMass
  real(kind=doubleR),intent(in)::v
  
  real(kind=doubleR)::r
  real(kind=doubleR),external::Radius

  r=Radius(v)
  select case (filter)
  case (0)
     aMass=(2*acos(-1.))**1.5*r**3
  case (1)
     aMass=6.*acos(-1.)**2*r**3
  case (2)
     aMass=4.*acos(-1.)*r**3/3.
  case default
     print *,'Unknown smoothing in function AMASS'
     stop
  end select

  aMass=aMass*2.778d11*Hubble100**2*Omega0

  return
end function aMass


function aM_star(z)

! Typical collapsing mass at redshift z

! DIMENSION: M_sun

  use pinocchio_common
  implicit none
  real(kind=doubleR) :: aM_star
  real(kind=doubleR),intent(in)::z
  real(kind=doubleR),external::aMass,grow

  aM_star=aMass((delta_c/grow(z))**2)
  
  return
end function aM_star




! **************
! POWER SPECTRUM
! **************


function window(x)

! Window function

!     N.B:  x =: kr !!!

!     filter = 0: Gaussian smoothing
!     filter = 1: SKS smoothing
!     filter = 2: top-hat smoothing

! DIMENSIONLESS

  use pinocchio_common
  implicit none
  real(kind=doubleR) :: window
  real(kind=doubleR),intent(in)::x
  real(kind=doubleR),external::theta

  select case (filter)
     case (0)                       ! Gaussian
        window=exp(-x*x/2.)
     case (1)                       ! sharp k-space
        window=theta(1.-x)
     case (2)                       ! top-hat
        if (x.lt.3.e-3) then
           window=1.
        else if (x.lt.10.9041) then
           window=3.*(sin(x)/x**3-cos(x)/x**2)
        else
           window=0.
        endif
     case default
        call abort_code ('unknown filter in function window')
   end select

  return
end function window


function power(fk)

! PowerSpectrum = [-3,1]:  power-law power spectrum
! PowerSpectrum >= 10.: Power spectrum according to Efstathiou, Bond & White (1992)   
!          (Gamma=PowerSpectrum-10. if PowerSpectrum>10)
! PowerSpectrum <=-10.: Power spectrum according Bardeen et al. (1986)
!          (Gamma=-PowerSpectrum-10. if PowerSpectrum<-10)

! NEW PowerSpectrum=7 : wmap cmbfast result

! N.B.: fk IS SUPPOSED TO BE IN Mpc^-1

  use pinocchio_common

  implicit none
  real(kind=doubleR) :: power
  real(kind=doubleR),intent(in)::fk

! local variables
  real(kind=doubleR)::pk,gamma,q
  logical,save::firsttime=.true.
  integer(kind=singleI)::nn
!  integer(kind=singleI),parameter::nwmap=201
!  real(kind=doubleR),dimension(nwmap):: kw,Pw
!  real(kind=doubleR)::Trans,kk,TT,T1,lfk,lastk,lastp
!  character(len=80)::string
! Eisenstein & Hui
  real(kind=doubleR),parameter::Teta_27=1.0104
  real(kind=doubleR)::Omegac,Oh2,b1,b2,zd,Rd,zeq,Req,keq,s,ks,alc,bec,f,Tc,beb,bno,kst,ksi,Tb,Tr,y,alb,Ob2
  real(kind=doubleR),external::T0
  if (fk.eq.0.) then
     power=0.
     return
  endif

  if (PowerSpectrum<=1.1 .and. PowerSpectrum>=0.9) then
! Eisenstein & Hu fit, PowerSpectrum gives the primordial index
     Omegac = Omega0-Omegab
     Oh2 = Omega0*Hubble100**2
     Ob2 = Omegab*Hubble100**2
     b1  = 0.313*Oh2**(-0.419)*(1+0.607*Oh2**0.674)
     b2  = 0.238*Oh2**0.223
     zd  = 1291.*Oh2**0.251*(1.+b1*Ob2**b2)/(1.+0.659*Oh2**0.828)
     Rd  = 31.5*Ob2/(Teta_27**4*0.001*zd)
     zeq = 2.5e4*Oh2/Teta_27**4
     Req = 31.5*Ob2/(Teta_27**4*0.001*zeq)
     keq = 7.46d-2*Oh2/Teta_27**2
     s   = 1.633*log((sqrt(1.+Rd)+sqrt(Rd+Req))/(1+sqrt(Req)))/(keq*sqrt(Req))  ! 2sqrt(6)/3=1.633
     ks  = fk*s
     q   = fk*Teta_27**2/Oh2
     alc = ((46.9*Oh2)**0.670*(1.+(32.1*Oh2)**(-0.532)))**(-Omegab/Omega0) * &
           ((12.0*Oh2)**0.424*(1.+(45.0*Oh2)**(-0.582)))**(-(Omegab/Omega0)**3)
     bec = 1./(1.+(0.944/(1.+(458.*Oh2)**(-0.708)))*((Omegac/Omega0)**((0.395*Oh2)**(-0.0266))-1.))
     f   = 1./(1+(ks/5.4)**4)
     Tc  = f*T0(q,1.,bec)+(1.-f)*T0(q,alc,bec)
     beb = 0.5+Omegab/Omega0+(3.-2.*Omegab/Omega0)*sqrt((17.2*Oh2)**2+1.)
     bno = 8.41*Oh2**0.435
     kst = ks/(1.+(bno/ks)**3)**(0.3333)
     ksi = 1.6*Ob2**0.52*Oh2**0.73*(1.+(10.4*Oh2)**(-0.95))
     y   = (1.+zeq)/(1+zd)
     alb = 2.07*keq*s*(1+Rd)**(-0.75)*(y*(-6.*sqrt(1.+y)+(2.+3.*y)*log((sqrt(1.+y)+1.)/(sqrt(1.+y)-1.))))
     Tb  = (T0(q,1.,1.)/(1.+(ks/5.2)**2)+alb/(1.+(beb/ks)**3)*exp(-(fk/ksi)**1.4))*sin(kst)/kst
     Tr  = (Omegab*Tb+Omegac*Tc)/Omega0
     pk  = fk**PowerSpectrum*Tr**2

  else if (PowerSpectrum.ge.-3..and.PowerSpectrum.le.1.) then
     if (fk.gt.5.d-4.and.fk.lt.5.d4) then
        pk=fk**PowerSpectrum
     else
        pk=0.
     endif
  else if (PowerSpectrum.ge.10.) then
     if (PowerSpectrum.eq.10.) then
        gamma=Omega0*Hubble100
     else
        gamma=PowerSpectrum-10.
     endif
     q=fk/Hubble100/gamma
     pk=q/(1.+(6.4*q+(3.*q)**1.5+(1.7*q)**2)**1.13)**(2./1.13)
  else if (PowerSpectrum.le.-10.) then
     if (PowerSpectrum.eq.-10.) then
        gamma=Omega0*Hubble100
     else
        gamma=-PowerSpectrum-10.d0
     endif
     q=fk/Hubble100/gamma
     pk=log(1.+2.34*q)**2*(1.+3.89*q+(16.1*q)*(16.1*q)+(5.64*q)**3+ &
          (6.71*q)**4)**(-.5)/q
  else
     call abort_code('Wrong PowerSpectrum in function POWER')

  endif
  power=pk*csp

  return
end function power


function T0(q,a,b)
  use numbers
  implicit none
  real(kind=doubleR) :: T0
  real(kind=doubleR),intent(in)::q,a,b
  real(kind=doubleR)::ll,C

  ll = log(exp(1.)+1.8*b*q)
  C  = 14.2/a+386./(1.+69.9*q**1.08)
  T0 = ll/(ll+C*q**2)
  return
end function T0


subroutine normalization()

! Normalization of the power spectrum to sigma_8

  use pinocchio_common
  implicit none
  real(kind=doubleR),external::aLambda_once
  
  integer(kind=singleI)::old_filter

  old_filter = filter
  filter     = 2
  csp        = 1.d0
  csp        = Sigma8*Sigma8/aLambda_once(8.d0/Hubble100)
  filter     = old_filter

  return
end subroutine normalization



! *************
! MASS FUNCTION
! *************


function Size(m)

! Radius corresponding to mass m
! m: M_sun

! DIMENSION: Mpc

  ! use run_module
  use pinocchio_common
  use pinocchio_constants
  implicit none
  real(kind=doubleR) :: Size
  real(kind=doubleR),intent(in)::m

  real(kind=doubleR)::rho

  rho = 2.778d11*Hubble100**2*Omega0
  select case (filter)
  case (0)
     Size=(m/(2.d0*pi)**1.5d0/rho)**(1.d0/3.d0)
  case (1)
     Size=(m/(6.d0*pi**2*rho))**(1.d0/3.d0)
  case (2)
     Size=(m/(4.d0*pi*rho/3.d0))**(1.d0/3.d0)
  case default
     print *,'Unknown smoothing in function SIZE'
     stop
  end select

  return
end function Size


function Variance(m)

! Variance corresponding to mass m
! m: M_sun

! DIMENSIONLESS

  ! use run_module
  use pinocchio_common
  use pinocchio_constants
  implicit none
  real(kind=doubleR) :: Variance
  real(kind=doubleR),intent(in)::m
  real(kind=doubleR),external::alambda

! local variables
  real(kind=doubleR)::r,rho

  rho = 2.778d11*Hubble100**2*Omega0
  select case (filter)
  case (0)
     r=(m/(2.d0*pi)**1.5d0/rho)**(1.d0/3.d0)
  case (1)
     r=(m/(6.d0*pi**2*rho))**(1.d0/3.d0)
  case (2)
     r=(m/(4.d0*pi*rho/3.d0))**(1.d0/3.d0)
  case default
     stop 'Unknown smoothing in function VARIANCE'
  end select

  Variance=aLambda(r)

  return
end function Variance


function dLambda(r)

! d Log Lambda/d Log r on scale r, with smoothing as specified by Lambda_start
! r: Mpc

! DIMENSIONLESS

  ! use run_module
  use pinocchio_common
  use Lambda_module
  implicit none
  real(kind=doubleR) :: dLambda
  real(kind=doubleR),intent(in)::r

! local variables
  real(kind=doubleR)::rlog

  if (firsttime) then
     call Lambda_start()
     firsttime=.false.
  endif

  rlog=log10(max(r,tiny(r)))
  call interpola(rv,dldr,nbin,rlog,dLambda)

  return
end function dLambda


function dodl(v)

! dOmega/dLambda

!     flag_ps=0:  Press  & Schechter (1974)
!     flag_ps=1:  Sheth & Tormen (2001)
!     flag_ps=2:  Jenkins et al. (2001)
!     flag_ps=3:  Warren et al. (2006)
!     flag_ps=4:  Reed et al. (2007)

! DIFFERENTIALS: dLambda
! DIMENSIONLESS

  ! use run_module
  use pinocchio_common
  implicit none
  real(kind=doubleR) :: dodl
  real(kind=doubleR),intent(in)::v

! local variables
  real(kind=doubleR)::espo,sv,espo2,espo3,c

  sv=sqrt(v)
  c=0.39894228d0             ! c=1./sqrt(2.*pi)
  espo=delta_c/sv
  
  select case (flag_ps)
  case (0)
     dodl=exp(-.5d0*espo**2)*espo*c
  case (1)
     espo2=0.84083292d0*espo
     dodl=0.3222d0*c*espo2*exp(-.5d0*espo2**2)*(1.d0+ 1.d0/espo2**0.6d0)
  case (2)
     espo3=-log(sv)
     dodl=.5d0*0.315d0*exp(-abs(espo3+0.61d0)**3.8d0)
  case (3)
     dodl=0.5d0*0.7234d0*(sv**(-1.625d0)+0.2538d0)*exp(-1.1982d0/v)
  case (4)
     espo2=0.84083292d0*espo
     dodl=0.3222d0*c*espo2*exp(-.54d0*espo2**2)*(1.d0+ 1.d0/espo2**0.6d0 + 0.2d0*exp(-((-log(sv)-0.4d0)**2/0.72d0)))
  case default
     stop 'Unknown flag_ps in function DODL'
  end select
         
  return
end function dodl


function an_h(mh,z)

! Number density of halos of mass mh (M_sun) at redshift z

! N.B.: LINEAR OUTPUT

! DIMENSION: Mpc^{-3} M_sun^{-1}
! DIFFERENTIALS: dMh

  ! use run_module
  use pinocchio_common
  implicit none
  real(kind=doubleR) :: an_h
  real(kind=doubleR),intent(in)::mh,z
  real(kind=doubleR),external::Size,Variance,grow,dodl,dLambda

! local variables
  real(kind=doubleR)::rho,v,r

  rho  = 2.778d11*Hubble100**2*Omega0
  v    = Variance(mh)*(grow(z))**2
  r    = Size(mh)
  an_h = rho*dodl(v)*abs(dLambda(r)/3.d0)/mh**2

  return
end function an_h



! ***************
! OTHER FUNCTIONS
! ***************


function theta(x)

! Heavyside's theta function

  use numbers
  implicit none
  real(kind=doubleR) :: theta
  real(kind=doubleR),intent(in)::x

  if (x > 0.) then
     theta=1.d0
  else
     theta=0.d0
  endif
  return
end function theta


function asinh(x)

  use numbers
  implicit none
  real(kind=doubleR) :: asinh
  real(kind=doubleR),intent(in)::x

  asinh=log(x+sqrt(x*x+1.d0))

  return
end function asinh


function acosh(x)

  use numbers
  implicit none
  real(kind=doubleR) :: acosh
  real(kind=doubleR),intent(in)::x

  if (x.gt.1.) then
     acosh=log(x+sqrt(x*x-1.d0))
  else
!     print *,'Undefined ACOSH function'
     acosh=0.
  endif

  return
end function acosh
