!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


! This file contains the main program of the fmax-2.2 code

program fmax

! fmax-2.2 version
! - parallel version with MPICH libraries
! - works with fftw-3
! - handles non-cubic boxes with non-power-of-two grid points
! - interfaced with HDF5 and SPRNG packages
! - new computation of the smoothing radii
! - improved output
! - generates the initial conditions
! - coming soon: handles up to 5 nested grids
!
!     ========================================================================
!     Summary: Compute merger history for dark matter haloes in a CDM
!              Cosmology, for a given realisation of a linear density field,
!              generated or specified on a rectangular grid
!     
!              fmax-2.2:    compute inverse collapse time for each grid point
!              fragment-2.2:  use output from fmax-2.2 to determine merger history
!
!     Implementation:
!     This implementation uses the MPI message passing algorithm
!     for use on shared or distrbuted parallel systems. It uses 
!     fftw (see http://www.fftw.org/) to perform fast Fourier transforms.
!     This programme will work in serial by omitting the -DMPI compilation option.
!     Parallel and serial results should be identical. The parallelisation spreads
!     planes to different processors (z-planes in real space, y-planes in Fourier
!     space)
!
!     Detail:
!     This programme computes the inverse collapse time Fmax using
!     ellipsoidal collapse, on a rectangular grid.
!     It also computes the Zeldovich velocities, Velx, Vely, Velz 
!     at the corresponding smoothing scales.
!     Fmax, Velx, Vely and Velz are written to separate files.
!     This data can be input to the fragment-2.2 code to compute the merging history
!     of the forming haloes
!
!     See the DOCUMENTATION file for more details
!     
!     Input is read from file 'fmax.info'


! Fmax-2.2 is organized as follows:
! 
! - The subroutine initialization() does all the necessary initializations 
!   for the run. It is contained in the file fmax-init-2.2.F90
! 
! - The main body of the code is a loop on smoothing radii.  The loop is 
!   divided into four parts.
! 
! - In the first part the second derivative tensor of all particles and for 
!   each nested box is computed.
! 
! - In the second part the collapse time for all particles of the 
!   highest-resolution box is computed.
! 
! - In the third part the displacement tensor of all particles and for each 
!   nested box is computed.
! 
! - In the fourth part the Zel'dovich velocities of all the collapsed
!   particles are computed
! 
! - After the loop on smoothing radii is finished, the density is computed
!   and all the  relevant matrices are saved to the disc.
!
!   calls: initialization                               fmax-init
!          compute_density(ismooth,1)                   fmax-subs
!          compute_second_derivatives(ismooth, igrid)   fmax-subs
!          assign_fftw_variables(ngrid)                 fmax-subs
!          allocate_plane(*)                            fmax-alloc
!          compute_collapse_times(ismooth,0,0,0)        fmax-subs
!          deallocate_plane(*)                          fmax-alloc
!          remove_second_derivatives()                  fmax-subs
!          compute_first_derivatives(ismooth, igrid)    fmax-subs
!          compute_velocities(ismooth,0,0,0)            fmax-subs
!          remove_first_derivatives()                   fmax-subs
!          compute_density(nsmooth,igrid)               fmax-subs
!          write_out()                                  fmax-subs

  use numbers
  use fmax_smoothing_radii ; use fmax_grids
  use pinocchio_mpi        ; use fmax_run
  implicit none

! local variables
  integer(kind=singleI):: ismooth, igrid, ipos, ix, iy, iz
  character(len=120)    :: info

  character(len=24):: fdate
  external fdate
#ifdef MPI
  real(kind=doubleR)   :: cputime(0:10)
#endif

! NB MEMTEST runs a memory test on the machine and then stops
!    IDLTEST runs a test on the derivatives of the deformation tensor and then stops


! **************************************************
!                   STARTING MPI
! **************************************************
#ifdef MEMTEST
#ifndef MPI
  stop 'MEMTEST option works only with MPI'
#endif
#endif

#ifdef MPI
  call mpi_init(ierr)
  call mpi_comm_rank(mpi_comm_world, proc_id, ierr)
  call mpi_comm_size(mpi_comm_world, nproc, ierr)
#else
  call message ('MPI is not active')
  nproc   = 1
  proc_id = 0
#endif

  call message(' ******* Starting Pinocchio 2.2 ******')
  write(info,'(''['',a24,''] Starting run'')') fdate()
  call message (trim(info))
  write (info,'(''.. mpi started, # processors = '',i6)') nproc
  call message(info)

#ifdef MPI
! timing of the code
  cputime(0) = mpi_wtime()
#endif

! ******************************
! routines to initialize the run
! ******************************
  call initialization()


! ************************
! CYCLE ON SMOOTHING RADII
! ************************
#if defined(MEMTEST) || defined(IDLTEST)
  do ismooth = nsmooth,nsmooth
#else
  do ismooth = 1, nsmooth
#endif

     call message ("")
     write(info,'(''['',a24,''] Starting smoothing radius '',i5,'' of '',i5)') fdate(),ismooth,nsmooth
     call message (trim(info))
     write(info,'(''Smoothing radius: '',i3,2(1x,f15.8))') &
          ismooth,Smoothing_Radii(ismooth),sqrt(Smoothing_Variances(ismooth))
     call message (trim(info))

#ifdef MEMTEST
     call compute_density(ismooth,1)
#endif


! *********************************************************************
! PART 1: 
! Compute the second derivatives of the potential for all the grids
! *********************************************************************
#ifdef MPI
     cputime(2) = mpi_wtime()               ! timing
#endif

     do igrid = 1, ngrid
        if (ngrid>1) then
           write(info,'(''Grid '',i1)') igrid
           call message(trim(info))
        endif

#ifdef MPI
       cputime(1) = mpi_wtime()            ! timing
#endif

       write(info,'(''['',a24,''] Computing second derivatives'')') fdate()
       call message (trim(info))
       call compute_second_derivatives(ismooth, igrid)

#ifdef MPI
       cputime(1) = mpi_wtime()-cputime(1) ! timing
       write(info,'(''['',a24,''] Done second derivatives, cpu time = '',e12.4)') fdate(), cputime(1)
       call message (trim(info))
#endif

    enddo

#ifdef MPI
    cputime(2) = mpi_wtime() - cputime(2)  ! timing
    write(info,'(''['',a24,''] All second derivatives, cpu time = '',e12.4)') fdate(), cputime(2)
    call message (trim(info))
#endif

#ifdef MEMTEST
     call message(' finished memtest ')
     call finish (' SUCCESS')
#endif


! ***********************************************************************
! PART 2:
! Compute the collapse times by summing the grids over many positions
! ***********************************************************************
#ifndef IDLTEST

#ifdef MPI
     cputime(3) = mpi_wtime()
#endif

! allocates the matrices to store a single plane of the highest resolution box
     call assign_fftw_variables(ngrid)
     call allocate_plane(2)

! for a single grid calls the routine only once
     write(info,'(''['',a24,''] Computing collapse times'')') fdate()
     call message (trim(info))
     if (ngrid==1) then
        call compute_collapse_times(ismooth,0,0,0)
     else
! in this case a list of positions to place the small box has been specified
        if (list_of_positions) then
           do ipos = 1, npos
              call compute_collapse_times(ismooth, &
                   positions(ipos,1), positions(ipos,2), positions(ipos,3))
           enddo
        else
! in this case the box is placed on all the possible positions of the biggest box
           do iz = 0, Lgrid(1,1)-1, Lgrid_overlap(1,1)
              do iy = 0, Lgrid(1,2)-1, Lgrid_overlap(1,2)
                 do ix = 0, Lgrid(1,3)-1, Lgrid_overlap(1,3)
                    call compute_collapse_times(ismooth,ix,iy,iz)
                 enddo
              enddo
           enddo

        endif
     endif
#ifdef MPI
     cputime(3) = mpi_wtime() - cputime(3)      ! timing
     write(info,'(''['',a24,''] Done computing collapse times, cputime = '',e12.4)') fdate(),cputime(3)
     call message (trim(info))
#endif

! deallocates the matrices for the plane
     call deallocate_plane(2)

! removes the tmp files of second derivatives
     write(info,'(''['',a24,''] Remove second derivatives'')') fdate()
     call message (trim(info))
     call remove_second_derivatives()
     write(info,'(''['',a24,''] Done remove second derivatives'')') fdate()
     call message (trim(info))


#endif
! *****************************************************************
! PART 3: 
! computes the first derivatives of the potential for all the grids
! *****************************************************************
#ifdef MPI
     cputime(4) = mpi_wtime()                  ! timing
#endif

     write(info,'(''['',a24,''] Computing first derivatives'')') fdate()
     call message (trim(info))
     do igrid = 1, ngrid
        if (ngrid>1) then
           write(info,'(''Grid '',i1)') igrid
           call message(trim(info))
        endif
        call compute_first_derivatives(ismooth, igrid)
     enddo
#ifdef MPI
     cputime(4) = mpi_wtime() - cputime(4)     ! timing
     write(info,'(''['',a24,''] Done computing first derivatives, cptime = '',e12.4)') fdate(),cputime(4)
     call message (trim(info))
#endif


#ifdef IDLTEST
     call message(' finished idl_test ')
     call finish (' SUCCESS')
#endif

! *******************************************************************
! PART 4:
! it computes the velocities by summing the grids over many positions
! *******************************************************************
#ifdef MPI
     cputime(5) = mpi_wtime()                 ! timing
#endif

! allocates the matrices to store a single plane for the highest resolution box
     call assign_fftw_variables(ngrid)
     call allocate_plane(1)

! for a single grid calls the routine only once
     write(info,'(''['',a24,''] Computing velocities'')') fdate()
     call message (trim(info))
     if (ngrid==1) then
        call compute_velocities(ismooth,0,0,0)
     else
! in this case a list of positions to place the small box has been specified
        if (list_of_positions) then
           do ipos = 1, npos
              call compute_velocities(ismooth, &
                   positions(ipos,1), positions(ipos,2), positions(ipos,3))
           enddo
        else
! in this case the box is placed on all the possible positions of the biggest box
           do iz = 0, Lgrid(1,1)-1, Lgrid_overlap(1,1)
              do iy = 0, Lgrid(1,2)-1, Lgrid_overlap(1,2)
                 do ix = 0, Lgrid(1,3)-1, Lgrid_overlap(1,3)
                    call compute_velocities(ismooth,ix,iy,iz)
                 enddo
              enddo
           enddo

        endif
     endif
#ifdef MPI
     cputime(5) = mpi_wtime() - cputime(5)         ! timing
     write(info,'(''['',a24,''] Done computing velocities, cputime = '',e12.4)') fdate(),cputime(5)
     call message (trim(info))
#endif

! deallocates the matrices for the plane
     call deallocate_plane(1)

! removes the tmp files of first derivatives
     write(info,'(''['',a24,''] Remove first derivatives'')') fdate()
     call message (trim(info))
     call remove_first_derivatives()
     write(info,'(''['',a24,''] Done remove first derivatives'')') fdate()
     call message (trim(info))


! *******************************
! END OF CYCLE ON SMOOTHING RADII
! *******************************
     call message ('Smoothing radius completed')
#ifdef MPI
     call mpi_barrier(mpi_comm_world,ierr)
#endif
  enddo


! ***********************************************************************************
! FINAL PART:
! Compute the density and output density and output the Fmax and Vel* fields
! ***********************************************************************************
#ifdef MPI
  cputime(6) = mpi_wtime()                    ! timing
#endif

  do igrid = 1, ngrid
     if (.not.read_IC(igrid)) then
        call compute_density(nsmooth,igrid)
     endif
  enddo

#ifdef MPI
  cputime(6) = mpi_wtime() - cputime(6)       ! timing
  write(info,'('' cputime compute_density =  '',e12.4)') cputime(6)
  call message(trim(info))
#endif

! Writes the results in files in the Data/ directory
!  call allocate_plane(0)
  call write_out()
!  call deallocate_plane(0)

#ifdef MPI
! final cpu time
  cputime(0) = mpi_wtime() - cputime(0)
  write(info,'(''['',a24,''] Finishing, total cputime = '',e12.4)') fdate(),cputime(0)
  call message (trim(info))
#endif

! finalize MPI
  call finish('Fmax done!')

end program fmax




function fdate ()
  
  character(len=24) :: fdate
  character(len=8)  :: date
  character(len=10) :: time
  character(len=5)  :: timezone
  integer                   :: values(8)
  
  call date_and_time (date, time, timezone, values)
  
  write (fdate, &
       &   "(I2.2,'-',I2.2,'-',I4.4,' ',I2.2,':',I2.2,'.',I2.2,'     ')") &
       &   values(2), values(3), values(1), values(5), values(6), &
       &   values(7)
  
  return
end function fdate
                                                                                                                                     
 
