!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


! This file contains the routines that allocate matrices
!
! ALLOCATION STRATEGY FOR THE BIG MATRICES
!
! The matrices that consume most memory are the following:
! (NB: nplane = nx_fftw x ny_fftw)
! 
!  name                type         size         allocated by            module
! ---------------------------------------------------------------------------------
! vector              doubleR  total_local_size  allocate_vector     fftw_variables
! fftw_work           doubleR  total_local_size  allocate_vector     fftw_variables
! plane               singleR*     nplane        allocate_tmp_plane    tmp_plane
! differentials_plane doubleR  3 or 6 x nplane   allocate_plane      fmax_one_plane
! Fmax_plane          doubleR      nplane        allocate_plane      fmax_one_plane
! Rmax_plane          singleI      nplane        allocate_plane      fmax_one_plane
! vmax_plane          doubleR     3 x nplane     allocate_plane      fmax_one_plane
!
! * note: it is doubleR with the DPTMP flag
!
! These arrays are allocated as follows in the code:
!
! PART 1:  for each grid - vector, fftw_work, plane
! PART 2:  for the last grid - differentials_plane(1:6), Fmax_plane, Rmax_plane, plane
! PART 3:  for each grid - vector, fftw_work, plane
! PART 4:  for the last rgid - differentials_plane(1:3), vmax_plane, Rmax_plane, plane
! density: for each non-generated grid - vector, fftw_work, plane
! write_out: for the last grid - plane
!
! The arrays containing the sections of the bigger box are neglected here,
! but their could be rather large in specific cases.
!
!
! SUBROUTINES AND FUNCTIONS IN THIS FILE
!
! - subroutine allocate_vector()
!     allocates the vectors necessary to perform differentiations
!     calls: allocate_tmp_plane()                     fmax-alloc
! - subroutine deallocate_vector()
!     deallocates the vectors
!     calls: deallocate_tmp_plane()                   fmax-alloc
! - subroutine allocate_tmp_plane()
!     allocates the tmp plane for reading and writing files
! - subroutine deallocate_tmp_plane()
!     deallocates the tmp plane
! - subroutine allocate_tmp_integer_plane()
!     allocates the tmp int_plane for reading and writing files
! - subroutine deallocate_tmp_integer_plane()
!     deallocates the tmp int_plane
! - subroutine allocate_plane(ider)
!     allocates all the matrices necessary to handle a plane in real space
!     (differentials, Fmax, Rmax, vmax, boxes)
!     calls: allocate_differentials_plane             fmax-alloc
!            allocate_Fmax_plane                      fmax-alloc
!            allocate_Rmax_plane                      fmax-alloc
!            allocate_vmax_plane                      fmax-alloc
!            allocate_box1                            fmax-alloc
!            allocate_box2                            fmax-alloc
!            allocate_box3                            fmax-alloc
!            allocate_box4                            fmax-alloc
! - subroutine allocate_differentials_plane(ncomp,Lx,Ly,total)
! - subroutine allocate_Fmax_plane(Lx,Ly,total)
! - subroutine allocate_Rmax_plane(Lx,Ly,total)
! - subroutine allocate_vmax_plane(ncomp,Lx,Ly,total)
! - subroutine allocate_box1(ncomp,Lx,Ly,Lz,total)
! - subroutine allocate_box2(ncomp,Lx,Ly,Lz,total)
! - subroutine allocate_box3(ncomp,Lx,Ly,Lz,total)
! - subroutine allocate_box4(ncomp,Lx,Ly,Lz,total)
! - subroutine deallocate_plane(ider)
!     deallocates what had been allocated above
! - subroutine allocate_grid_vectors(saved_seed)
!     allocates the vectors needed to handle the grids
! - subroutine allocate_positions()
!     allocate the vectors needed to store the positions of the highest res. box
! - subroutine allocate_smoothing_radii()
!     allocate the vectors for the smoothing radii and variances


subroutine allocate_vector()

! This subroutine allocates the vector necessary for Part 2 and 4 of the main code
! i.e. the computation of derivatives with fftw.
! It allocates:
! - vector(total_local_size)     [in k-space]
! - fftw_work(total_local_size)  [workspace for fftw]
! - plane(nx_fftw*ny_fftw)       [z-plane in real-space]
! - plane_id(ny_fftw)            [ids for y-planes in k-space]

  use fftw_variables
  use pinocchio_mpi
  use alloc_flag
  implicit none

! local variables
  character(len=80)  :: info
  real(kind=doubleR) :: total_allocated
  real(kind=doubleR), parameter :: Gbyte=1024.d0*1024.d0*1024.d0

! this check is hopefully useless; the allocate_vector should never be called
! if the vector is already allocated
  if (allocated(vector)) &
     call abort_code ('error in allocate_vector: vector already allocated')

! allocation of the fftw vector
  total_allocated = total_local_size
  if (alloc_verbose) call message ('allocating vector...')
  allocate(vector(total_local_size),stat=ierr)
  if (alloc_verbose) call message ('vector allocated')
  if (ierr>0) then
     call abort_code('error in ALLOCATE_VECTOR: unable to allocate vector')
  endif

! allocation of the fftw workspace
  total_allocated = total_allocated + total_local_size
  if (allocated(fftw_work)) &
       call abort_code ('error in allocate_vector: fftw_work array already allocated')
  if (alloc_verbose) call message ('allocating fftw_work...')
  allocate(fftw_work(total_local_size),stat=ierr)
  if (alloc_verbose) call message ('fftw_work allocated')
  if (ierr>0) then
     call abort_code('error in ALLOCATE_VECTOR: unable to allocate workspace')
  endif

! allocation of tmp plane
  total_allocated = total_allocated + nx_fftw*ny_fftw
  call allocate_tmp_plane(0)

! allocate ids for random number streams
  if (allocated(plane_id)) &
       call abort_code ('error in allocate_vector: plane_id array already allocated')
  if (alloc_verbose) call message ('allocating plane_id...')
  allocate(plane_id(ny_fftw),stat=ierr)
  if (alloc_verbose) call message ('plane_id allocated')
  if (ierr>0) &
     call abort_code('error in allocate_vector: cannot allocate plane_id')

! message
  if(alloc_verbose) then
     write (info,'('' allocated arrays of size: '',f8.3,'' Gbytes'')') total_allocated*8/GByte
     call message(trim(info))
  endif

  return
end subroutine allocate_vector


subroutine deallocate_vector()

! De-allocates the vectors allocated by allocate_vector

  use fftw_variables
  use pinocchio_mpi
  use alloc_flag
  implicit none

! local variables
  character(len=80) :: info

  deallocate(vector)
  deallocate(fftw_work)
  deallocate(plane_id)
  call deallocate_tmp_plane()

  if(alloc_verbose) then
     write (info,'('' de-allocated fftw vector, work space and plane'')')
     call message(trim(info))
  end if

  return
end subroutine deallocate_vector


subroutine allocate_tmp_plane(igrid)

! allocate the tmp plane for reading and writing files (both tmp and i/o)
! NB: plane is used to read and write (plane by plane) both tmp files and 
!     output files, with the exceptions of tmp Rmax files (being Rmax an integer)
!     and the double precision fftw vector in read_ and save_kspace_vector;
!     plane is usually singleR, unless the flag DPTMP is specified

  use numbers
  use fftw_variables, ONLY: nx_fftw, ny_fftw, nz_fftw
  use fmax_grids
  use tmp_plane
  use alloc_flag
  implicit none
  integer(kind=singleI),intent(in)::igrid
  integer(kind=singleI)::ierr

  if(allocated(plane)) &
       call abort_code ('error in allocate_tmp_plane: plane already allocated')
  if (alloc_verbose) call message('allocating plane...')
  if (igrid==0) then
     allocate(plane(nx_fftw,ny_fftw),stat=ierr)
  else
     allocate(plane(Lgrid(igrid,1),Lgrid(igrid,2)), stat=ierr)
  endif
  if (alloc_verbose) call message('plane allocated')
  if (ierr>0) then
     call abort_code('error in allocate_tmp_plane: unable to allocate plane')
  endif

  return
end subroutine allocate_tmp_plane


subroutine deallocate_tmp_plane()

! deallocate the tmp plane

  use tmp_plane
  implicit none

  deallocate (plane)

  return
end subroutine deallocate_tmp_plane


subroutine allocate_tmp_integer_plane(igrid)

! allocate the tmp integer plane for reading and writing files (both tmp and i/o)
! NB: this plane is used to read and write (plane by plane) the Rmax field

  use numbers
  use fftw_variables, ONLY: nx_fftw, ny_fftw, nz_fftw
  use fmax_grids
  use tmp_plane
  use alloc_flag
  implicit none
  integer(kind=singleI),intent(in)::igrid
  integer(kind=singleI)::ierr

  if(allocated(int_plane)) &
       call abort_code ('error in allocate_tmp_integer_plane: int_plane already allocated')
  if (alloc_verbose) call message('allocating int_plane...')
  if (igrid==0) then
     allocate(int_plane(nx_fftw,ny_fftw),stat=ierr)
  else
     allocate(int_plane(Lgrid(igrid,1),Lgrid(igrid,2)), stat=ierr)
  endif
  if (alloc_verbose) call message('int_plane allocated')
  if (ierr>0) then
     call abort_code('error in allocate_tmp_integer_plane: unable to allocate int_plane')
  endif

  return
end subroutine allocate_tmp_integer_plane


subroutine deallocate_tmp_integer_plane()

! deallocate the tmp int_plane

  use tmp_plane
  implicit none

  deallocate (int_plane)

  return
end subroutine deallocate_tmp_integer_plane


subroutine allocate_plane(ider)

! Allocates the matrices necessary to handle a plane; used by Part 2 and 4
! of the main code

  use numbers
  use fmax_grids
  use alloc_flag
!  use fftw_variables, ONLY: fftw_verbose
  implicit none
  integer(kind=singleI),intent(in)::ider

! local variables
  character(len=80)  :: info
  real(kind=doubleR) :: total_allocated
  real(kind=doubleR), parameter :: Gbyte=1024.d0*1024.d0*1024.d0

  total_allocated=0.d0

  if (ider == 0) then
! in this case it allocates the matrices needed to read in Fmax and vmax
! and write them out to files


     call allocate_vmax_plane(3,Lgrid_local(ngrid,1),Lgrid_local(ngrid,2),total_allocated)
     call allocate_Fmax_plane(Lgrid_local(ngrid,1),Lgrid_local(ngrid,2),total_allocated)

  else if (ider == 1) then

! in this case it allocates the matrices necessary to compute velocities

     call allocate_differentials_plane(3,Lgrid_local(ngrid,1),Lgrid_local(ngrid,2),total_allocated)
     call allocate_vmax_plane(3,Lgrid_local(ngrid,1),Lgrid_local(ngrid,2),total_allocated)
     call allocate_Rmax_plane(Lgrid_local(ngrid,1),Lgrid_local(ngrid,2),total_allocated)
     if (ngrid>1) &
          call allocate_box1(3,Lgrid_overlap(ngrid-1,1),Lgrid_overlap(ngrid-1,2) &
          ,Lgrid_overlap(ngrid-1,3),total_allocated)
     if (ngrid>2) &
          call allocate_box2(3,Lgrid_overlap(ngrid-2,1),Lgrid_overlap(ngrid-2,2) &
          ,Lgrid_overlap(ngrid-2,3),total_allocated)
     if (ngrid>3) &
          call allocate_box3(3,Lgrid_overlap(ngrid-2,1),Lgrid_overlap(ngrid-2,2) &
          ,Lgrid_overlap(ngrid-2,3),total_allocated)
     if (ngrid>4) &
          call allocate_box4(3,Lgrid_overlap(ngrid-4,1),Lgrid_overlap(ngrid-4,2) &
          ,Lgrid_overlap(ngrid-4,3),total_allocated)

  else if (ider == 2) then

! in this case it allocates the matrices necessary to compute collapse times

     call allocate_differentials_plane(6,Lgrid_local(ngrid,1),Lgrid_local(ngrid,2),total_allocated)
     call allocate_Fmax_plane(Lgrid_local(ngrid,1),Lgrid_local(ngrid,2),total_allocated)
     call allocate_Rmax_plane(Lgrid_local(ngrid,1),Lgrid_local(ngrid,2),total_allocated)
     if (ngrid>1) &
          call allocate_box1(6,Lgrid_overlap(ngrid-1,1),Lgrid_overlap(ngrid-1,2) &
          ,Lgrid_overlap(ngrid-1,3),total_allocated)
     if (ngrid>2) &
          call allocate_box2(6,Lgrid_overlap(ngrid-2,1),Lgrid_overlap(ngrid-2,2) &
          ,Lgrid_overlap(ngrid-2,3),total_allocated)
     if (ngrid>3) &
          call allocate_box3(6,Lgrid_overlap(ngrid-2,1),Lgrid_overlap(ngrid-2,2) &
          ,Lgrid_overlap(ngrid-2,3),total_allocated)
     if (ngrid>4) &
          call allocate_box4(6,Lgrid_overlap(ngrid-4,1),Lgrid_overlap(ngrid-4,2) &
          ,Lgrid_overlap(ngrid-4,3),total_allocated)

  else
     call abort_code('wrong call of allocate_plane')
  endif

! message
  if(alloc_verbose) then
     write (info,'('' allocated arrays of size: '',f8.3,'' Gbytes'')') total_allocated*8./GByte
     call message(trim(info))
  endif

  return
end subroutine allocate_plane


subroutine allocate_differentials_plane(ncomp,Lx,Ly,total_allocated)
  use numbers
  use fmax_one_plane
  use alloc_flag
  implicit none
  integer(kind=singleI),intent(in) :: ncomp,Lx,Ly
  real(kind=doubleR),intent(inout) :: total_allocated
  integer(kind=singleI)::ierr

  if (allocated(differentials_plane)) then
     call abort_code('error: differentials_plane already allocated')
  endif
  total_allocated = total_allocated + ncomp*Lx*Ly
  if (alloc_verbose) call message('allocating differentials_plane...')
  allocate (differentials_plane(ncomp,0:Lx-1,0:Ly-1),stat=ierr)
  if (alloc_verbose) call message('differentials_plane allocated')
  if (ierr>0) then
     call abort_code('unable to allocate differentials_plane')
  endif

  return
end subroutine allocate_differentials_plane


subroutine allocate_Fmax_plane(Lx,Ly,total_allocated)
  use numbers
  use fmax_one_plane
  use alloc_flag
  implicit none
  integer(kind=singleI),intent(in) :: Lx,Ly
  real(kind=doubleR),intent(inout) :: total_allocated
  integer(kind=singleI)::ierr

  if (allocated(Fmax_plane)) then
     call abort_code('error: Fmax_plane already allocated')
  endif
  total_allocated = total_allocated + Lx*Ly
  if (alloc_verbose) call message('allocating Fmax_plane...')
  allocate (Fmax_plane(0:Lx-1,0:Ly-1),stat=ierr)
  if (alloc_verbose) call message('Fmax_plane allocated')
  if (ierr>0) then
     call abort_code('unable to allocate Fmax_plane')
  endif

  return
end subroutine allocate_Fmax_plane


subroutine allocate_Rmax_plane(Lx,Ly,total_allocated)
  use numbers
  use fmax_one_plane
  use alloc_flag
  implicit none
  integer(kind=singleI),intent(in) :: Lx,Ly
  real(kind=doubleR),intent(inout) :: total_allocated
  integer(kind=singleI)::ierr

  if (allocated(Rmax_plane)) then
     call abort_code('error: Rmax_plane already allocated')
  endif
  total_allocated = total_allocated + 0.5*Lx*Ly
  if (alloc_verbose) call message('allocating Rmax_plane...')
  allocate (Rmax_plane(0:Lx-1,0:Ly-1),stat=ierr)
  if (alloc_verbose) call message('Rmax_plane allocated')
  if (ierr>0) then
     call abort_code('unable to allocate Rmax_plane')
  endif

  return
end subroutine allocate_Rmax_plane


subroutine allocate_vmax_plane(ncomp,Lx,Ly,total_allocated)
  use numbers
  use fmax_one_plane
  use alloc_flag
  implicit none
  integer(kind=singleI),intent(in) :: ncomp,Lx,Ly
  real(kind=doubleR),intent(inout) :: total_allocated
  integer(kind=singleI)::ierr

  if (allocated(vmax_plane)) then
     call abort_code('error: vmax_plane already allocated')
  endif
  total_allocated = total_allocated + ncomp*Lx*Ly
  if (alloc_verbose) call message('allocating vmax_plane...')
  allocate (vmax_plane(ncomp,0:Lx-1,0:Ly-1),stat=ierr)
  if (alloc_verbose) call message('vmax_plane allocated')
  if (ierr>0) then
     call abort_code('unable to allocate vmax_plane')
  endif

  return
end subroutine allocate_vmax_plane


subroutine allocate_box1(ncomp,Lx,Ly,Lz,total_allocated)
  use numbers
  use fmax_box_sections
  use alloc_flag
  implicit none
  integer(kind=singleI),intent(in) :: ncomp,Lx,Ly,Lz
  real(kind=doubleR),intent(inout) :: total_allocated
  integer(kind=singleI)::ierr

  if (allocated(box1)) then
     call abort_code('error: box1 already allocated')
  endif
  total_allocated=total_allocated + ncomp*Lx*Ly*Lz
  if (alloc_verbose) call message('allocating box1...')
  allocate(box1(ncomp,0:Lx-1,0:Ly-1,0:Lz-1),stat=ierr)
  if (alloc_verbose) call message('box1 allocated')
  if (ierr>0) then
     call abort_code('unable to allocate vector box1')
  endif

  return
end subroutine allocate_box1


subroutine allocate_box2(ncomp,Lx,Ly,Lz,total_allocated)
  use numbers
  use fmax_box_sections
  use alloc_flag
  implicit none
  integer(kind=singleI),intent(in) :: ncomp,Lx,Ly,Lz
  real(kind=doubleR),intent(inout) :: total_allocated
  integer(kind=singleI)::ierr

  if (allocated(box2)) then
     call abort_code('error: box2 already allocated')
  endif
  total_allocated=total_allocated + ncomp*Lx*Ly*Lz
  if (alloc_verbose) call message('allocating box2...')
  allocate(box2(ncomp,0:Lx-1,0:Ly-1,0:Lz-1),stat=ierr)
  if (alloc_verbose) call message('box2 allocated')
  if (ierr>0) then
     call abort_code('unable to allocate vector box2')
  endif

  return
end subroutine allocate_box2


subroutine allocate_box3(ncomp,Lx,Ly,Lz,total_allocated)
  use numbers
  use fmax_box_sections
  use alloc_flag
  implicit none
  integer(kind=singleI),intent(in) :: ncomp,Lx,Ly,Lz
  real(kind=doubleR),intent(inout) :: total_allocated
  integer(kind=singleI)::ierr

  if (allocated(box3)) then
     call abort_code('error: box3 already allocated')
  endif
  total_allocated=total_allocated + ncomp*Lx*Ly*Lz
  if (alloc_verbose) call message('allocating box3...')
  allocate(box3(ncomp,0:Lx-1,0:Ly-1,0:Lz-1),stat=ierr)
  if (alloc_verbose) call message('box3 allocated')
  if (ierr>0) then
     call abort_code('unable to allocate vector box3')
  endif

  return
end subroutine allocate_box3


subroutine allocate_box4(ncomp,Lx,Ly,Lz,total_allocated)
  use numbers
  use fmax_box_sections
  use alloc_flag
  implicit none
  integer(kind=singleI),intent(in) :: ncomp,Lx,Ly,Lz
  real(kind=doubleR),intent(inout) :: total_allocated
  integer(kind=singleI)::ierr

  if (allocated(box4)) then
     call abort_code('error: box4 already allocated')
  endif
  total_allocated=total_allocated + ncomp*Lx*Ly*Lz
  if (alloc_verbose) call message('allocating box4...')
  allocate(box4(ncomp,0:Lx-1,0:Ly-1,0:Lz-1),stat=ierr)
  if (alloc_verbose) call message('box4 allocated')
  if (ierr>0) then
     call abort_code('unable to allocate vector box4')
  endif

  return
end subroutine allocate_box4


subroutine deallocate_plane(ider)

! deallocates the matrices allocated by allocate_plane

  use fmax_grids
  use fmax_one_plane
  use fmax_box_sections
  implicit none
  integer(kind=singleI),intent(in)::ider

  deallocate (differentials_plane)
  deallocate (Rmax_plane)
  if (ider==1) then
     deallocate (vmax_plane)
  else
     deallocate (Fmax_plane)
  endif
  if (ngrid>1) deallocate (box1)
  if (ngrid>2) deallocate (box2)
  if (ngrid>3) deallocate (box3)
  if (ngrid>4) deallocate (box4)

  return
end subroutine deallocate_plane


subroutine allocate_grid_vectors()

! Allocate and initialize vectors that store information on the grids

  use numbers
  use fmax_grids
  use fmax_box_sections
  use alloc_flag
  use fmax_run
  implicit none

  character(len=1)::label

  if (alloc_verbose) call message ('allocating grid vectors...')
! basics
  allocate (Lgrid(ngrid,3))             ! grid sizes
  allocate (BoxSizeGrids(ngrid,3))      ! box sizes
  allocate (InterPartDistGrids(ngrid))  ! inter-particle distances
  allocate (k_cutoff(ngrid))            ! cutoffs in k-space
  allocate (flag_cutoff(ngrid))         ! flag for the cutoff (spherical or cubic)
  allocate (npart(ngrid))               ! number of particles in the grid
  allocate (nplane(ngrid))              ! number of particles in a z-plane
  allocate (coarse_level(ngrid,3))      ! number of particles overlapping with the smallest box (real-space)
  allocate (Lgrid_overlap(ngrid,3))     ! as above, but integer (>=1)

! local grids in real space
  allocate (Lgrid_local(ngrid,3))       ! dimension of local grid
  allocate (npart_local(ngrid))         ! number of particles in the local grid
  allocate (Lgrid_z_start(ngrid))       ! starting z-position of local grid

! local grids in the Fourier space
  allocate (Lgrid_fft(ngrid,3))         ! dimension of local grid
  allocate (npart_fft(ngrid))           ! number of particles in the local grid
  allocate (Lgrid_y_start(ngrid))       ! starting y-position of local grid

! fftw plans
  allocate (saved_forward_plan(ngrid))
  allocate (saved_reverse_plan(ngrid))

! ICs
  allocate (read_IC(ngrid))             ! flag for reading ICs
  allocate (InputDensityFile(ngrid))    ! IC filename (if any)

! for box interpolations
  allocate(deltax(ngrid))
  allocate(deltay(ngrid))
  allocate(deltaz(ngrid))
  if (alloc_verbose) call message ('grid vectors allocated')

! Initializations
  k_cutoff(ngrid)=1.d30
  if (saved_seed == 0) then
     read_IC(ngrid)=.true.
     write(label,'(i1)')ngrid
     InputDensityFile(ngrid)='Data/Density_g'//label//'.d'
  else
     read_IC(ngrid)=.false.
     InputDensityFile(ngrid)=''
  endif

  return
end subroutine allocate_grid_vectors


subroutine allocate_positions()

! allocates the array of positions of the small box

  use numbers
  use fmax_grids
  use alloc_flag
  implicit none

  if (alloc_verbose) call message ('allocating positions...')
  allocate(positions(npos,3))
  if (alloc_verbose) call message ('positions allocated')

  return
end subroutine allocate_positions


subroutine allocate_smoothing_radii()

! Allocates vectors for smoothing radii

  use numbers
  use fmax_smoothing_radii
  use alloc_flag
  implicit none

  if (alloc_verbose) call message ('allocating smoothing radii...')
  allocate(Smoothing_Radii(nsmooth))
  allocate(Smoothing_Variances(nsmooth))
  allocate(True_Variances(nsmooth)) ; True_Variances = 0.d0
  if (alloc_verbose) call message ('smoothing radii allocated')

  return
end subroutine allocate_smoothing_radii
