!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


! This file contains the routines that handle multiple nested boxes;
! these are called mainly in Part 2 and 4 of the main code.
!
!
! SUBROUTINES AND FUNCTIONS IN THIS FILE
!
! - subroutine get_one_box(igrid,xpos,dx,ypos,dy,zpos,dz,derivative)
!     gets one derivative from a large-scale box
!     calls: get_component_plane(...)                 fmax-tmpfiles
!            allocate_tmp_plane()                     fmax-alloc
!            deallocate_tmp_plane()                   fmax-alloc
! - subroutine get_boxes(ix_lb,iy_lb,iz_lb,derivative)
!     drives the acquisition of derivatives from large-scale boxes
!     calls: get_one_box(...)                           fmax-boxes
! - subroutine sum_over_boxes(local_x,local_y,local_z,derivative)
!     sums the deformation tensor or displacement vector over the
!     large-scale  boxes
!     calls: interpolate3D(...)                       fmax-boxes
! - real(kind=doubleR) function interpolate3D(x,y,z,array)
!     performs a trilinear interpolation in a cubic array


subroutine get_one_box(igrid,xpos,dx,ypos,dy,zpos,dz,derivative)

! gets one derivative from a large-scale section of the grid
! igrid: grid ID
! xpos, ypos, zpos: posizion of starting particle of the section
! dx, dy, dz: width of the section to recover
! derivative: first or second derivative tensor

! SALVARE I PIANI!!!!!!!!!


  use fmax_grids
  use fmax_box_sections
  use tmp_plane
  implicit none
  integer(kind=singleI),intent(in)::igrid,xpos,dx,ypos,dy,zpos,dz,derivative

! local variables
  integer(kind=singleI)::components,ix,iy,iz,xplane,yplane,zplane,icomp

  print *,'chiamata get_one_box ',igrid,xpos,dx,ypos,dy,zpos,dz,derivative

! allocates tmp plane for reading components
  call allocate_tmp_plane(igrid)

! Number of components according to the derivative
  select case (derivative)
  case (1)
     components=3
  case (2)
     components=6
  end select

  do iz= 0, dz-1
 
! checks that the plane is required by the processor, if not sets the boxes to 0.
     if ((iz+1)*Lgrid_overlap(igrid,3) < Lgrid_z_start(ngrid) .or. &
         iz*Lgrid_overlap(igrid,3) > Lgrid_z_start(ngrid)+Lgrid_local(ngrid,3)) then
        
        do icomp = 1, components
           if (igrid==1) box1(icomp,:,:,iz) = 0.
           if (igrid==2) box2(icomp,:,:,iz) = 0.
           if (igrid==3) box3(icomp,:,:,iz) = 0.
           if (igrid==4) box4(icomp,:,:,iz) = 0.
        enddo

     else

! z-component with PBCs
        zplane = zpos + iz
        if (zplane > Lgrid(igrid,3)) zplane = mod(zplane,Lgrid(igrid,3))

        do icomp = 1, components

           print *,'chiamo get_component_plane ',igrid, zpos+iz, derivative, icomp

! gets a large-scale plane
           call get_component_plane(igrid, zpos+iz, derivative, icomp)

! y-component with PBCs
           do iy = 0, dy-1
              yplane = ypos + iy
              if (yplane > Lgrid(igrid,2)) yplane = mod(yplane,Lgrid(igrid,2))

! x-component with PBCs
              do ix = 0, dx-1
                 xplane = xpos + ix
                 if (xplane > Lgrid(igrid,1)) xplane = mod(xplane,Lgrid(igrid,1))

! saves the required section in the box* array
                 if (igrid==1) box1(icomp,ix,iy,iz) = plane(xplane,yplane)
                 if (igrid==2) box2(icomp,ix,iy,iz) = plane(xplane,yplane)
                 if (igrid==3) box3(icomp,ix,iy,iz) = plane(xplane,yplane)
                 if (igrid==4) box4(icomp,ix,iy,iz) = plane(xplane,yplane)

              enddo
           enddo
        enddo
     endif
  enddo

  call deallocate_tmp_plane()

end subroutine get_one_box


subroutine get_boxes(ix_lb,iy_lb,iz_lb,derivative)

! drives the acquisition of derivatives from large-scale boxes
! ix_lb, iy_lb and iz_lb: position within the largest box of the origin of the 
!    highest resolution box. 
!    NB: IN UNITS OF THE LARGEST BOX GRID, from 0 to Lgrid(1,*)-1
! derivative: first or second derivative tensors
!
! NB: we assume that the intermediate boxes fill the the volume of the largest
!     box with PBCs

  use fmax_grids
  use fmax_box_sections
  implicit none
  integer(kind=singleI),intent(in)::ix_lb,iy_lb,iz_lb,derivative

! local variables
  integer(kind=singleI)::xpos,ypos,zpos,dx,dy,dz,Lgrid_scale(3),igrid

! nothing to do in the case of a single box
  if (ngrid==1) return

  print *,'chiamata get_boxes ',ix_lb,iy_lb,iz_lb,derivative

  do igrid=1,ngrid-1
     Lgrid_scale = Lgrid(igrid,:)/Lgrid(1,:)

! positions and widths to recover
! NB the width of the box is increased by one grid point to allow interpolations
! NB PBCs are applied in get_one_box
     xpos = ix_lb*Lgrid_scale(1) ; dx=Lgrid_overlap(igrid,1)+1
     ypos = iy_lb*Lgrid_scale(2) ; dy=Lgrid_overlap(igrid,2)+1
     zpos = iz_lb*Lgrid_scale(3) ; dz=Lgrid_overlap(igrid,3)+1

!!!! other choice: only the section overlapping with the slice is loaded
!!!     zpos = (iz_lb + int(dble(Lgrid_z_start(ngrid))/dble(Lgrid(ngrid,3))&
!!!       *dble(Lgrid_overlap(igrid,3)))) * Lgrid_scale(3)
!!!     dz   = int(dble(Lgrid_local(ngrid,3))/dble(Lgrid(ngrid,3)) &
!!!          *dble(Lgrid_overlap(igrid,3))) + 1

! records the widths of the box
     deltax(igrid) = dx ; deltay(igrid) = dy ; deltaz(igrid) = dz

     print *,'chiamo get_one_box ',igrid,xpos,dx,ypos,dy,zpos,dz,derivative

! reads the box from the tmp files
     call get_one_box(igrid,xpos,dx,ypos,dy,zpos,dz,derivative)
  enddo

!!!! other boxes
!!!  do igrid=2,ngrid-1
!!!     Lgrid_scale = Lgrid(igrid,:)/Lgrid(1,:)
!!!
!!!     xpos = ix_lb*Lgrid_scale(1) 
!!!     if (xpos>=Lgrid(igrid,1)) xpos=mod(xpos,Lgrid(igrid,1))
!!!     dx   = (deltax(1)-1)*Lgrid_scale(1) + 1
!!!
!!!     ypos = iy_lb*Lgrid_scale(2) 
!!!     if (ypos>=Lgrid(igrid,2)) ypos=mod(ypos,Lgrid(igrid,2))
!!!     dy   = (deltay(1)-1)*Lgrid_scale(2) + 1
!!!
!!!     zpos = (iz_lb + int(dble(Lgrid_z_start(ngrid))/dble(Lgrid(ngrid,3))&
!!!       *dble(Lgrid_overlap(1,3)))) * Lgrid_scale(3)
!!!     if (zpos>=Lgrid(igrid,3)) zpos=mod(zpos,Lgrid(igrid,3))
!!!     dz   = (deltaz(1)-1)*Lgrid_scale(3) + 1
!!!
!!!     deltax(igrid) = dx ; deltay(igrid) = dy ; deltaz(igrid) = dz
!!!
!!!     print *,'chiamo get_one_box ',igrid,xpos,dx,ypos,dy,zpos,dz,derivative
!!!
!!!     call get_one_box(igrid,xpos,dx,ypos,dy,zpos,dz,derivative)
!!!  enddo

  return
end subroutine get_boxes


subroutine sum_over_boxes(global_x,global_y,global_z,derivative)

! sum a derivative tensor over all the boxes
! global_x, global_y, global_z is the position within the highest-resolution grid

  use fmax_box_sections
  use fmax_grids
  implicit none
  integer(kind=singleI),intent(in)::global_x,global_y,global_z,derivative

! local variables
  real(kind=doubleR)::tointerpolate(2,2,2),x,y,z
  integer(kind=singleI)::ncomp,igrid,icomp,xpos_in_box,ypos_in_box,zpos_in_box
  real(kind=doubleR),external::interpolate3D

  if (ngrid==1) return

  print *,'chiamata sum_over_boxes ',global_x,global_y,global_z,derivative

  if (derivative==1) then
     ncomp=3
  else if (derivative==2) then
     ncomp=6
  else
     call abort_code('Wrong call of sum_over_boxes')
  endif

  do igrid=1,ngrid-1


     x = dble(global_x) / dble(Lgrid_overlap(igrid,1))
     y = dble(global_y) / dble(Lgrid_overlap(igrid,2))
     z = dble(global_z) / dble(Lgrid_overlap(igrid,3))

     xpos_in_box = int(x) ; x = x - xpos_in_box
     ypos_in_box = int(y) ; y = y - ypos_in_box 
     zpos_in_box = int(z) ; z = z - zpos_in_box 

     do icomp=1,ncomp
        if (igrid==1) tointerpolate = box1(icomp &
             ,xpos_in_box:xpos_in_box+1 &
             ,ypos_in_box:ypos_in_box+1 &
             ,zpos_in_box:zpos_in_box+1 )
        if (igrid==2) tointerpolate = box2(icomp &
             ,xpos_in_box:xpos_in_box+1 &
             ,ypos_in_box:ypos_in_box+1 &
             ,zpos_in_box:zpos_in_box+1 )
        if (igrid==3) tointerpolate = box3(icomp &
             ,xpos_in_box:xpos_in_box+1 &
             ,ypos_in_box:ypos_in_box+1 &
             ,zpos_in_box:zpos_in_box+1 )
        if (igrid==4) tointerpolate = box4(icomp &
             ,xpos_in_box:xpos_in_box+1 &
             ,ypos_in_box:ypos_in_box+1 &
             ,zpos_in_box:zpos_in_box+1 )

        diff_ten(icomp) = diff_ten(icomp) &
             + interpolate3D(x,y,z,tointerpolate) !/InterPartDistGrids(igrid)

        print *,'update: ',igrid,icomp,diff_ten(icomp)

     enddo

  enddo

  return
end subroutine sum_over_boxes


real(kind=doubleR) function interpolate3D(x,y,z,array)

! tri-linear interpolation of a 2^3 array
! x, y and z are assumed in the range [0,1)

  use numbers
  implicit none
  real(kind=doubleR),intent(in)::x,y,z,array(2,2,2)

  interpolate3D = array(1,1,1) * (1.d0-x) * (1.d0-y) * (1.d0-z) &
                + array(2,1,1) *       x  * (1.d0-y) * (1.d0-z) &
                + array(1,2,1) * (1.d0-x) *       y  * (1.d0-z) &
                + array(2,2,1) *       x  *       y  * (1.d0-z) &
                + array(1,1,2) * (1.d0-x) * (1.d0-y) *       z  &
                + array(2,1,2) *       x  * (1.d0-y) *       z  &
                + array(1,2,2) * (1.d0-x) *       y  *       z  &
                + array(2,2,2) *       x  *       y  *       z  

  return
end function interpolate3D
