!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

! This file contains the wrappers for fftw
!
!
! SUBROUTINES AND FUNCTIONS IN THIS FILE
! - subroutine real_derivative(first_derivative, second_derivative)
!     reads or generates a potential, differentiates it in the k-space
!     and calls the wrapper for fftw
!     calls: reverse_Transform()                      fmax-fftw
!            index_kspace(i,j,k)                      fmax-fftw
!            greens_function(...)                     fmax-fftw
!            init_random(streamnum)                   random
!            power(k)                                 cosmo
!            random(stream)                           random
! - integer(kind=doubleI) function index_rspace(local_x, local_y, local_z)
!     index of fftw vector in real-space
! - integer(kind=doubleI) function index_kspace(local_x, local_y, local_z)
!     index of fftw vector in k-space
! - real(kind=doubleR) function greens_function(diff_comp, k_squared, &
!                                    first_derivative,second_derivative)
!     Green's function for the differentiation
! - subroutine init_fftw(igrid)
!     creates the plans for FFTW and gets the number of planes per processor
!     calls: create_plan()                            fmax-fftw
! - subroutine create_plan()
!     creates a plan for fftw
!     calls: rfftw3d_f77_mpi_create_plan(...)         fftw
!            rfftwnd_f77_mpi_local_sizes(...)         fftw
!            rfftw3d_f77_create_plan(...)             fftw
! - subroutine forward_transform()
!     wrapper for a forward transform with fftw
!     calls: rfftwnd_f77_mpi(...)                     fftw
!            rfftwnd_f77_one_real_to_complex(...)     fftw
! - subroutine reverse_transform ()
!     wrapper for a reverse transform with fftw
!     calls: rfftwnd_f77_mpi(...)                     fftw
!            rfftwnd_f77_one_complex_to_real(...)     fftw


subroutine real_derivative(first_derivative, second_derivative)

! reads or generates a potential, differentiates it in the k-space
! and calls the wrapper for fftw

  use pinocchio_mpi
  use pinocchio_constants
  use fftw_variables
  use fmax_run
  use fmax_real_derivative

  implicit none
  character(len=24):: fdate
  external fdate
#ifdef SPRNG
!!#include <mpif.h>
#include "sprng_f.h"
#endif
  integer(kind=singleI),intent(in)::first_derivative, second_derivative

! local variables
  logical::swap
  integer(kind=singleI),parameter::Ne=16
  integer(kind=singleI)::local_x,local_y,local_z,ix,iy,iz,ixx, iyy, izz,iyp,istart
  integer(kind=singleI)::istream
  integer(kind=singleI)::index,nxhalf,nyhalf,nzhalf
!  integer(kind=singleI),save :: first_call=0
  integer(kind=singleI),external :: init_random
  integer(kind=singleI)::iloop
  integer(kind=doubleI),external :: index_kspace
!
  real(kind=doubleR):: kx,ky,kz,kxnorm,kynorm,kznorm,green,spectrum &
       ,amplitude,k_cutoff,k_squared,k_module,phase,smoothing &
       ,random1, random2, tmp, k_physical, norm
  real(kind=doubleR),dimension(0:3)::diff_comp=1.d0
  real(kind=doubleR),external::power,random,greens_function
!
  character(len=80)::info, outfile

#ifdef MPI
  real(kind=doubleR) :: cpu(10)
#endif

  integer(kind=singleI):: streamnum, nstreams, sprng_seed, junk
#ifdef SPRNG
  SPRNG_POINTER stream, stream1, stream2
#else
  integer(kind=singleI)::stream
#endif

! normalization of the k-space modes
  norm = 2.d0*pi**2 / cell_size**3 * dble(nx_fftw)*dble(ny_fftw)*dble(nz_fftw)

! k vectors
  kxnorm   = 2.*pi/dble(nx_fftw)
  kynorm   = 2.*pi/dble(ny_fftw)
  kznorm   = 2.*pi/dble(nz_fftw)

! Nyquist frequencies
  nxhalf = nx_fftw/2
  nyhalf = ny_fftw/2
  nzhalf = nz_fftw/2

! Initialize random number sequence
! using sprng:  number of independent streams to generate, and starting seed
  nstreams    = nz_fftw  
#ifdef SPRNG
  sprng_seed  = saved_seed
#endif

! initialise plane-id's for determining random stream so that large scale power 
! is independent of resolution
  index = 0
  plane_id(1) = index
  do iy=1, ny_fftw/2
     iyp = -iy + ny_fftw
     index = index + 1 ;  plane_id(iy+1)  = index
     if(iy .lt. ny_fftw/2) then
        index = index + 1 ;  plane_id(iyp+1) = index
     endif
  enddo

!!$  ! test random number generator
!!$  do iy=0, ny_fftw-1
!!$     ! random_seed for this y-plane
!!$     streamnum  = plane_id(iy + 1)
!!$#ifdef SPRNG
!!$     stream     = init_sprng(streamnum,nstreams,sprng_seed,SPRNG_DEFAULT)
!!$     if(proc_id .eq. 0 .and. iy.eq.0 .and. first_call.eq.0) then
!!$        call message(' real_derivative: sprng random generator initialised (GOOD): ')
!!$        junk = print_sprng(stream)
!!$     endif
!!$#else
!!$     stream     = init_random(streamnum)
!!$     if(proc_id .eq. 0 .and. iy.eq.0 .and. first_call.eq.0)) then
!!$        call message('real_derivative:  ran3 random generator initialised (AVOID): ')
!!$     endif
!!$#endif
!!$     ! write some values for this stream
!!$     write (info,'(''random_test_plane='',i4.4,''_procs='',i4.4)') iy, proc_id
!!$     open(unit=1,file=info,status='unknown')
!!$     do iloop=0,64
!!$#ifdef SPRNG              
!!$        random1   = sprng(stream)
!!$        random2   = sprng(stream)
!!$#else
!!$        random1   = random(stream)
!!$        random2   = random(stream)
!!$#endif
!!$        write (1,'(i5,1x,2(e16.8,1x))') iloop, random1, random2
!!$     enddo
!!$#ifdef SPRNG
!!$     junk = free_sprng(stream)
!!$#endif
!!$     close(1)
!!$  enddo
!!$  call message('random sequence written')
!!$  first_call = 1

! for first derivatives the real and imaginary parts must be swapped
  swap=.false.
  if ((first_derivative==0 .and. second_derivative> 0) .or. &
      (first_derivative> 0 .and. second_derivative==0))     &
       swap=.true.

#ifdef MEMTEST
  call message(' starting generation of grid ')
#endif

#ifdef MPI
  cpu(1) = MPI_WTIME()
#endif

#ifndef MEMTEST  

! NB: ix, iy and iz are global coordinates of the box (from 0 to N-1)
!     ix: [0,N/2]
!     iy, iz: [0,N-1]
!     iyy and izz are unfolded, they run from -N/2+1 to N/2 (ixx = ix)
!     local_x, local_y and local_z are local coordinates of the slice
!     (for the x and z coordinates they are the same as ix and iz)


! loop over k-space indices

! loop over planes in y
  write(info,'(''['',a24,''] real_derivative: init random numbers '')') fdate()
  call message (trim(info))
  loopy: do local_y=0,local_ny_after_transform-1
     iy  = local_y + local_y_start_after_transform
     iyy = iy ; if (iy > nyhalf) iyy = iy-ny_fftw
     ky  = kynorm*dble(iyy)

! start random sequence for this y-plane
     if (generate_field) then
        streamnum  = plane_id(iy + 1)
#ifdef SPRNG
        stream     = init_sprng(streamnum,nstreams,sprng_seed,SPRNG_DEFAULT)
#else
        stream     = init_random(saved_seed,ThisGrid,streamnum)
#endif
     endif
     
! in this case the loop is done on concentric frames in the ixx-izz space
     loop: do iloop = 0, max(nzhalf,nxhalf)
        
        loopz: do izz = -iloop,iloop

           if(abs(izz) .eq. iloop) then
              istart = 0
           else
              istart = iloop
           endif
           
           loopx: do ixx = istart, iloop
              
! the loop cycles if the point is not on the iloop frame (which means that 
! either ixx or abs(izz) are not equal to iloop) or if the loop is going on 
! non-existent points (this happens for non-cubic boxes)
              if (abs(izz) /= iloop .and. ixx /= iloop) then
                 write (*,*) ' error in loop ',iloop,ixx,izz
                 call abort_code(' error ')
              endif
              
              if ( izz <= -nzhalf .or. izz >nzhalf  &
                   .or. ixx < 0 .or. ixx > nxhalf) cycle 
              
              iz = izz ; if (izz<0) iz = izz + nz_fftw
              local_z = iz
              kz = kznorm*dble(izz)
              
              ix = ixx
              local_x = ix
              kx = kxnorm*dble(ixx)
              
              k_squared  = kx**2 + ky**2 + kz**2
              k_module   = sqrt(k_squared)
              k_physical = k_module/cell_size
              
! corresponding index of real part in vector (imaginary in index + 1)
              index = index_kspace(local_x, local_y, local_z)
#ifdef IDLTEST
              if (index .lt. 0 .or. index .gt. total_local_size-1) call abort_code(' error in index ')
#endif

              if(generate_field) then
#ifdef SPRNG
                 random1   = sprng(stream)
                 random2   = sprng(stream)
#else
                 random1   = random(stream)
                 random2   = random(stream)
#endif
              endif

              if (k_squared /= 0.) then

! Gaussian smoothing window
                 smoothing = exp(-0.5d0*k_squared*Rsm_grid**2)

                 if (generate_field) then
! power spectrum with high- and low-frequency cutoffs
                    spectrum  = power(k_physical) * norm         &
                         *        exp(-(k_module/upper_k_cutoff)**Ne) 
                    if (lower_k_cutoff > 0.) spectrum = spectrum &
                         *(1.d0 - exp(-(k_module/lower_k_cutoff)**Ne))
                    
                    amplitude = sqrt(-spectrum*log(random1)) * smoothing
#ifdef WHITE_NOISE
                    amplitude = sqrt(-log(random1))
                    if(k_module > pi*32) amplitude = 0.d0
#endif
! phase is set to 0 for k-vectors on Nyquist edge
                    if ((ix==0 .or. ix==nxhalf) .and. &
                         (iy==0 .or. iy==nyhalf) .and. &
                         (iz==0 .or. iz==nzhalf)) then
                       phase = 0.d0
                    else
                       phase = 2.d0*pi*random2
                    endif

                    vector(index)   = amplitude*cos(phase)
                    vector(index+1) = amplitude*sin(phase)

                 else
                    
                    vector(index)   = vector(index)   * smoothing
                    vector(index+1) = vector(index+1) * smoothing

                 endif

! the components are stored in the vectors that control the differentiation
#ifdef  WHITE_NOISE
                 green     = 1.d0
#else
                 diff_comp = (/1.d0,kx,ky,kz/)
                 green     = greens_function(diff_comp, k_squared, first_derivative, second_derivative)
#endif

                 vector(index)   =  green * vector(index)    ! real
                 vector(index+1) =  green * vector(index+1)  ! imaginary
              else
! fundamental mode
                 vector(index)   = 0.
                 vector(index+1) = 0.
              endif
              
              if (swap) then
                 tmp             = vector(index+1)
                 vector(index+1) = vector(index)     ! imaginary
                 vector(index)   = - tmp             ! real
              endif
           enddo loopx
        enddo loopz
     enddo loop
     
  
     if (generate_field) then
#ifdef SPRNG
! release random number stream
        junk = free_sprng(stream)
#endif
     endif
  enddo loopy
#endif


!!!! release seeds of ran3
!!!#ifndef SPRNG
!!!  deallocate(stream_seed)
!!!#endif  

#ifdef MPI
  cpu(1) = MPI_WTIME() - cpu(1)
#endif

#ifdef MEMTEST
  call message(' ... done  generation of grid ')
  write (info,'('' timing for generation of grid : '',f8.2))') cpu(1)
  call message(info)
#endif
  
! reverse complex_to_real FFT
  write(info,'(''['',a24,''] real_derivative: starting fft '')') fdate()
  call message (trim(info))
#ifdef MEMTEST
  call message(' starting reverse transform ')
#endif

#ifdef WHITE_NOISE  
  norm      = dble(nx_fftw)*dble(ny_fftw)*dble(nz_fftw)
  norm      = sqrt(norm)
  vector    = norm * vector
#endif

#ifdef MPI
  cpu(2) = MPI_WTIME()
#endif

  call reverse_transform ()

#ifdef MPI
  cpu(2) = MPI_WTIME() - cpu(2)
#endif
  
  write(info,'(''['',a24,''] real_derivative: done fft '')') fdate()
  call message (trim(info))
#ifdef MEMTEST
  call message(' done reverse transform ')
  write (info,'('' timing for generating grid = '',f8.2,'' and for reverse fft : '',f8.2))') cpu(1), cpu(2)
  call message(info)
#endif

  return
end subroutine real_derivative


function index_rspace(local_x, local_y, local_z)

! index of vector in real-space

  use numbers
  use fftw_variables, ONLY: nx_fftw, ny_fftw, nz_fftw
  implicit none
  integer(kind=singleI), intent(in) :: local_x, local_y, local_z
  integer(kind=doubleI) :: index_rspace
  
#ifdef MPI
  index_rspace = 1 + local_x + local_y * (nx_fftw+2) + local_z * (nx_fftw+2) * ny_fftw
#else
  index_rspace = 1 + local_x + local_y * (nx_fftw+2) + local_z * (nx_fftw+2) * ny_fftw
#endif
end function index_rspace


function index_kspace(local_x, local_y, local_z)

! index of vector in k-space

  use numbers
  use fftw_variables, ONLY: nx_fftw, ny_fftw, nz_fftw
  implicit none
  integer(kind=singleI), intent(in) :: local_x, local_y, local_z
  integer(kind=doubleI) :: index_kspace

#ifdef MPI
  index_kspace = 1 + 2*local_x + (nx_fftw+2)*(local_z + local_y*nz_fftw)
#else
  index_kspace = 1 + 2*local_x + (nx_fftw+2)*(local_y + local_z*ny_fftw)
#endif

end function index_kspace

function greens_function(diff_comp, k_squared, first_derivative, second_derivative)

! green function for the differentiation

  use numbers
  implicit none
  real(kind=doubleR) :: greens_function
  integer(kind=singleI),intent(in)::first_derivative,second_derivative
  real(kind=doubleR),intent(in)::diff_comp(0:3)
  real(kind=doubleR),intent(in)::k_squared

! local variables
  real(kind=doubleR)::sign,kvect(3),green00
  integer(kind=singleI)::der


! in this case the greens_function is simply 1
  if (first_derivative == -1 .and. second_derivative == -1) then
     greens_function = 1.d0
     return
  end if

#ifdef IDLTEST
  kvect = diff_comp(1:3)
  ! no derivative, pure potential
  green00 = 2.d0*(cos(kvect(1))+cos(kvect(2))+cos(kvect(3))-3.d0)
  if(abs(green00) > 0) green00=1.d0/green00
  if (first_derivative==0 .and. second_derivative==0) then
     greens_function = 1.d0

! first derivatives
  else if (first_derivative==0 .or. second_derivative==0) then
     der = max(first_derivative, second_derivative)
     greens_function = sin(kvect(der))

! second derivatives
  else if (first_derivative == second_derivative) then
     greens_function = 2.d0*(cos(kvect(first_derivative))-1.d0) 
  else
     greens_function = -2.d0 * sin(kvect(first_derivative)) &
                             * sin(kvect(second_derivative))
  endif
  greens_function = greens_function * green00
#else
  if (first_derivative==0 .and. second_derivative==0) then
     greens_function = - diff_comp(first_derivative)*diff_comp(second_derivative)/k_squared
  else
     greens_function =   diff_comp(first_derivative)*diff_comp(second_derivative)/k_squared
  endif

#endif

  return
end function greens_function


subroutine init_fftw(igrid)

! Creates the plans for FFTW and gets the number of planes per processor

  use fmax_grids
  use pinocchio_mpi
  use fftw_variables
 
  implicit none
  integer(kind=singleI),intent(in)::igrid

! local variables
  integer(kind=singleI)::igrid2
  character(len=80)::info
  integer(kind=singleI):: nx, ny, nz
!  integer(kind=doubleR):: fp, rp

! checks whether the plans have been computed for an equal grid
  do igrid2=1,igrid-1

     if (Lgrid(igrid,1) == Lgrid(igrid2,1) .and. &
         Lgrid(igrid,2) == Lgrid(igrid2,2) .and. &
         Lgrid(igrid,3) == Lgrid(igrid2,3)) then

        write (info,'(''Grid '',i1,'' is equal to grid '',i1)')igrid,igrid2
        call message(info)

        saved_forward_plan(igrid)  = saved_forward_plan(igrid2)
        saved_reverse_plan(igrid)  = saved_reverse_plan(igrid2)

        Lgrid_local(igrid,:) = Lgrid_local(igrid2,:)
        Lgrid_fft(igrid,:)   = Lgrid_fft(igrid2,:)
        Lgrid_z_start(igrid) = Lgrid_z_start(igrid2)
        Lgrid_y_start(igrid) = Lgrid_y_start(igrid2)
        npart_fft(igrid)     = npart_fft(igrid2)
        npart_local(igrid)   = npart_local(igrid2)

        return
     endif
  enddo

! else create plans
  nx_fftw = Lgrid(igrid,1) ; ny_fftw = Lgrid(igrid,2) ; nz_fftw = Lgrid(igrid,3)

  call create_plan()

! local grid in real space for each processor
  saved_forward_plan(igrid) = forward_plan
  saved_reverse_plan(igrid) = reverse_plan
  Lgrid_local(igrid,1:2)    = Lgrid(igrid,1:2)
  Lgrid_local(igrid,3)      = local_nz
  Lgrid_z_start(igrid)      = local_z_start
  npart_local(igrid)        = product(Lgrid_local(igrid,:))
! local grid in Fourier space for each processor
  Lgrid_fft(igrid,1)        = Lgrid(igrid,1)/2+1
  Lgrid_fft(igrid,2)        = local_ny_after_transform
  Lgrid_fft(igrid,3)        = Lgrid(igrid,3)
  Lgrid_y_start(igrid)      = local_y_start_after_transform
  npart_fft(igrid)          = total_local_size

  return
end subroutine init_fftw


subroutine create_plan()

! Creates a plan for fftw

  use pinocchio_mpi
  use fftw_variables
  implicit none

#ifdef MPI
  call rfftw3d_f77_mpi_create_plan(forward_plan ,mpi_comm_world  & 
       , nx_fftw, ny_fftw, nz_fftw &
       , FFTW_REAL_TO_COMPLEX , FFTW_ESTIMATE + FFTW_IN_PLACE)

  call rfftw3d_f77_mpi_create_plan(reverse_plan ,mpi_comm_world &
       , nx_fftw, ny_fftw, nz_fftw &
       , FFTW_COMPLEX_TO_REAL, FFTW_ESTIMATE)

  call rfftwnd_f77_mpi_local_sizes(forward_plan, local_nz, local_z_start, &
       local_ny_after_transform, local_y_start_after_transform, total_local_size)

#else
  call rfftw3d_f77_create_plan(forward_plan & !,mpi_comm_world  & 
       , nx_fftw, ny_fftw, nz_fftw &
       , FFTW_REAL_TO_COMPLEX , FFTW_ESTIMATE + FFTW_IN_PLACE)

  call rfftw3d_f77_create_plan(reverse_plan & !,mpi_comm_world &
       , nx_fftw, ny_fftw, nz_fftw &
       , FFTW_COMPLEX_TO_REAL, FFTW_ESTIMATE + FFTW_IN_PLACE)

  local_nz = nz_fftw
  local_z_start = 0
  local_ny_after_transform = ny_fftw
  local_y_start_after_transform = 0
  total_local_size = (nx_fftw +2) * ny_fftw * nz_fftw

#endif

  return
end subroutine create_plan


subroutine forward_transform()

! wrapper for a forward transform with fftw

  use pinocchio_mpi
  use fftw_variables
  implicit none

  
#ifdef MPI
  call rfftwnd_f77_mpi(forward_plan, 1, vector, fftw_work, use_work_fftw &
       , FFTW_TRANSPOSED_ORDER)
#else
  call rfftwnd_f77_one_real_to_complex(forward_plan, vector, vector)
#endif

  return
end subroutine forward_transform


subroutine reverse_transform ()

! wrapper for a reverse transform with fftw

  use pinocchio_mpi
  use fftw_variables
  implicit none
  real(kind=doubleR) :: norm

! normalization
  norm = 1.d0/dble(nx_fftw)/dble(ny_fftw)/dble(nz_fftw)

#ifdef MPI
  call rfftwnd_f77_mpi(reverse_plan, 1, vector &
       , fftw_work, use_work_fftw, FFTW_TRANSPOSED_ORDER)
#else
  call rfftwnd_f77_one_complex_to_real(reverse_plan, vector, vector) 
#endif


  vector = vector * norm

end subroutine reverse_transform
