!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

! This file contains the routines that initialize the fmax run

! SUBROUTINES AND FUNCTIONS IN THIS FILE
!
! - subroutine initialization()
!     is a driver that calls all the other routines:
!     calls: set_run()                                fmax-init
!            set_cosmology()                          fmax-init
!            set_grids()                              fmax-init
!            set_smoothing_radii()                    fmax-init
!            set_old_smoothing_radii()                fmax-init
!            test_allocations()                       fmax-init
!            open_tmp_files()                         fmax-tmpfiles
!            check_no_output()                        fmax-init
!            read_density_files()                     fmax-init
! - subroutine set_run()
!     reads the run information from the fmax.info, $run.info, $run.grids and
!     $run.pos files
!     calls: allocate_grid_vectors()                  fmax-alloc
!            allocate_positions()                     fmax-alloc
! - subroutine check_fs()
!     checks that the filesystem is ready for the run
!     calls: check_dir                                fmax-init
! - subroutine set_cosmology()
!     normalizes the power spectrum and outputs it in an ascii file
!     calls: normalization()                          cosmo
!            grow(z)                                  cosmo
!            power(k)                                 cosmo
! - subroutine set_grids()
!     initializes the vectors that keep information on the nested grids
!     calls: init_fftw(igrid)                         fmax-fftw
! - subroutine set_smoothig_radii()
!     computes the smoothing radii for the run
!     calls: allocate_smoothing_radii()               fmax-alloc
!            aLambda(r)                               cosmo
!            Radius(Lambda)                           cosmo
!            grow(z)                                  cosmo
! - subroutine set_old_smoothing_radii()
!     performs the old 1.0 computation
!     calls: allocate_smoothing_radii()               fmax-alloc
!            aLambda(r)                               cosmo
!            Radius(Lambda)                           cosmo
! - subroutine test_allocations()
!     performs all the allocations, to verify that memory is sufficient
!     calls: assign_fftw_variables(igrid)             fmax-subs
!            allocate_vector()                        fmax-alloc
!            deallocate_vector()                      fmax-alloc
!            allocate_plane(*)                        fmax-alloc
!            deallocate_plane(*)                      fmax-alloc
! - subroutine read_density_files()
!     reads in the linear density for the grids that need it
!     calls: assign_fftw_variables(igrid)             fmax-subs
!            allocate_vector()                        fmax-alloc
!            read_one_density_field(igrid,file)       fmax-rw
!            forward_transform()                      fmax-fftw
!            save_kspace_vector(igrid)                fmax-tmpfiles
!            deallocate_vector()                      fmax-alloc
! - subroutine check_dir(filename)
!     checks the existence of a directory and aborts if it does not exist


subroutine initialization()

! Initializations for the Fmax run

  implicit none

! acquires the run basic information
  call set_run()

! checks that the filesystem is ready for the run
  call check_fs()

! sets the cosmological parameters and normalizes the power spectrum
  call set_cosmology()

! sets the information for all the grids to be used and creates fftw plans
  call set_grids()

! sets the smoothing radii
#ifdef OLD_RADII
  call set_old_smoothing_radii()
#else
  call set_smoothing_radii()
#endif

! tests that there is enough space for allocations
  call test_allocations()

! opens all the scratch tmp files
  call open_tmp_files()

! reads the initial density fields
  call read_density_files()


! bye!
  return
end subroutine initialization


subroutine set_run()

! Reads all the basic information for the run
! file: fmax.info         contains: basic passwords for the run
! file: $run.info         contains: cosmology and box size
! file: $run.grids        contains: info on the other grids
! file: $run.pos          contains: positions to place the smallest box

  use numbers
  use pinocchio_mpi
  use pinocchio_common
  use fmax_grids
  use fmax_smoothing_radii
  use fmax_run
  use alloc_flag

  implicit none
! local variables
  integer(kind=singleI)       :: iproc, Lgrid1, igrid, ipos, flag_read_IC
  real(kind=doubleR),parameter:: small=1.e-6
  real(kind=doubleR)          :: overlap, Mtot
  character(len=80)           :: word
  character(len=1)            :: label
  logical                     :: cubic=.false., success=.false., exist

  integer::ii

! selects the running processor
  do iproc=0,nproc-1
     if(iproc .eq. proc_id) then

! Checks that the Data/, tmp_local/ and tmp_global/ directories exist



! **********************
! READING fmax.info FILE
! **********************

! finds name of run
        open (10,file='fmax.info',status='old',err=30)
        read (10,'(a6)')run

! reads commands from the fmax.info file
        ngrid              = 0
        saved_seed         = 0
        list_of_positions  = .false.
!        optimize_ram       = .false.
        final_redshift     = 0.
        limit_in_delta     = -100.
        alloc_verbose      = .false.
        use_scratch_for_derivatives = .false.
        use_scratch_for_all         = .false.

        do while (.true.)
           read (10,'(a)',end=20,err=50) word        
           if (trim(word)=='') cycle

           if (trim(word)=='MULTIPLE_GRIDS') then
              read (10,*,end=50,err=50) ngrid

           else if (trim(word)=='GENERATE_DENSITY_FIELD') then
              read (10,*,end=50,err=50) random_seed
              saved_seed = abs(random_seed)

           else if (trim(word)=='LIST_OF_POSITIONS') then
              list_of_positions=.true.

!           else if (trim(word)=='OPTIMIZE_RAM') then
!              optimize_ram  = .true.

           else if (trim(word)=='ALLOC_VERBOSE') then
              alloc_verbose=.true.

           else if (trim(word)=='FINAL_REDSHIFT') then
              read (10,*,end=50,err=50) final_redshift

           else if (trim(word)=='LIMIT_IN_DELTA') then
              read (10,*,end=50,err=50) limit_in_delta

           else if (trim(word)=='USE_SCRATCH_FOR_DERIVATIVES') then
              use_scratch_for_derivatives=.true.

           else if (trim(word)=='USE_SCRATCH_FOR_ALL') then
              use_scratch_for_derivatives=.true.
              use_scratch_for_all        =.true.

           else if (trim(word)=='OUTPUT_RMAX') then
              output_rmax=.true.              

! add here more choices

           else
! warns in case of unknown commands
              call message('WARNING: command unkown in fmax.info file')
              call message(word)
           endif

        enddo
20      close(10)

! Checks consistency and outputs choices
        if (ngrid<0) then
           call message ('ERROR: the number of grids must be positive')
           call abort_code   ('set_run: please correct the fmax.info file and run again')
        endif
        if (ngrid>5) then
           call message ('ERROR: this version handles no more than 5 boxes')
           call abort_code   ('set_run: please correct the fmax.info file and run again')
        endif
        if (final_redshift<-1.) then
           call message ('wrong final_redshift specified, assuming 0')
           final_redshift=0.
        endif

        if (ngrid==0 .or. ngrid==1) then
           call message ('Single grid run')
           ngrid = 1
           if (list_of_positions) then
              call message ('ERROR: list of positions specified in a single-grid run')
              call abort_code   ('set_run: please correct the fmax.info file and run again')
           endif
        else
           write(word,'(''Multi-grid run with '',i3,'' grids'')')ngrid
           call message (word)
           if (ngrid>5) then
              call message ('We do not suggest such a large number of grids!')
              call abort_code('')
           endif
        endif

        if (saved_seed>0) then
           write (word,'(''ICs will be generated with the seed '',i12)') saved_seed
           call message (word)
        else
           call message ('ICs will be read from the files:')
           do igrid=1,ngrid
              write(word,'(''Data/Density_g'',i1,''.d'')')igrid
              call message(word)
           enddo
        endif


        write (word,'(''Final redshift for the run: ''f7.3)') final_redshift
        call message (word)

        write (word,'(''The limit in delta for computing Fmax is: ''f10.5)') limit_in_delta
        call message (word)

! **********************
! READING $run.info file
! **********************

! allocates vectors for dimensions of boxes (valid for general ngrid-values)
        call allocate_grid_vectors()

! NB the $run.info file refers to the last, highest-resolution grid (igrid=ngrid)
        success=.false.
        do while(.not.success)
           open  (unit=10,file=run//'.info',status='old',err=40)
           read(10,*)BoxSize(1)
           read(10,*)Omega0
           read(10,*)OmegaLambda
           read(10,*)PowerSpectrum
           read(10,*)Sigma8
           read(10,*)Hubble100
           if(cubic)then
              read(10,*,err=41) Lgrid1
              success = .true.
              Lgrid(ngrid,:)  = Lgrid1
           else
              read(10,*,err=41) Lgrid(ngrid,:)
              success  = .true.
           endif
           
41         close(10)
           if(success) then
              InterPartDist = BoxSize(1)/dble(Lgrid(ngrid,1))
              BoxSize(:)    = InterPartDist*Lgrid(ngrid,:)
              InterPartDistGrids(ngrid) = InterPartDist
              BoxSizeGrids(ngrid,:) = BoxSize
              Mtot   = 2.778e11*Hubble100**2*Omega0*product(sngl(BoxSize))
              Mpart  = dble(Mtot/float(product(Lgrid)))

           endif
           cubic = .true.
        enddo

! *************
! Reading grids
! *************
        if (ngrid>1) then

! reads dimensions of all boxes (but the last one) from run.grids file
           open (unit=10,file=run//'.grids',status='old',err=60)
           do igrid=1,ngrid-1
              read (10,*,end=60,err=60) Lgrid(igrid,:),BoxSizeGrids(igrid,1) &
                   ,k_cutoff(igrid) ,flag_cutoff(igrid), flag_read_IC
              InterPartDistGrids(igrid) = BoxSizeGrids(igrid,1)/dble(Lgrid(igrid,1))
              BoxSizeGrids(igrid,:)     = InterPartDistGrids(igrid)*Lgrid(igrid,:)
              if (flag_read_IC == 1) then
                 read_IC(igrid)=.true.
                 write(label,'(i1)') igrid
                 InputDensityFile(igrid) = 'Data/Density_g'//label//'.d'
              else
                 read_IC(igrid)=.false.
                 InputDensityFile(igrid) = ''
              endif

           enddo
           close(10)

        endif

! *************************
! Reading list of positions
! *************************
        if (list_of_positions) then
           open(10,file=run//'.pos',status='old',err=70)
           read(10,*,end=70,err=70)npos
           call allocate_positions()
           do ipos=1,npos
              read(10,*,end=70,err=70)positions(ipos,:)
           enddo
           close(10)
        endif

! ***
! end
! ***

     endif
#ifdef MPI
     call mpi_barrier(mpi_comm_world,ierr)
#endif
  enddo
 
  call message('.. set run done')

! bye!
 return
30 call message(' ERROR: file fmax.info not found ')
  call abort_code   (' set_run: file not found ')
40 call message(' ERROR: file '//run//'.info not found ')
  call abort_code   (' set_run: file not found ')
50 call message(' ERROR in reading fmax.info file')
  call abort_code   (' set_run: please correct the fmax.info file and run again')
60 call message(' ERROR in reading file '//run//'.grids')
  call abort_code   (' set_run: please correct '//run//'.grids and run again')
70 call message(' ERROR in reading file '//run//'.pos')
  call abort_code   (' set_run: please correct '//run//'.pos and run again')

end subroutine set_run


subroutine set_cosmology()
! Initializes the cosmology

  use pinocchio_common
  use fmax_run
  use pinocchio_mpi
  implicit none
  real(kind=doubleR),external::grow,power

  real(kind=doubleR)::ppp

  call normalization()
  filter=0
  final_growing_mode = grow(final_redshift)

  if (proc_id==0) then
!     OUTPUT COSMOLOGICI: LAMBDA, P(k)
  endif

  call message('.. Power-spectrum normalization computed ')
  
  return
end subroutine set_cosmology


subroutine set_grids()

! computes relevant quantities for handling multiple grids
! including the plans for fftw

  use fmax_grids
  use fmax_run
  implicit none

! local variables
  integer(kind=singleI):: igrid, ix, iy, iz, ipos

  do igrid=1,ngrid
! number of grid points (particles) in a grid
     npart(igrid) = product(Lgrid(igrid,:))

! dimension of plane in real space
     nplane(igrid) = Lgrid(igrid,1)*Lgrid(igrid,2)

! number of grid particles overlapping with the smallest grid
     coarse_level(igrid,:) = BoxSizeGrids(ngrid,:)/InterPartDistGrids(igrid)

! dimension of the sub-box overlapping with the smallest one
     where (((coarse_level(igrid,:)+0.1) >= 1.))
        Lgrid_overlap(igrid,:) = int(coarse_level(igrid,:)+0.1)
     elsewhere
        Lgrid_overlap(igrid,:) = 1
     end where

! initializes fftw and local grids
     call init_fftw(igrid)

  enddo

  Lgrid_total = nint(BoxSizeGrids(1,:)/InterPartDistGrids(ngrid))
  npart_total = product(Lgrid_total)

  call message('.. set_grids done')

  return
end subroutine set_grids


subroutine set_old_smoothing_radii()

! Sets the smoothing radii
! This is taken from Pinocchio-1.0 and is used for testing purposes

  use numbers
  use fmax_smoothing_radii
  use fmax_run
  use pinocchio_common
  use pinocchio_mpi
  use fmax_grids

  implicit none
  integer(kind=singleI)::i
  real(kind=doubleR),parameter::varmin=-1.1d0
  real(kind=doubleR)::rmax,rmin
  real(kind=doubleR),external::aLambda,Radius

  call message('WARNING: using set_old_smoothing_radii')

  if (Lgrid(ngrid,1)==256) then
     nsmooth=26
  else if (Lgrid(ngrid,1)==128) then
     nsmooth=15
  else if (Lgrid(ngrid,1)==64) then
     nsmooth=10
  else
     call abort_code('Unknown grid in SET_OLD_SMOOTHING_RADII')
  endif

  call allocate_smoothing_radii()

  rmax=Radius(10.**varmin)
  rmin=InterPartDist/6.

  do i=1,nsmooth-1
     Smoothing_Radii(i)=10.**(log10(rmax)-(i-1.)* &
          (log10(rmax)-log10(rmin))/dble(nsmooth))
     Smoothing_Variances(i)=aLambda(Smoothing_Radii(i))
  enddo
  Smoothing_Radii(nsmooth)=0.d0
  Smoothing_Variances(nsmooth)=aLambda(InterPartDist/10.d0)

  return
end subroutine set_old_smoothing_radii


subroutine set_smoothing_radii()

! Decides the smoothing radii for the run

! we chose smoothing scales such that the variance of the
! smoothed density field (the square of the mass variance) 
! ranges between sigma_min and sigma_max, in
! logarithmic steps of width step_var

! var_min is chosen so that, at the final_redshift, delta=1.686 is 
! an n-sigma event, i.e. var_min = (1.686/nsigma)**2/grow(final_redshift)**2

! var_max is estimated as the variance of the Gaussian-smoothed field
! at 1/6th of the inter-particle distance

! the last variance value refers to the unsmoothed field 

  use fmax_smoothing_radii
  use fmax_run
  use pinocchio_common
  use pinocchio_mpi

  implicit none

! local variables
  integer(kind=singleI)::i
  real(kind=doubleR),parameter::nsigma=6.0
  real(kind=doubleR),parameter::var_min_z0=(1.686/nsigma)**2
  real(kind=doubleR),parameter::step_var=0.15
  real(kind=doubleR)::rmin,rmax,var_min,var_max
  real(kind=doubleR),external::aLambda,Radius,grow
  character(len=80)::info

  var_min = var_min_z0/final_growing_mode**2
  rmax    = Radius(var_min)
  rmin    = InterPartDist/6.
  var_max = aLambda(rmin)
  nsmooth = (log10(var_max)-log10(var_min))/step_var+2

  write(info,'(''Minimum and maximum variance, nsmooth: '',2(1x,f12.6),1x,i3)') &
       var_min,var_max,nsmooth
  call message(trim(info))

  call allocate_smoothing_radii(nsmooth)

  do i=1,nsmooth-1
     Smoothing_Variances(i)=10.**(log10(var_min)+step_var*(i-1))
     Smoothing_Radii(i)=Radius(Smoothing_Variances(i))
  enddo
  Smoothing_Variances(nsmooth)=aLambda(InterPartDist/10.)
  Smoothing_Radii(nsmooth)=0.
  
  call message('.. set_smoothing_radii: done ')

  return
end subroutine set_smoothing_radii


subroutine test_allocations()

! allocates and deallocates matrices to test that memory is enough

  use fmax_grids
  use fftw_variables, ONLY: fftw_verbose
  use fmax_smoothing_radii
  use alloc_flag
  implicit none

! local variables
  integer(kind=singleI)::ismooth,igrid
  logical::old_alloc_verbose

  call message ('Starting allocation test (alloc_verbose inactive)...')

  fftw_verbose = 0
  old_alloc_verbose = alloc_verbose
  alloc_verbose = .false.
  do ismooth=1,nsmooth

     do igrid=1,ngrid
        call assign_fftw_variables(igrid)
        call allocate_vector()
        call deallocate_vector()
     enddo

     call allocate_plane(2)
     call allocate_tmp_plane(ngrid)
     call deallocate_tmp_plane()
     call deallocate_plane(2)

     do igrid=1,ngrid
        call assign_fftw_variables(igrid)
        call allocate_vector()
        call deallocate_vector()
     enddo

     call allocate_plane(1)
     call allocate_tmp_plane(ngrid)
     call deallocate_tmp_plane()
     call deallocate_plane(1)

  enddo
  fftw_verbose = 1
  alloc_verbose = old_alloc_verbose
  call message('Allocation test done')

  return
end subroutine test_allocations


subroutine check_fs()

! Checks that the filesystem is ready for the run

  use numbers
  use pinocchio_common
  use fmax_run
  use fmax_grids
  use save_directory
  implicit none

! local variables
  character(len=80):: filename
  logical::flag
  integer(kind=singleI)::igrid

  call check_dir('Data')
  call check_dir(local_directory)
!  call check_dir(common_directory)

  filename='Data/Fmax.'//run//'.d'
  inquire(file=filename,exist=flag)
  if (flag) &
       call abort_code ('ERROR in check_fs: existent '//trim(filename))

  filename='Data/Velx.'//run//'.d'
  inquire(file=filename,exist=flag)
  if (flag) &
       call abort_code ('ERROR in check_fs: existent '//trim(filename))

  filename='Data/Vely.'//run//'.d'
  inquire(file=filename,exist=flag)
  if (flag) &
       call abort_code ('ERROR in check_fs: existent '//trim(filename))

  filename='Data/Velz.'//run//'.d'
  inquire(file=filename,exist=flag)
  if (flag) &
       call abort_code ('ERROR in check_fs: existent '//trim(filename))

  do igrid=1,ngrid
     inquire(file=InputDensityFile(igrid),exist=flag)
     if (.not.read_IC(igrid)) then
        if (flag) call abort_code ('ERROR in check_fs: file Data/'&
             //trim(InputDensityFile(igrid))//' should not exist')
     else
        if (.not.flag) call abort_code ('ERROR in check_fs: file Data/'&
             //trim(InputDensityFile(igrid))//' not found')
     endif
  enddo

  return
end subroutine check_fs


subroutine check_dir(filename)

! checks the existence of a directory and aborts if it does not exist

  use numbers
  implicit none
  character(len=*),intent(in)::filename
  integer(kind=singleI),external::direxist

  integer(kind=singleI)::i,j

  i=len_trim(trim(filename))
  j=direxist(filename,i)

  if (j==0) then
     call abort_code('ERROR: directory '//trim(filename)//' does not exist')
  else
     call message('Found directory '//trim(filename))
  endif

  return
end subroutine check_dir



subroutine read_density_files()

! Reads the ICs if it is required
  use pinocchio_mpi
  use fmax_grids
  use fmax_run
  implicit none

! local variables
  integer(kind=singleI)::igrid
  
  do igrid= 1,ngrid
     if (read_IC(igrid)) then

! allocates the vector
        call assign_fftw_variables(igrid)
        call allocate_vector()

! Reads the density file
        call read_one_density_field(igrid,InputDensityFile(igrid))

! forward transform
        call forward_transform()

! saves the result to a tmp file
        call save_kspace_vector(igrid)

        call deallocate_vector()

     endif
  enddo

  return
end subroutine read_density_files
