!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

! Modules for fmax-2.2

module pinocchio_common
! This module contains the basic properties of the (highest-resolution) box, 
! the cosmological parameters, the run name and other parameters needed by
! cosmo-dp-2.2
  use numbers
! Cosmological parameters
  real(kind=doubleR):: Omega0  
  real(kind=doubleR):: OmegaLambda
  real(kind=doubleR):: Hubble100
  real(kind=doubleR):: Sigma8
  real(kind=doubleR):: PowerSpectrum  ! flag for the transfer function
  real(kind=doubleR):: csp            ! normalization constant for P(k)
  real(kind=doubleR):: delta_c        ! delta_c Press & Schechter parameter
  integer(kind=singleI):: filter      ! flag for window filter
  integer(kind=singleI):: flag_ps     ! flag for the analytical mass function
! Basic properties of the highest resolution box
  real(kind=doubleR):: InterPartDist  ! comoving inter-particle distance
  real(kind=doubleR):: BoxSize(3)     ! 3D comoving size
  real(kind=doubleR):: Mpart          ! particle mass
!
  character(len=6)  :: run, kernel    ! name of the run

  real(kind=doubleR),parameter::Omegab=0.0456   ! baryonic Omega
end module pinocchio_common

module pinocchio_constants
! CONSTANTS
  use numbers
  real(kind=doubleR), parameter :: pi   =3.1415926535897931d0
  real(kind=doubleR), parameter :: twopi=2.d0*pi
end module pinocchio_constants

module pinocchio_mpi
! numbers required to handle mpi
  use numbers
#ifdef MPI
  include 'mpif.h'
#endif
  integer(kind=singleI) :: ierr, k_offset, proc_id, nproc
end module pinocchio_mpi

#ifdef HDF5
module pinocchio_hdf5
! numbers required to handle hdf5
  use numbers
  integer(kind=singleI) :: file_handle, group_handle, attribute_handle
end module pinocchio_hdf5
#endif

module fmax_run
! info on the run
  use numbers
  logical::list_of_positions !,optimize_ram
  logical::use_scratch_for_derivatives=.false.,use_scratch_for_all=.false.
  logical::output_rmax=.false.
  integer(kind=singleI)::random_seed, saved_seed
  real(kind=doubleR)::final_redshift, limit_in_delta, final_growing_mode
end module fmax_run


module fmax_grids
  use numbers
! information on multiple grids
! NB: the code deals with ngrid grids (ngrid<=5)
!     the index goes in order of increasing mass resolution or decreasing particle mass
!     the highest-resolution grid is always the last (ngrid-th) one
!     some information is in common with pinocchio_common
!
! DEFINITIONS:
!   The GLOBAL box is divided into SLICES, each slice is given to a processor.
!   Each slice contains one or more PLANES, with sizes 
!   Slicing in real-space is on the z direction 
!     (so each slice contains a number of z-planes)
!   Slicing in k-space is on the y direction 
!     (so each slice contains a number of y=planes)

! number of grids
  integer(kind=singleI)::ngrid
! 3D lengths of the global grids (number of particles or grid vertices per side)
  integer(kind=singleI),allocatable,dimension(:,:):: Lgrid
! 3D lengths of the slice assigned to the processor
  integer(kind=singleI),allocatable,dimension(:,:):: Lgrid_local
! 3D length of the k-space slice assigned to the processor
  integer(kind=singleI),allocatable,dimension(:,:):: Lgrid_fft
! Number of particles per side of grid number igrid that overlap with 
!   the highest resolution box (number ngrid) (>=1)
  integer(kind=singleI),allocatable,dimension(:,:):: Lgrid_overlap
! Floating-point value of Lgrid_overlap (can be <1)
  real(kind=doubleR),allocatable,dimension(:,:)   :: coarse_level
! starting z-plane for the real-space slice
  integer(kind=singleI),allocatable,dimension(:)  :: Lgrid_z_start
! starting y-plane for the k-space slice
  integer(kind=singleI),allocatable,dimension(:)  :: Lgrid_y_start
! comoving box sizes
  real(kind=doubleR),allocatable,dimension(:,:)   :: BoxSizeGrids
! comoving inter-particle distances
  real(kind=doubleR),allocatable,dimension(:)     :: InterPartDistGrids
! comoving lower cutoff to apply to the box
  real(kind=doubleR),allocatable,dimension(:)     :: k_cutoff
! kind of cutoff (0=rectangular, 1=spherical)
  integer(kind=singleI),allocatable,dimension(:)  :: flag_cutoff
! number of positions to place the highest resolution box
  integer(kind=singleI)                           :: npos
! positions to place the lowest resolution box
  integer(kind=singleI),allocatable,dimension(:,:):: positions
! number of particles in the global grid
  integer(kind=doubleI),allocatable,dimension(:)  :: npart
! number of particles in the real-space slice
  integer(kind=doubleI),allocatable,dimension(:)  :: npart_local
! number of particles in the k-space slice
  integer(kind=doubleI),allocatable,dimension(:)  :: npart_fft
! number of particles in a z-plane
  integer(kind=singleI),allocatable,dimension(:)  :: nplane
! Total number of grid vertices for a box of the comoving size of igrid=1 and
!   the inter-particle distance of igrid=ngrid
  integer(kind=singleI)                           :: Lgrid_total(3)
! number of particles in the "total" box
  integer(kind=doubleI)                           :: npart_total
! fftw plans
  integer(kind=doubleI),allocatable,dimension(:)  :: saved_forward_plan, saved_reverse_plan
! Flag for reading or generating initial conditions for the grids
  logical,allocatable,dimension(:)                :: read_IC
! Files with the input initial conditions (if relevant)
  character(len=30),allocatable,dimension(:)      :: InputDensityFile
end module fmax_grids


module fftw_variables
! These module contains the variables that are passed to the fftw wrapper.
! They are re-initializes each time that a grid is considered
  use numbers
!  include '/opt/local/fftw/64/current/include/fftw_f77.i'
  include 'fftw_f77.i'
! grid size (saved in Lgrid(igrid,:))
  integer(kind=singleI)     :: nx_fftw, ny_fftw, nz_fftw
! fftw plans (saved in saved_forward_plan and saved_reverse_plan)
  integer(kind=doubleI)     :: forward_plan, reverse_plan
! number of real-space z-planes and starting z-plane for the slice
!   (saved in Lgrid_local(igrid,3) and Lgrid_z_start) 
  integer(kind=singleI)     :: local_nz, local_z_start 
! number of k-space y-planes and starting y-plane for the slice
!   (saved in Lgrid_fft(igrid,2) and Lgrid_y_start)
  integer(kind=singleI)     :: local_ny_after_transform, local_y_start_after_transform 
! ids of the y-planes of the slice
  integer(kind=singleI), allocatable:: plane_id(:)
! total number of particles in the slice (saved in npart_fft)
  integer(kind=singleI)     :: total_local_size
! flag for extra fftw-related output
  integer(kind=singleI)     :: fftw_verbose
! grid considered (equivalent to igrid)
  integer(kind=singleI)     :: ThisGrid

! These are the largest matrices to be allocated by fmax
! vector contains the field to process with fftw (either in real- or k-space)
! fftw_work is a workspace for fftw
  real(kind=doubleR),allocatable,dimension(:)::vector
  real(kind=doubleR),allocatable,dimension(:)::fftw_work
! this flag tells fft to use the allocated workspace
  integer(kind=singleI),parameter :: use_work_fftw=1
end module fftw_variables


module fmax_one_plane
! plane for the computation of the collapse times
  use numbers
  real(kind=doubleR),allocatable,dimension(:,:,:)::differentials_plane,vmax_plane
  real(kind=doubleR),allocatable,dimension(:,:)::Fmax_plane
  integer(kind=singleI),allocatable,dimension(:,:)::Rmax_plane
end module fmax_one_plane


module fmax_smoothing_radii
! smoothing radii used by fmax, theorical and actual variances of the smoothed
! linear density field
  use numbers
  integer(kind=singleI)::nsmooth
  real(kind=doubleR),allocatable,dimension(:)::Smoothing_Radii, Smoothing_Variances, True_variances
end module fmax_smoothing_radii


module fmax_box_sections
! sections of the large-scale boxes, to be used for the interpolation
  use numbers
  real(kind=doubleR),allocatable,dimension(:,:,:,:)::box1, box2, box3, box4
! ??????????????????????
  integer(kind=singleI),allocatable,dimension(:)::deltax,deltay,deltaz
! deformation tensor or displacement vector, to be summed over the boxes
  real(kind=doubleR)::diff_ten(6)
end module fmax_box_sections


module fmax_real_derivative
! variables needed by real_derivative
  use numbers
  logical::generate_field
  real(kind=doubleR)::Rsm_grid  ! smoothing radius in grid coordinates
  real(kind=doubleR)::upper_k_cutoff, lower_k_cutoff   ! cutoffs in k-space, grid coordinates
  real(kind=doubleR)::cell_size ! inter-particle distance
end module fmax_real_derivative


module tmp_plane
! plane used for reading and writing tmp and output files
  use numbers
#if defined(DPTMP) || defined(IDLTEST)
  real(kind=doubleR),allocatable,dimension(:,:)::plane
#else
  real(kind=singleR),allocatable,dimension(:,:)::plane
#endif
  integer(kind=singleI),allocatable,dimension(:,:)::int_plane
end module tmp_plane


module save_directory
  use numbers
! name of directory for the files
!  character(len=24)::local_directory='/export/disc/2/simulate/'
  character(len=10)::local_directory ='tmp_local/'
  character(len=11)::common_directory='tmp_common/'
end module save_directory


module alloc_flag
  logical::alloc_verbose=.true.
end module alloc_flag

