!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


! This file contains the routines for reading and writing files
! If the complier option is set, it uses HDF5.
! Otherwise, it can read and write binary (default) or ascii (debug)
!
! NB Data are always written plane by plane through the plane array
!    defined in module tmp_plane and allocated by allocate_tmp_ldsplane;
!    this array is singleR, unless the DPTMP flag is specified
!
!
! SUBROUTINES AND FUNCTIONS IN THIS FILE
!
! - subroutine write_fmax_fields(file,file_flag,igrid)
!     drives the writing of Fmax, vmax and density matrices to files
!     calls: open_file_to_write(iproc,unit,file)             fmax-rw
!            read_tmp_plane(filename)                        fmax-tmpfiles
!            write_plane(unit,global_z)                      fmax-rw
!            read_tmp_integer_plane(filename)                fmax-tmpfiles
!            write_integer_plane(unit,global_z)              fmax-rw
!            close_file(unit)                                fmax-rw
!            write_string(i)                                 fmax-rw
! - subroutine read_one_density_field(igrid,file)
!     reads a density file from the disc and puts it in the fftw vector
!     calls: open_file_to_read(iproc,unit,file)              fmax-rw
!            read_plane(unit,global_z)                       fmax-rw
!            plane_to_vector(local_z)                        fmax-tmpfiles
!            close_file(unit)                                fmax-rw
!            write_string(i)                                 fmax-rw
! - subroutine open_file_to_write(iproc,unit,file)
!     opens a file (in fortran or HDF5 mode) and writes the headers if relevant
!     calls: hdf5_open_file(unit, file)                      hdf5-wrapper
!            hdf5_create_file(unit, file)                    hdf5-wrapper
!            rw_cosmology(write)                             fmax-rw
!            rw_grid(write)                                  fmax-rw
!            rw_fmax(write)                                  fmax-rw
!            write_pedantic_header                           fmax-rw
!            write_cosm_header                               fmax-rw
! - subroutine open_file_to_read(iproc,unit,file)
!     opens a file (in fortran or HDF5 mode) and reads the headers if relevant
!     calls: hdf5_open_file(unit, file, readonly=.true.)     hdf5-wrapper
!            rw_cosmology(read)                              fmax-rw
!            rw_grid(read)                                   fmax-rw
!            rw_fmax(read)                                   fmax-rw
!            read_pedantic_header                            fmax-rw
!            read_cosm_header                                fmax-rw
! - subroutine close_file(unit)
!     closes a file
!     calls: hdf5_close_file(unit)                           hdf5-wrapper
! - subroutine write_plane(unit,iplane)
!     writes a plane to a file in standard or HDF5 format
!     calls: hdf5_create_group(unit,field)                   hdf5-wrapper
!            hdf5_write_data(unit,field,plane)               hdf5-wrapper
!            write_string(i)                                 fmax-rw
! - subroutine read_plane(unit,iplane)
!            Writes a plane in standard or hdf5 format. 
!            In this case the plane is given as an argument.
!            This is used by fragment-2.2
!     calls: hdf5_read_data(unit,field,plane)                hdf5-wrapper
!            write_string(i)                                 fmax-rw
! - subroutine write_integer_plane(unit,iplane)
!     writes an integer plane to a file in standard or HDF5 format
!     calls: hdf5_create_group(unit,field)                   hdf5-wrapper
!            hdf5_write_data(unit,field,plane)               hdf5-wrapper
!            write_string(i)                                 fmax-rw
! - subroutine read_integer_plane(unit,iplane)
!            Writes an integer plane in standard or hdf5 format. 
!     calls: hdf5_read_data(unit,field,plane)                hdf5-wrapper
!            write_string(i)                                 fmax-rw
! - subroutine read_plane_fragment(unit,iplane)
!     reads a plane from a file in standard or HDF5 format
!     this is used by fragment-2.2
!     calls: hdf5_read_data(unit,field,plane)                hdf5-wrapper
!            write_string(i)                                 fmax-rw
! - subroutine write_cosm_header()
!     writes a cosmological header to a standard format file
! - subroutine read_cosm_header()
!     reads a cosmological header from a standard format file
! - subroutine write_pedantic_header()
!     writes a consistency header to a standard format file
! - subroutine read_pedantic_header()
!     reads a consistency header from a standard format file
! - subroutine rw_grid(action)
!     reads/writes a header with grid properties from/to a HDF5 file
!     calls: hdf5_create_group(unit,field)                   hdf5-wrapper
!            hdf5_write_data(unit,field,plane)               hdf5-wrapper
!            hdf5_read_data(unit,field,plane)                hdf5-wrapper
! - subroutine rw_cosmology(action)
!     reads/writes a header with cosmological parameters from/to a HDF5 file
!     calls: hdf5_create_group(unit,field)                   hdf5-wrapper
!            hdf5_write_data(unit,field,plane)               hdf5-wrapper
!            hdf5_read_data(unit,field,plane)                hdf5-wrapper
! - subroutine rw_fmax(action)
!     reads/writes a header with run parameters from/to a HDF5 file
!     calls: hdf5_create_group(unit,field)                   hdf5-wrapper
!            hdf5_write_data(unit,field,plane)               hdf5-wrapper
!            hdf5_read_data(unit,field,plane)                hdf5-wrapper
! - character(len=4) function write_string(i)
!     writes a number from 0 to 9999 into a 4 char. string padded with 0s


! This file contains also these auxiliary subroutines to handle MPI functions:
! - subroutine message(messg)
! - subroutine abort_code(messg)
! - subroutine finish(messg)


#ifdef ASCII_READ
#  define READ_STATEMENT read(1,*,err=20)
#else
#  define READ_STATEMENT read(1,err=20)
#endif
#ifdef ASCII_WRITE
#  define WRITE_STATEMENT write(1,*,err=20)
#else
#  define WRITE_STATEMENT write(1,err=20)
#endif


subroutine write_fmax_fields(file,file_flag,igrid)

! writes to file
! consistency header
! cosmological parameters
! data(Lgrid(1),Lgrid(2),Lgrid(3)) slice in (x,y) planes

  use pinocchio_mpi
  use pinocchio_common
  use fmax_one_plane
  use fmax_grids
  use save_directory
  implicit none

! external variables
  character(*)         , intent(in) :: file
  integer(kind=singleI), intent(in) :: file_flag,igrid

! local variables
  integer(kind=singleI)    :: iproc, local_z, global_z, unit, tmp_unit
  logical                  :: first_time=.true.,scratch_flag
  character(len=80)        :: info
  character(len=80)        :: filename

  if (file_flag<0 .or. file_flag>5) call abort_code('Wrong call of write_fmax_fields')

  if (first_time) then
#ifndef HDF5
#ifdef ASCII_WRITE
     call message('writing ascii (debug)')
#else
     call message('writing binary')
#endif
#else
     call message('writing in HDF5 format')
#endif
     first_time=.false.
  endif

  do iproc=0,nproc-1
     if(iproc .eq. proc_id) then

! open the file and write the headers
        unit=1
        call open_file_to_write(iproc,unit,file)

! write data plane by plane
        do local_z=0,Lgrid_local(ngrid,3)-1
           global_z = local_z + Lgrid_z_start(ngrid)

           if (file_flag<=4) then
              call make_filename(-file_flag,0,0,global_z,filename,tmp_unit,scratch_flag)
           else 
              call make_filename(igrid,2,0,global_z,filename,tmp_unit,scratch_flag)
           endif

           if (file_flag/=4) then
! read the plane
              call read_tmp_plane(filename,tmp_unit,scratch_flag)

! write the plane
              call write_plane(unit,global_z,file)
           else
! read the plane
              call read_tmp_integer_plane(filename,tmp_unit,scratch_flag)

! write the plane
              call write_integer_plane(unit,global_z,file)
           endif

        enddo
        call close_file(unit)
     endif

#ifdef MPI
! MPI barrier
     call mpi_barrier(mpi_comm_world,ierr)
#endif
  enddo

! exit
  info = ' written '//trim(file)
  call message(trim(info))
  return
20 call message(' error writing output file '//trim(file))
  call abort_code(' write_fmax_fields ')
end subroutine write_fmax_fields


subroutine read_one_density_field(igrid,file)

! reads the linear density for grid igrid from a file
! and stores it in the fftw vector
  use fftw_variables, only : nx_fftw, ny_fftw, nz_fftw
  use pinocchio_mpi
  use pinocchio_common
  use fmax_one_plane
  use fmax_grids
  use save_directory
  use tmp_plane
  implicit none

! external variables
  integer(kind=singleI), intent(in) :: igrid
  character(*)         , intent(in) :: file

! local variables
  logical::flag
  integer(kind=singleI)    :: iproc, local_x, local_y, local_z, index, global_z, unit
  integer(kind=singleI)::Lx,Ly,Lz,ii
  logical                  :: first_time=.true.
  character(len=80)        :: info
  character(len=80)        :: filename
  real(kind=doubleR)       :: local_variance, local_average, global_variance, global_average


  if (first_time) then
#ifdef IDLTEST
     call message(' reading input file from idl ')
#else
#ifndef HDF5
#ifdef ASCII_WRITE
     call message('reading ascii (debug)')
#else
     call message('reading binary')
#endif
#else
     call message('reading in HDF5 format')
#endif
#endif
     first_time=.false.
  endif

#ifdef IDLTEST
  inquire(file=file,exist=flag)
  if (.not.flag) call abort_code ('File '//trim(file)//' does not exist')
  procs: do iproc=0,nproc-1
     if(iproc .eq. proc_id) then
        open(unit=1,file=file,form='unformatted',status='old')
        read (1) Lx, Ly, Lz
        if(Lx .ne. nx_fftw .or. Ly .ne. ny_fftw .or. Lz .ne. nz_fftw) &
             call abort_code(' size of grid is not correct')
! skip to planes to read
        do local_z=0,Lgrid_z_start(igrid)-1
           read (1)
           read (1)
        enddo
        local_variance = 0.0
        local_average  = 0.0
        do local_z=0,Lgrid_local(igrid,3)-1
           global_z = local_z + Lgrid_z_start(igrid)
           read (1) ii
           if(ii .ne. global_z) &
                call abort_code(' read_one_denisty_field: reading wrong plane')
           read (1) plane
           call plane_to_vector(local_z)
           
           local_variance = local_variance + sum(plane**2)
           local_average  = local_average  + sum(plane)
           
        enddo
        close(1)
     endif
#ifdef MPI
! MPI barrier
     call mpi_barrier(mpi_comm_world,ierr)
#endif
  enddo procs

#else
!
  do iproc=0,nproc-1
     if(iproc .eq. proc_id) then

! open the file and read the headers
        unit=1
        call open_file_to_read(iproc,unit,file,.false.)

! read data plane by plane
        do local_z=0,Lgrid_z_start(igrid)-1
           call read_plane(unit,-1)
        enddo
        do local_z=0,Lgrid_local(igrid,3)-1
           global_z = local_z + Lgrid_z_start(igrid)

           call read_plane(unit,global_z)
           call plane_to_vector(local_z)

           local_variance = local_variance + sum(plane**2)
           local_average  = local_average  + sum(plane)

        enddo
        call close_file(unit)
     endif

#ifdef MPI
! MPI barrier
     call mpi_barrier(mpi_comm_world,ierr)
#endif
  enddo
#endif ! idltest

#ifdef MPI
  CALL MPI_REDUCE(local_variance, global_variance, 1, MPI_DOUBLE_PRECISION, MPI_SUM, 0, mpi_comm_world, ierr)
  CALL MPI_REDUCE(local_average , global_average , 1, MPI_DOUBLE_PRECISION, MPI_SUM, 0, mpi_comm_world, ierr)
#else
  global_variance = local_variance
  global_average  = local_average
#endif

  if(proc_id .eq. 0) then
     global_variance = global_variance / dble(npart(igrid))
     global_average = global_average / dble(npart(igrid))
     write (*,*) ' variance = ',global_variance
     write (*,*) ' average = ',global_average
  endif

! exit
  info = ' read input file '//trim(file)
  call message(trim(info))
  return
20 call message(' error reading input file '//trim(file))
  call abort_code('read_one_density_field')
end subroutine read_one_density_field


subroutine open_file_to_write(iproc,unit,file)

! opens a file for writing in standard or hdf5 mode
! processor 0 writes the headers

  use numbers
#ifdef HDF5
  use hdf5_wrapper; use pinocchio_hdf5
#endif
  implicit none
  integer(kind=singleI),intent(in)::iproc
  integer(kind=singleI),intent(inout)::unit
  character(len=*),intent(in)::file
#ifdef ASCII_WRITE
  character(len=*),parameter :: form='formatted'
#else
  character(len=*),parameter :: form='unformatted'
#endif

! local variables
  logical::flag
  character,parameter :: read='read',write='write',test='test'

  call message('opening file '//trim(file)//' for writing')
#ifdef HDF5
  inquire(file=file,exist=flag)
  if (flag) then
     call hdf5_open_file(file_handle, file)
  else
     call hdf5_create_file(file_handle, file)
     call rw_cosmology(write)       
     call rw_grid(write)
     call rw_fmax(write)
  endif
#else

#ifdef ONE_FILE_PER_PLANE
  if (iproc==0) then
! processor 0 writes the header
     open(unit=unit,file=trim(file)//".header",status='unknown',form=form)
     call write_pedantic_header()
     call write_cosm_header()
     close(unit=unit)
  endif
#else
  if (iproc==0) then
! processor 0 opens file and writes header
     open(unit=unit,file=file,status='unknown',form=form)
     call write_pedantic_header()
     call write_cosm_header()
  else
! others append
     open(unit=unit,file=file,status='old',form=form,position='append', err=20)
  endif
#endif

#endif

  return
20 call abort_code('ERROR in open_file_to_write')
end subroutine open_file_to_write


subroutine open_file_to_read(iproc,unit,file,read_headers)

! opens a file for reading in standard or hdf5 mode
! processor 0 reads the headers

#ifdef HDF5
  use hdf5_wrapper; use pinocchio_hdf5
#endif
  use numbers
  implicit none
  integer(kind=singleI),intent(in)::iproc
  integer(kind=singleI),intent(inout)::unit
  character(len=*),intent(in)::file
  logical,intent(in)::read_headers

! local variables
  logical::flag
#ifdef ASCII_WRITE
  character(len=*),parameter :: form='formatted'
#else
  character(len=*),parameter :: form='unformatted'
#endif
  character,parameter :: read='read',write='write', test='test'

  inquire(file=file,exist=flag)
  if (.not.flag) call abort_code ('File '//trim(file)//' does not exist')

  call message('opening file '//trim(file)//' for reading')

#ifdef HDF5
  call hdf5_open_file(file_handle, file, readonly=.true.)
  if (iproc==0 .and. read_headers) then
     call rw_cosmology(test)
     call rw_grid(test)
     call rw_fmax(test)
  endif
  return
#endif
! open file and read header
  open(unit=unit,file=file,status='old',form=form)
!!  if (iproc==0) then
     call read_pedantic_header()
     call read_cosm_header()
!!  endif

  return
end subroutine open_file_to_read


subroutine close_file(unit)

! Close a file

  use numbers
#ifdef HDF5
  use hdf5_wrapper; use pinocchio_hdf5
#endif
  implicit none
  integer(kind=singleI),intent(in):: unit

#ifdef HDF5
  call hdf5_close_file(file_handle)
#else
  close(1)
#endif

  return
end subroutine close_file


subroutine write_plane(unit,iplane,file)

! Writes a plane in standard or hdf5 format
  
  use numbers
#ifdef HDF5
  use hdf5_wrapper; use pinocchio_hdf5
#endif
  use tmp_plane
  implicit none
  integer(kind=singleI),intent(in):: unit,iplane
  character(len=*),intent(in)     :: file
  character(len=4),external       :: write_string
  character(len=11)               :: field
  character(len=80)               :: varname
  character(len=5)                :: label
#ifdef ASCII_WRITE
  character(len=*),parameter :: form='formatted'
#else
  character(len=*),parameter :: form='unformatted'
#endif

#ifdef HDF5
  field='/Plane'//write_string(iplane)//'/'
  varname = trim(field)//'Data'

  call hdf5_create_group(file_handle,field)
!!$  write (*,*) ' created group: ',field
!!$  write (*,*) ' writing data : ',varname
  call hdf5_write_data(file_handle,varname,plane)
#else

#ifdef ONE_FILE_PER_PLANE
  if (iplane<10) then
     write(label,'(".000",i1)')iplane
  else if (iplane<100) then
     write(label,'(".00",i2)')iplane
  else if (iplane<1000) then
     write(label,'(".0",i3)')iplane
  else 
     write(label,'(".",i4)')iplane
  endif
  open(unit=unit,file=trim(file)//label,status='new',form=form)
#endif

! plane ID
  WRITE_STATEMENT iplane
! data
#ifdef ASCII_WRITE
  write (1,'(8(1x,e13.6))',err=20) plane
#else
  write (1,err=20) plane
#endif

#ifdef ONE_FILE_PER_PLANE
  close(unit)
#endif

#endif

  return
20 call abort_code ('ERROR in write_plane')
end subroutine write_plane


subroutine read_plane(unit,iplane)

! Writes a plane in standard or hdf5 format
  
  use numbers
#ifdef HDF5
  use hdf5_wrapper; use pinocchio_hdf5
#endif
  use tmp_plane
  implicit none
  integer(kind=singleI),intent(in):: unit,iplane
  character(len=4),external       :: write_string
  character(len=11)               :: field
  integer(kind=singleI)           :: ii
  character(len=80)               :: info
  integer(kind=singleI)           :: OneInt
  real(kind=doubleR)              :: OneReal
  character(len=80)               :: varname

#ifdef HDF5
  field='/Plane'//write_string(iplane)//'/'
  varname = trim(field)//'Data'
  call hdf5_read_data(file_handle,varname,plane)
  return
#endif
! plane ID
  READ_STATEMENT ii
  if (ii/=iplane .and. iplane .ge. 0) then
     write(info,'(''ERROR read_plane, index: '',2(1x,i9))')ii,iplane
     call abort_code (trim(info))
  endif
! data
#ifdef ASCII_READ
  read (1,*,err=20) plane
#else
  read (1,err=20) plane
#endif

  return
20 call abort_code ('ERROR in read_plane')
end subroutine read_plane


subroutine write_integer_plane(unit,iplane,file)

! Writes an integer plane in standard or hdf5 format
  
  use numbers
#ifdef HDF5
  use hdf5_wrapper; use pinocchio_hdf5
#endif
  use tmp_plane
  implicit none
  integer(kind=singleI),intent(in):: unit,iplane
  character(len=*),intent(in)     :: file
  character(len=4),external       :: write_string
  character(len=11)               :: field
  character(len=80)               :: varname
  character(len=5)                :: label
#ifdef ASCII_WRITE
  character(len=*),parameter :: form='formatted'
#else
  character(len=*),parameter :: form='unformatted'
#endif

#ifdef HDF5
  field='/Plane'//write_string(iplane)//'/'
  varname = trim(field)//'Data'
  call hdf5_create_group(file_handle,field)
!!$  write (*,*) ' created group: ',field
!!$  write (*,*) ' writing data : ',varname
  call hdf5_write_data(file_handle,varname,int_plane)
#else

#ifdef ONE_FILE_PER_PLANE
  if (iplane<10) then
     write(label,'(".000",i1)')iplane
  else if (iplane<100) then
     write(label,'(".00",i2)')iplane
  else if (iplane<1000) then
     write(label,'(".0",i3)')iplane
  else 
     write(label,'(".",i4)')iplane
  endif
  open(unit=unit,file=trim(file)//label,status='new',form=form)
#endif

! plane ID
  WRITE_STATEMENT iplane
! data
#ifdef ASCII_WRITE
  write (1,'(8(1x,i12))',err=20) int_plane
#else
  write (1,err=20) int_plane
#endif

#ifdef ONE_FILE_PER_PLANE
  close(unit)
#endif

#endif

  return
20 call abort_code ('ERROR in write_integer_plane')
end subroutine write_integer_plane


subroutine read_integer_plane(unit,iplane)

! Writes an integer plane in standard or hdf5 format
  
  use numbers
#ifdef HDF5
  use hdf5_wrapper; use pinocchio_hdf5
#endif
  use tmp_plane
  implicit none
  integer(kind=singleI),intent(in):: unit,iplane
  character(len=4),external       :: write_string
  character(len=11)               :: field
  integer(kind=singleI)           :: ii
  character(len=80)               :: info
  integer(kind=singleI)           :: OneInt
  real(kind=doubleR)              :: OneReal
  character(len=80)               :: varname

#ifdef HDF5
  field='/Plane'//write_string(iplane)//'/'
  varname = trim(field)//'Data'
  call hdf5_read_data(file_handle,varname,int_plane)
  return
#endif
! plane ID
  READ_STATEMENT ii
  if (ii/=iplane .and. iplane .ge. 0) then
     write(info,'(''ERROR read_integer_plane, index: '',2(1x,i9))')ii,iplane
     call abort_code (trim(info))
  endif
! data
#ifdef ASCII_READ
  read (1,*,err=20) int_plane
#else
  read (1,err=20) int_plane
#endif

  return
20 call abort_code ('ERROR in read_integer_plane')
end subroutine read_integer_plane


subroutine read_plane_fragment(unit,iplane,Lx,Ly,plane)

! Writes a plane in standard or hdf5 format
! In this case the plane is given as an argument and not in a module
! This is used by fragment-2.2
  
  use numbers
#ifdef HDF5
  use hdf5_wrapper; use pinocchio_hdf5
#endif
  implicit none
  integer(kind=singleI),intent(in):: unit,iplane,Lx,Ly
  real(kind=singleR),dimension(Lx,Ly),intent(inout)  :: plane
  character(len=4),external       :: write_string
  character(len=11)               :: field
  integer(kind=singleI)           :: ii
  character(len=80)               :: info, var_name
  integer(kind=singleI)           :: OneInt
  real(kind=doubleR)              :: OneReal

#ifdef HDF5
  field='/Plane'//write_string(iplane)//'/'
  var_name = trim(field)//'Data'
  call hdf5_read_data(file_handle,var_name,plane)
  return
#endif
! plane ID
  READ_STATEMENT ii
  if (ii/=iplane .and. iplane .ge. 0) then
     write(info,'(''ERROR read_plane, index: '',2(1x,i9))')ii,iplane
     call abort_code (trim(info))
  endif
! data
#ifdef ASCII_READ
  read (1,*,err=20) plane
#else
  read (1,err=20) plane
#endif

  return
20 call abort_code ('ERROR in read_plane')
end subroutine read_plane_fragment


subroutine write_cosm_header()

! write cosmological parameters
  use pinocchio_constants
  use pinocchio_common
  use fftw_variables, ONLY: nx_fftw, ny_fftw, nz_fftw
  implicit none

! local variables
  real(kind=singleR)::Omega0s,OmegaLambdas,Hubble100s,Sigma8s,PowerSpectrums,InterPartDists

!!!  Omega0s        = Omega0
!!!  OmegaLambdas   = OmegaLambda
!!!  Hubble100s     = Hubble100
!!!  Sigma8s        = Sigma8
!!!  PowerSpectrums = PowerSpectrum
!!!  InterPartDists = InterPartDist
!!!
!!!  WRITE_STATEMENT Omega0s,OmegaLambdas,Hubble100s,Sigma8s,PowerSpectrums
!!!  WRITE_STATEMENT L
!!!  WRITE_STATEMENT InterPartDists

  WRITE_STATEMENT nx_fftw, ny_fftw, nz_fftw

  return
20 call message(' error writing cosmological header ')
  call abort_code('write_cosm_header  ')
end subroutine write_cosm_header


subroutine read_cosm_header()

! write cosmological parameters
  use pinocchio_constants
  use pinocchio_common
  use fftw_variables, ONLY: nx_fftw, ny_fftw, nz_fftw
  implicit none

! local variables
!  real(kind=singleR)::Omega0s,OmegaLambdas,Hubble100s,Sigma8s,PowerSpectrums,InterPartDists
  integer(kind=singleI)::Lx,Ly,Lz
  character(len=80)::info

!!!  Omega0s        = Omega0
!!!  OmegaLambdas   = OmegaLambda
!!!  Hubble100s     = Hubble100
!!!  Sigma8s        = Sigma8
!!!  PowerSpectrums = PowerSpectrum
!!!  InterPartDists = InterPartDist
!!!
!!!  READ_STATEMENT Omega0s,OmegaLambdas,Hubble100s,Sigma8s,PowerSpectrums
!!!  READ_STATEMENT L
!!!  READ_STATEMENT InterPartDists

  READ_STATEMENT Lx, Ly, Lz
  if (Lx /= nx_fftw .or. Ly /= ny_fftw .or. Lz /= nz_fftw) then
     call message('WARNING: SEVERE PROBLEM')
     call message('Inconsistent grid data in fortran file')
     write (info,'(''Parameters'',5(1x,i9))')Lx,Ly,Lz
     call message(trim(info))
  else
     call message('grid lengths match')
  endif

  return
20 call message(' error reading cosmological header ')
  call abort_code('read_cosm_header  ')
end subroutine read_cosm_header


subroutine write_pedantic_header()

  ! writes a header to check consistency of binary numbers

  implicit none
!!!  integer(kind=singleI),parameter ::integ=1
!!!  integer(kind=doubleI),parameter ::long=1
!!!  real(kind=singleR)   ,parameter ::single=1.
!!!  real(kind=doubleR)   ,parameter ::double=1.d0
!!!  complex(kind=4),parameter ::scompl=(1.,1.)
!!!  complex(kind=8),parameter ::dcompl=(1.d0,1.d0)
!!!
!!!  WRITE_STATEMENT 1
!!!  WRITE_STATEMENT integ                   ! integer(kind=4)
!!!  WRITE_STATEMENT 2
!!!  WRITE_STATEMENT long                    ! integer(kind=8)
!!!  WRITE_STATEMENT 3
!!!  WRITE_STATEMENT single                  ! real(kind=singleR)
!!!  WRITE_STATEMENT 4
!!!  WRITE_STATEMENT double                  ! real(kind=doubleR) 
!!!  WRITE_STATEMENT 5
!!!  WRITE_STATEMENT scompl                ! complex(kind=4)
!!!  WRITE_STATEMENT 6
!!!  WRITE_STATEMENT dcompl                ! complex(kind=8)

  return
20 call message(' error writing pedantic header ')
  call abort_code(' write_pedantic_header ')
end subroutine write_pedantic_header

subroutine read_pedantic_header()
  implicit none
  return
end subroutine read_pedantic_header
!!!logical function read_pedantic_header()
!!!
!!!  ! checks whether definition of 
!!!  ! integer*4 integer *8
!!!  ! real*4    real*8
!!!  ! complex*4 complex*8
!!!  ! are same as in data file
!!!  ! 
!!!
!!!  implicit none
!!!  integer(kind=4),parameter ::integ=1
!!!  integer(kind=8),parameter ::long=1
!!!  real(kind=singleR)   ,parameter ::single=1.
!!!  real(kind=doubleR)   ,parameter ::double=1.d0
!!!  complex(kind=4),parameter ::scompl=(1.,1.)
!!!  complex(kind=8),parameter ::dcompl=(1.d0,1.d0)
!!!
!!!  integer(kind=4) ::read_integ
!!!  integer(kind=8) ::read_long
!!!  real(kind=singleR)    ::read_single

#ifdef HDF5
subroutine rw_grid(action)
  use hdf5_wrapper; use pinocchio_hdf5
  use fmax_grids
  use fftw_variables, ONLY: ThisGrid
  implicit none
  character(*), intent(in) :: action

  character,parameter :: read='read',write='write',test='test'
  character(len=80)::info
  integer(kind=singleI)::ngrid1,ThisGrid1,Lx,Ly,Lz

  if(action .eq. write) then
     CALL hdf5_create_group(file_handle,'/Grid/')  
     call hdf5_write_data(file_handle,'/Grid/Ngrid',ngrid)
     call hdf5_write_data(file_handle,'/Grid/ThisGrid',ThisGrid)  
     call hdf5_write_data(file_handle,'/Grid/nx',Lgrid(ThisGrid,1))
     call hdf5_write_data(file_handle,'/Grid/ny',Lgrid(ThisGrid,2))
     call hdf5_write_data(file_handle,'/Grid/nz',Lgrid(ThisGrid,3))
  elseif(action .eq. read) then
     call hdf5_read_data(file_handle,'/Grid/Ngrid',ngrid)
     call hdf5_read_data(file_handle,'/Grid/ThisGrid',ThisGrid)  
     call hdf5_read_data(file_handle,'/Grid/nx',Lgrid(ThisGrid,1))
     call hdf5_read_data(file_handle,'/Grid/ny',Lgrid(ThisGrid,2))
     call hdf5_read_data(file_handle,'/Grid/nz',Lgrid(ThisGrid,3))
  elseif(action .eq. test) then
     !! warning: inserted by hand
     ngrid = 1
     ThisGrid = 1
     !! 
     call hdf5_read_data(file_handle,'/Grid/Ngrid',ngrid1)
     call hdf5_read_data(file_handle,'/Grid/ThisGrid',ThisGrid1)  
     call hdf5_read_data(file_handle,'/Grid/nx',Lx)
     call hdf5_read_data(file_handle,'/Grid/ny',Ly)
     call hdf5_read_data(file_handle,'/Grid/nz',Lz)
  else
     call abort_code(' invalid action in rw_grid')
  endif
end subroutine rw_grid


subroutine rw_cosmology(action)
  use hdf5_wrapper; use pinocchio_hdf5
  use pinocchio_common
  implicit none
  character(*), intent(in) :: action
  character(len=200)        :: info
  character,parameter :: read='read',write='write',test='test'

  real(kind=doubleR)::diff,par1,par2,par3,par4,par5(3),par6,par7,par8
  real(kind=doubleR),parameter::accuracy=1.e-3

  if (action == write) then

     call hdf5_create_group(file_handle,'/Cosmology/')
     call hdf5_write_data(file_handle,'Cosmology/Omega0',Omega0)
     call hdf5_write_data(file_handle,'Cosmology/OmegaLambda',OmegaLambda)
     call hdf5_write_data(file_handle,'Cosmology/Hubble100',Hubble100)
     call hdf5_write_data(file_handle,'Cosmology/Sigma8',Sigma8)
     call hdf5_write_data(file_handle,'Cosmology/BoxSize',BoxSize)
     call hdf5_write_data(file_handle,'Cosmology/PowerSpectrum',PowerSpectrum)
     call hdf5_write_data(file_handle,'Cosmology/Mpart',Mpart)
     call hdf5_write_data(file_handle,'Cosmology/InterPartDist',InterPartDist)

  else if (action == read) then

     call hdf5_read_data(file_handle,'Cosmology/Omega0',Omega0)
     call hdf5_read_data(file_handle,'Cosmology/OmegaLambda',OmegaLambda)
     call hdf5_read_data(file_handle,'Cosmology/Hubble100',Hubble100)
     call hdf5_read_data(file_handle,'Cosmology/Sigma8',Sigma8)
     call hdf5_read_data(file_handle,'Cosmology/BoxSize',BoxSize)
     call hdf5_read_data(file_handle,'Cosmology/PowerSpectrum',PowerSpectrum)
     call hdf5_read_data(file_handle,'Cosmology/Mpart',Mpart)
     call hdf5_read_data(file_handle,'Cosmology/InterPartDist',InterPartDist)

  else if (action == test) then

     call hdf5_read_data(file_handle,'Cosmology/Omega0',     par1)
     call hdf5_read_data(file_handle,'Cosmology/OmegaLambda',par2)
     call hdf5_read_data(file_handle,'Cosmology/Hubble100',  par3)
     call hdf5_read_data(file_handle,'Cosmology/Sigma8',     par4)
     call hdf5_read_data(file_handle,'Cosmology/BoxSize',    par5)
     call hdf5_read_data(file_handle,'Cosmology/PowerSpectrum',par6)
     call hdf5_read_data(file_handle,'Cosmology/Mpart',      par7)
     call hdf5_read_data(file_handle,'Cosmology/InterPartDist',par8)
     
     diff = maxval( abs( &
          ((/par1,par2,par3,par4,par5(:),par6,par7,par8/)&
       -  (/Omega0,OmegaLambda,Hubble100,Sigma8,BoxSize(:),PowerSpectrum,Mpart,InterPartDist/))&
     / abs((/Omega0,OmegaLambda,Hubble100,Sigma8,BoxSize(:),PowerSpectrum,Mpart,InterPartDist/))))


     if (diff>accuracy) then
        call message('WARNING: SEVERE PROBLEM')
        call message('Inconsistent grid data in hdf5 file')
        write (info,'(''Parameters '',10(1x,e12.6))') par1,par2,par3,par4,par5(:),par6,par7,par8
        call message(trim(info))
        !
        call message('run parameters should be : ')
        write (info,'(''Parameters '',10(1x,e12.6))') Omega0,OmegaLambda,Hubble100,Sigma8,BoxSize(:)&
             ,PowerSpectrum,Mpart,InterPartDist
        call message(trim(info))
     endif

  else
 
    call abort_code(' invalid action in rw_cosmology ')

  endif
end subroutine rw_cosmology


subroutine rw_fmax(action)

! Run parameters

  use hdf5_wrapper; use pinocchio_hdf5
  use pinocchio_common
  use fmax_run
  implicit none
  character(*), intent(in) :: action
  character,parameter :: read='read',write='write',test='test'

  if (action .eq. write) then

     CALL hdf5_create_group(file_handle,'/Fmax/')
     call hdf5_write_data(file_handle,'Fmax/random_seed',random_seed)
     call hdf5_write_data(file_handle,'Fmax/saved_seed',saved_seed)
     call hdf5_write_data(file_handle,'Fmax/final_redshift',final_redshift)
     call hdf5_write_data(file_handle,'Fmax/limit_in_delta',limit_in_delta)
     call hdf5_write_data(file_handle,'Fmax/final_growing_mode',final_growing_mode)

  else if (action == read) then

     call hdf5_read_data(file_handle,'Fmax/random_seed',random_seed)
     call hdf5_read_data(file_handle,'Fmax/saved_seed',saved_seed)
     call hdf5_read_data(file_handle,'Fmax/final_redshift',final_redshift)
     call hdf5_read_data(file_handle,'Fmax/limit_in_delta',limit_in_delta)
     call hdf5_read_data(file_handle,'Fmax/final_growing_mode',final_growing_mode)

  else if (action == test) then
     return
  else
     call abort_code(' invalid action in rw_cosmology ')
  endif
end subroutine rw_fmax
#endif

character(len=4) function write_string(i)

  use numbers
  implicit none
  integer(kind=singleI),intent(in)::i

  write(write_string,'(i4)')i
  if (i<=9) then 
     write_string(1:3)='000'
  else if (i<=99) then
     write_string(1:2)='00'
  else if (i<=999) then
     write_string(1:1)='0'
  endif
  return
end function write_string
  
! ****************
! MPI I/O routines
! ****************

subroutine message(messg)
  use pinocchio_mpi
  implicit none
  !     
  character(*) :: messg
  
  if(proc_id .eq. 0) then 
     write (*,*) messg
     call flush(6)
  endif
end subroutine message
      
subroutine abort_code(messg)
  use pinocchio_mpi
  implicit none
  !    
  character(*) :: messg
  
  write (*,'(a)') messg
#ifdef MPI
  write (*,*) ' mpi_abort called by processor: ',proc_id
  call mpi_abort(mpi_comm_world,ierr)
  call mpi_finalize(' .... stop ')
#else
  write (*,*) ' abort_code called by processor: ',proc_id
#endif
  stop
end subroutine abort_code

subroutine finish(messg)
  use pinocchio_mpi
  implicit none
  !    
  character(*) :: messg
  
  call message(messg)

#ifdef MPI
  call mpi_finalize(ierr)
#endif
  stop
end subroutine finish



#undef READ_STATEMENT
#undef WRITE_STATEMENT
