!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


! This file contains the routines that drive differentiations and compute
! collapse times and velocities, plus the routine write_out
!
!
! SUBROUTINES AND FUNCTIONS IN THIS FILE
!
! - subroutine assign_fftw_variables(igrid) 
!     is called for initializing the grid variables used by compute_derivative(),
!     allocate_vector() and other allocation or i/o routines
! - subroutine compute_density(ismooth,igrid) 
!     is used to compute the density on the grid igrid from the potential (with FFT)
!     calls: assign_fftw_variables(igrid)             fmax-subs
!            allocate_vector()                        fmax-alloc
!            compute_derivative(ismooth,igrid,-1,-1)  fmax-subs
!            save_slice(igrid,2,0)                    fmax-tmpfiles
!            deallocate_vector()                      fmax-alloc
! - subroutine compute_first_derivatives(ismooth,igrid) 
!     computes the first derivatives of the potential
!     calls: assign_fftw_variables(igrid)             fmax-subs
!            allocate_vector()                        fmax-alloc
!            compute_derivative(ismooth,igrid,ia,0)   fmax-subs
!            save_slice(igrid,1,ia)                   fmax-tmpfiles
!            deallocate_vector()                      fmax-alloc
! - subroutine remove_first_derivatives() 
!     removes tmp files of first derivatives by writing a single number in them
!     calls: remove_slice                             fmax-tmpfiles
! - subroutine compute_second_derivatives(ismooth,igrid)
!     computes the second derivatives of the potential
!     calls: assign_fftw_variables(igrid)             fmax-subs
!            allocate_vector()                        fmax-alloc
!            compute_derivative(ismooth,igrid,ia,ib)  fmax-subs
!            save_slice(igrid,2,ider)                 fmax-tmpfiles
!            deallocate_vector()                      fmax-alloc
! - subroutine remove_second_derivatives()
!     removes tmp files of second derivatives by writing a single number in them
!     calls: remove_slice                             fmax-tmpfiles
! - subroutine compute_derivative(ismooth,igrid,first_derivative,second_derivative)
!     calls the fftw wrapper for the derivative. It reads the k-space
!     linear density from a file if necessary, and computes the variance
!     calls: assign_fftw_variables(igrid)             fmax-subs
!            get_kspace_vector(igrid)                 fmax-tmpfiles
!            real_derivative(first_d., second_d.)     fmax-fftw
! - subroutine compute_collapse_times(ismooth,ix_lb,iy_lb,iz_lb)
!     sums the deformation tensor over the boxes and computes the collapse time
!     for all particles [ix_lb, iy_lb and iz_lb is the position within the big box]
!     calls: get_boxes(ix_lb,iy_lb,iz_lb,2)           fmax-box
!            allocate_tmp_plane                       fmax-alloc
!            get_diff_plane(ngrid,iz,2)               fmax-tmpfiles
!            get_Fmax_plane(iz)                       fmax-tmpfiles
!            get_Rmax_plane(iz)                       fmax-tmpfiles
!            sum_over_boxes(ix,iy,iz,2)               fmax-box
!            function inverse_collapse_time(diff_ten,fail)   fmax-subs
!            save_Fmax_plane(iz)                      fmax-tmpfiles
!            save_Rmax_plane(iz)                      fmax-tmpfiles
!            deallocate_tmp_plane                     fmax-alloc
! - real(kind=doubleR) function inverse_collapse_time(def_ten,fail)
!     computes the inverse collapse time of a particle given the deformation tensor
!     calls: ord(x1,x2,x3)
!            function ell(l1,l2,l3,del,det)
! - real(kind=doubleR) function ell(l1,l2,l3,del,det)
!     computes the ellipsoidal collapse time
! - subroutine ord(x1,x2,x3)
!      puts three numbers in decreasing order
! - subroutine compute_velocities(ismooth,ix_lb,iy_lb,iz_lb)
!     sums the displacement tensor over the boxes and computes the Zel'dovich
!     velocities for all particles
!     [ix_lb, iy_lb and iz_lb is the position within the big box]
!     calls: get_boxes(ix_lb,iy_lb,iz_lb,1)           fmax-box
!            allocate_tmp_plane                       fmax-alloc
!            get_diff_plane(ngrid,iz,1)               fmax-tmpfiles
!            get_Rmax_plane(iz)                       fmax-tmpfiles
!            get_vmax_plane(iz)                       fmax-tmpfiles
!            sum_over_boxes(ix,iy,iz,1)               fmax-box
!            save_vmax_plane(iz)                      fmax-tmpfiles
!            deallocate_tmp_plane                     fmax-alloc
! - subroutine write_out()
!     writes the final results of fmax to output files
!     calls: assign_fftw_variables(ngrid)             fmax-subs
!            allocate_tmp_plane                       fmax-alloc
!            write_fmax_fields(filename,*, ngrid)     fmax-rw
!            deallocate_tmp_plane                     fmax-alloc
!            allocate_tmp_integer_plane               fmax-alloc
!            deallocate_tmp_integer_plane             fmax-alloc

subroutine assign_fftw_variables(igrid)

! Assigns the variables to be passed to the fftw wrapper

  use fftw_variables
  use fmax_grids
  use fmax_run
  implicit none
  integer(kind=singleI),intent(in)::igrid

  ThisGrid                      = igrid
  nx_fftw                       = Lgrid(igrid,1)
  ny_fftw                       = Lgrid(igrid,2)
  nz_fftw                       = Lgrid(igrid,3)
  total_local_size              = npart_fft(igrid)
  reverse_plan                  = saved_reverse_plan(igrid)
  forward_plan                  = saved_forward_plan(igrid)
  local_nz                      = Lgrid_local(igrid,3)
  local_z_start                 = Lgrid_z_start(igrid)
  local_ny_after_transform      = Lgrid_fft(igrid,2)
  local_y_start_after_transform = Lgrid_y_start(igrid)

  return
end subroutine assign_fftw_variables


subroutine compute_density(ismooth,igrid)

! Computes the density

  use fmax_run
  use fmax_grids
  use fmax_smoothing_radii
 
  implicit none
  integer(kind=singleI),intent(in)::ismooth,igrid

! allocates the vector for fft
  call assign_fftw_variables(igrid)
  call allocate_vector()

! the density is the laplacian of the potential
  call compute_derivative(ismooth,igrid,-1,-1)  ! (-1,-1) sets the greens function to be 1
  call save_slice(igrid,2,0)                    ! 0-th component of 2nd der. tensor = trace

  call deallocate_vector()

  return
end subroutine compute_density


subroutine compute_first_derivatives(ismooth,igrid)

! Computes the three components of the first derivative tensor
! of the peculiar rescaled gravitational potential;

  use fmax_run
  use fmax_grids
 
  implicit none
  integer(kind=singleI),intent(in)::ismooth,igrid
!
  character(len=24):: fdate
  external fdate

! local variables
  integer(kind=singleI)::ia
  character(len=80)::info

! allocates the vector for fft
  call assign_fftw_variables(igrid)
  call allocate_vector()

! computes all three first derivatives of the potential
! and saves them (the saving strategy is decided in optimized.F90)
  do ia=1,3

     write(info,'(''['',a24,''] Compute 1st derivative : '',i4)') fdate(),ia
     call message (trim(info))

     call compute_derivative(ismooth,igrid,ia,0)

     write(info,'(''['',a24,''] Save derivative : '',i4)') fdate(),ia
     call message (trim(info))

     call save_slice(igrid,1,ia)

  enddo

  call deallocate_vector()

  return
end subroutine compute_first_derivatives


subroutine remove_first_derivatives()

! Removes the tmp files for the second derivatives from the tmp memory

  use fmax_grids
  implicit none

  integer(kind=singleI)::igrid,ider

#ifndef IDLTEST
! for the IDL test tmp files are kept
  do igrid=1,ngrid
     do ider=1,3
        call remove_slice(igrid,1,ider)
     enddo
  enddo
#endif

  return
end subroutine remove_first_derivatives


subroutine compute_second_derivatives(ismooth,igrid)

! Computes the six components of the second derivative tensor
! of the peculiar rescaled gravitational potential;

  use pinocchio_mpi
  use fmax_run
  use fmax_grids

  implicit none
  integer(kind=singleI),intent(in)::ismooth,igrid
!
  character(len=24):: fdate
  external fdate

! local variables
  integer(kind=singleI)::ia,ib,ider
  character(len=80)::info

! allocates the vector for fft
  call assign_fftw_variables(igrid)
  call allocate_vector()

! computes all six second derivatives of the potential
! and saves them (the saving strategy is decided in optimized.F90)

  do ia=1,3
     do ib=ia,3
        if (ia==ib) then
           ider=ia
        else
           ider=ia+ib+1
        endif

        write(info,'(''['',a24,''] Compute 2nd derivative : '',i4)') fdate(),ider
        call message (trim(info))

        call compute_derivative(ismooth,igrid,ia,ib)

        write(info,'(''['',a24,''] Save derivative : '',i4)') fdate(),ider
        call message (trim(info))

        call save_slice(igrid,2,ider)

     enddo
  enddo
  call deallocate_vector()

  return
end subroutine compute_second_derivatives


subroutine remove_second_derivatives()

! Removes the tmp files for the second derivatives from the tmp memory

  use fmax_grids
  implicit none

  integer(kind=singleI)::igrid,ider

#ifndef IDLTEST
! for the IDL test tmp files are kept
  do igrid=1,ngrid
     do ider=1,6
        call remove_slice(igrid,2,ider)
     enddo
  enddo
#endif

  return
end subroutine remove_second_derivatives


subroutine compute_derivative(ismooth,igrid,first_derivative,second_derivative)

! This code computes a derivative of the (peculiar rescaled) 
! gravitational potential
! The field to transform will be put in vector contained in module fftw_variables
! first_derivative and second_derivative are integers used in the Green function
! they can assume the following values:
!   0 = no derivative
!   1 = differentiate over x-direction
!   2 = differentiate over y-direction
!   3 = differentiate over z-direction

! the field is treated as a vector in k-space, and is sliced in the y-component

  use pinocchio_constants
  use pinocchio_common
  use fmax_grids
  use fmax_run
  use fmax_smoothing_radii
  use fftw_variables
  use pinocchio_mpi
  use fmax_real_derivative
  implicit none
  integer(kind=singleI),intent(in)::ismooth,igrid,first_derivative,second_derivative
  character(len=24):: fdate
  external fdate

! local variables
!  real(kind=doubleR):: local_variance,global_variance,local_average,global_average
!  real(kind=doubleR)         :: random1
!  real(kind=doubleR),external:: random
  integer(kind=singleI)              :: ok
  character(len=80)    :: info
  character(len=1):: grid_label

  call assign_fftw_variables(igrid)

#ifdef MEMTEST
  if(first_derivative /= 1 .or. second_derivative /= 1) then
     call message(' memtest only does d^2/dx/dx ')
     return
  endif
#endif
  if(read_IC(igrid)) then

! Reads the FFT transformed grid from the tmp file
     write(info,'(''['',a24,''] read k-space vector from file '')') fdate()
     call message (trim(info))
     call get_kspace_vector(igrid)

!     local_variance = sum(vector**2)
!     local_average  = sum(vector)
!#ifdef MPI
!     CALL MPI_REDUCE(local_variance, global_variance, 1, MPI_DOUBLE_PRECISION, MPI_SUM, 0, mpi_comm_world, ierr)
!     CALL MPI_REDUCE(local_average , global_average , 1, MPI_DOUBLE_PRECISION, MPI_SUM, 0, mpi_comm_world, ierr)
!#else
!     global_variance = local_variance
!     global_average  = local_average
!#endif
!     if (proc_id == 0) then
!        global_variance = global_variance / dble(npart(ngrid))
!        global_average  = global_average  / dble(npart(ngrid))
!        write (*,*) 'average and rms: ',global_average, global_variance
!     endif

     generate_field = .false.
  else
! intialization of random seed
!     random_seed = -saved_seed + proc_id*23 ! ?????
!     random1     = random(random_seed)
     generate_field = .true.
  endif

  Rsm_grid  = Smoothing_radii(ismooth)/InterPartDistGrids(igrid)
  cell_size = InterPartDistGrids(igrid)

#ifdef IDLTEST
  lower_k_cutoff = 0.d0
  upper_k_cutoff = 0.8 * pi
#else
  if (igrid > 1) then
     lower_k_cutoff = k_cutoff(igrid-1)
  else
     lower_k_cutoff = 0.d0
  endif

  if (igrid < ngrid) then
     upper_k_cutoff = k_cutoff(igrid)
  else
     upper_k_cutoff = 0.8 * pi
  endif
#endif

  write(info,'(''['',a24,''] starting real_derivative '')') fdate()
  call message (trim(info))
  call real_derivative(first_derivative, second_derivative)

  return
end subroutine compute_derivative


subroutine compute_collapse_times(ismooth,ix_lb,iy_lb,iz_lb)

! Computes the collapse times for one smoothing radius and position
! by collecting information from all grids
! and then computing the collapse times in planes

  use numbers
  use pinocchio_mpi
  use fmax_box_sections
  use fmax_grids
  use fmax_one_plane
  use fmax_run
  use fmax_smoothing_radii
  implicit none
  integer(kind=singleI),intent(in)::ismooth,ix_lb,iy_lb,iz_lb

  real(kind=doubleR),external::inverse_collapse_time

! local variables
  integer(kind=singleI)::local_z,global_x,global_y,global_z,fail
  real(kind=doubleR):: Fnew
  real(kind=doubleR):: local_variance,global_variance,local_average,global_average,delta

!!!! collects information from all the other grids
!!!  call get_boxes(ix_lb,iy_lb,iz_lb,2)

! allocates tmp plane
  call allocate_tmp_plane(ngrid)

! initialize the variance of the slice
  local_variance = 0.d0
  local_average  = 0.d0

! loop on planes
  do local_z=0,Lgrid_local(ngrid,3)-1
     global_z=local_z+Lgrid_z_start(ngrid)

! reads the differentials for the plane
     call get_diff_plane(ngrid,global_z,2)

! reads Fmax if ismooth is not one
     if (ismooth > 1) then
        call get_Fmax_plane(global_z)
        call get_Rmax_plane(global_z)
     else
        Fmax_plane = -10.d0
        Rmax_plane = 0
     endif

! loop on the plane
     do global_y=0,Lgrid_local(ngrid,2)-1
        do global_x=0,Lgrid_local(ngrid,1)-1

! sum all contributions
           diff_ten(1:6) = differentials_plane(:,global_x,global_y)


!!!#ifndef IDLTEST
!!!           call sum_over_boxes(global_x,global_y,global_z,2)
!!!#endif

! Computation of the variance of the linear density field
           delta = sum(diff_ten(1:3))
           local_average  = local_average  + delta
           local_variance = local_variance + delta**2

#ifdef IDLTEST
! Here Fmax is just the smoothed density
           Fnew = delta
           Fmax_plane(global_x,global_y) = Fnew
           Rmax_plane(global_x,global_y) = 0
#else
! Computation of ellipsoidal collapse
           if (delta > limit_in_delta) then
              Fnew = inverse_collapse_time(diff_ten, fail)

              if (fail==1) &
                 call message('Warning: failure in inverse_collapse_time')

              if (Fmax_plane(global_x,global_y) < Fnew) then
                 Fmax_plane(global_x,global_y) = Fnew
                 Rmax_plane(global_x,global_y) = ismooth
              endif
           endif
#endif

! end of loop on the plane
        enddo
     enddo

! writes the result on disc if relevan
     call save_Fmax_plane(global_z)
     call save_Rmax_plane(global_z)

! end of loop on planes
  enddo

  call deallocate_tmp_plane()

! Saves the obtained variance
  global_variance = 0.d0
  global_average  = 0.d0
#ifdef MPI
  CALL MPI_REDUCE(local_variance, global_variance, 1, MPI_DOUBLE_PRECISION, MPI_SUM, 0, mpi_comm_world, ierr)
  CALL MPI_REDUCE(local_average , global_average , 1, MPI_DOUBLE_PRECISION, MPI_SUM, 0, mpi_comm_world, ierr)
#else
  global_variance = local_variance
  global_average  = local_average
#endif
  if (proc_id == 0) then
     global_variance = global_variance / dble(npart(ngrid))
     global_average  = global_average  / dble(npart(ngrid))
     True_variances(ismooth) = global_variance
     global_variance = sqrt(global_variance)
     write (*,*) 'average and rms of the smoothed density: ',global_average, global_variance
  endif

  call message ('collapse times done')

! bye!
  return
end subroutine compute_collapse_times


function inverse_collapse_time(deformation_tensor, fail)

! Computes the inverse collapse time of a grid point
! given its deformation tensor
  use numbers
  use pinocchio_constants

  implicit none
  real(kind=doubleR) :: inverse_collapse_time
  real(kind=doubleR),intent(in)::deformation_tensor(6)
  integer(kind=singleI),intent(out)::fail

! local variables
  integer(kind=singleI)::m
  real(kind=doubleR)::mu1,mu2,mu3,c,q,r,sq,t,x1,x2,x3
  real(kind=doubleR),external::ell

!  print *,'chiamata inverse_collapse_time ',deformation_tensor

! mu1, mu2 and mu3 are the principal invariants of the 3x3 tensor 
! of second derivatives.


! diagonalization of the tensor
  mu1=sum(deformation_tensor(1:3))
  mu2=.5*mu1*mu1
  do m=1,6
     if (m.le.3) then 
        c=.5
     else
        c=1.
     endif
     mu2=mu2-c*deformation_tensor(m)*deformation_tensor(m)
  enddo
  mu3=deformation_tensor(1)*deformation_tensor(2)*deformation_tensor(3) &
    + 2.*deformation_tensor(4)*deformation_tensor(5)*deformation_tensor(6) &
       - deformation_tensor(1)*deformation_tensor(6)*deformation_tensor(6) &
       - deformation_tensor(2)*deformation_tensor(5)*deformation_tensor(5) &
       - deformation_tensor(3)*deformation_tensor(4)*deformation_tensor(4)

  q=(mu1*mu1-3.*mu2)/9.

  if (q==0.) then
! in this case the tensor is already diagonal
     x1=deformation_tensor(1)
     x2=deformation_tensor(2)
     x3=deformation_tensor(3)
  else
     r=-(2.*mu1**3-9.*mu1*mu2+27.*mu3)/54.

     if (q*q*q.lt.r*r.or.q.lt.0.) then
        fail=1
        inverse_collapse_time=-10.d0
        return
     endif

     if(q < 0)then
        write (*,*) ' error: q<0'
        call abort_code(' q < 0 in inverse_collapse_times')
     endif
     sq=sqrt(q)     
     t=acos(r/q/sq)
     x1=-2.*sq*cos(t/3.)+mu1/3.
     x2=-2.*sq*cos((t+2.*pi)/3.)+mu1/3.
     x3=-2.*sq*cos((t+4.*pi)/3.)+mu1/3.

  endif

  call ord(x1,x2,x3)
  t=ell(x1,x2,x3,mu1,mu3)
     
! inverse collapse time
     
  if (t>0.) then
     inverse_collapse_time = 1./t
  else
     inverse_collapse_time = -10.
  endif

  fail=0
  return
end function inverse_collapse_time


function ell(l1,l2,l3,del,det)
  use numbers
  use pinocchio_constants
  implicit none
  real(kind=doubleR) :: ell
  real(kind=doubleR) :: l1,l2,l3,del,det

! This rouine computes the smallest non-negative solution of the 3rd
! order equation for the ellipsoid, and corrects it to reproduce the
! spherical collapse correctly.

! local variables
  real(kind=doubleR) :: a1,a2,a3,den,dis,r,q,sq,t,s1,s2,s3,corr

  den=det/126.+5.*l1*del*(del-l1)/84.
  if (l1.eq.0.) then
     ell=-.1
  else if (den.eq.0.) then
     if (del.eq.l1) then
        if (l1.gt.0) then
           ell=1./l1
        else
           ell=-.1
        endif
     else
        dis=7.*l1*(l1+6.*del)
        if (dis.lt.0.) then
           ell=-.1
        else
           ell=(7.*l1-sqrt(dis))/(3.*l1*(l1-del))
           if (ell.lt.0.) ell=-.1
        endif
     endif
  else
     a1=3.*l1*(del-l1)/14./den
     a2=l1/den
     a3=-1./den
     q=(a1*a1-3.*a2)/9.
     r=(2.*a1*a1*a1-9.*a1*a2+27.*a3)/54.
     if (r*r.gt.q*q*q) then
        sq=(sqrt(r*r-q*q*q)+abs(r))**(1./3.)
        ell=-abs(r)/r*(sq+q/sq)-a1/3.
        if (ell.lt.0.) ell=-.1
     else
        sq=sqrt(q)
        t=acos(r/q/sq)
        s1=-2.*sq*cos(t/3.)-a1/3.
        s2=-2.*sq*cos((t+2.*pi)/3.)-a1/3.
        s3=-2.*sq*cos((t+4.*pi)/3.)-a1/3.
        if (s1.lt.0.) s1=1.d10
        if (s2.lt.0.) s2=1.d10
        if (s3.lt.0.) s3=1.d10
        ell=min(s1,s2,s3)
        if (ell.eq.1.d10) ell=-.1
     endif
  endif
  if (del.gt.0..and.ell.gt.0.) then
     corr=-.364/del*exp(-6.5*(l1-l2)/del-2.8*(l2-l3)/del)
  else
     corr=0.
  endif
  ell=ell+corr
  return

end function ell


subroutine ord(a,b,c)

! orders a,b,c in decreasing order a>b>c
  use numbers
  implicit none
  real(kind=doubleR),intent(inout) :: a,b,c
  real(kind=doubleR)::a1,b1,c1
  
  a1=max(a,b,c)
  c1=min(a,b,c)
  b1=a+b+c-a1-c1
  a=a1
  b=b1
  c=c1
  return

end subroutine ord


subroutine compute_velocities(ismooth,ix_lb,iy_lb,iz_lb)

! Computes the collapse times for one smoothing radius and position
! by collecting information from all grids
! and then computing the collapse times in planes

  use fmax_box_sections
  use fmax_grids
  use fmax_one_plane
  use fmax_run
  use fmax_smoothing_radii

  implicit none
  integer(kind=singleI),intent(in)::ismooth,ix_lb,iy_lb,iz_lb

! local variables
  integer(kind=singleI)::local_z,global_x,global_y,global_z

!!!! collects information from all the other grids
!!!  call get_boxes(ix_lb,iy_lb,iz_lb,1)

! allocates tmp plane
  call allocate_tmp_plane(ngrid)

! loop on planes
  do local_z=0,Lgrid_local(ngrid,3)-1
     global_z=local_z+Lgrid_z_start(ngrid)

! reads the differentials for the plane
     call get_diff_plane(ngrid,global_z,1)

! reads the Rmax matrix, and the vmax vector if ismooth > 1
     call get_Rmax_plane(global_z)
     if (ismooth > 1) then
        call get_vmax_plane(global_z)
     else
        vmax_plane = 0.d0
     endif

! loop on the plane
     do global_y=0,Lgrid_local(ngrid,2)-1
        do global_x=0,Lgrid_local(ngrid,1)-1

           if (Rmax_plane(global_x,global_y) == ismooth) then
! sum all contributions
              diff_ten(1:3) = differentials_plane(:,global_x,global_y)
!!!#ifndef IDLTEST
!!!              call sum_over_boxes(global_x,global_y,global_z,1)
!!!#endif
              vmax_plane(:,global_x,global_y) = diff_ten(1:3)

           endif

! end of loop on the plane
        enddo
     enddo

! writes the result on disc
     call save_vmax_plane(global_z)

! end of loop on planes
  enddo

  call deallocate_tmp_plane()
  call message ('compute_velocities done')

! bye!
  return
end subroutine compute_velocities


subroutine write_out()

! output of the Fmax and vmax matrices

  use pinocchio_common
  use pinocchio_mpi
  use fmax_smoothing_radii
  use fmax_grids
  use fmax_one_plane
  use fmax_run
  implicit none

! local variables
  character(len=80):: filename
  integer(kind=singleI)::ismooth,igrid
  character(len=1):: label

! allocation of tmp plane
!  call assign_fftw_variables(ngrid)
  call allocate_tmp_plane(ngrid)

! Writes out Fmax and Vel* files
  filename='Data/Fmax.'//run//'.d'
  call write_fmax_fields(filename,0, ngrid)
  filename='Data/Velx.'//run//'.d'
  call write_fmax_fields(filename,1, ngrid)
  filename='Data/Vely.'//run//'.d'
  call write_fmax_fields(filename,2, ngrid)
  filename='Data/Velz.'//run//'.d'
  call write_fmax_fields(filename,3, ngrid)

! Writes out Rmax only if requested
  if (output_rmax) then
     call allocate_tmp_integer_plane(ngrid)
     filename='Data/Rmax.'//run//'.d'
     call write_fmax_fields(filename,4, ngrid)
     call deallocate_tmp_integer_plane()
  endif

! deallocate tmp plane
  call deallocate_tmp_plane()

! Writes the density if generated
  do igrid = 1, ngrid
     if (.not.read_IC(igrid)) then
        write(label,'(i1)')igrid
        filename='Data/Density_g'//label//'.d'

!        call assign_fftw_variables(igrid)
        call allocate_tmp_plane(igrid)
        call write_fmax_fields(filename, 5, igrid)
        call deallocate_tmp_plane()

     endif
  enddo

  if (proc_id == 0) then
     open(10,file=run//'.var',status='unknown')
     do ismooth = 1, nsmooth
        write(10,'(3(1x,f15.8))') Smoothing_Radii(ismooth) &
             , Smoothing_Variances(ismooth) &
             , True_Variances(ismooth)
     enddo
     close(10)
     write (*,*) 'Written file '//run//'.var'
  endif

  return
end subroutine write_out


subroutine compute_only_variance(ismooth,ix_lb,iy_lb,iz_lb)

! Computes the collapse times for one smoothing radius and position
! by collecting information from all grids
! and then computing the collapse times in planes

  use numbers
  use pinocchio_mpi
  use fmax_box_sections
  use fmax_grids
  use fmax_one_plane
  use fmax_run
  use fmax_smoothing_radii
  use tmp_plane
  implicit none
  integer(kind=singleI),intent(in)::ismooth,ix_lb,iy_lb,iz_lb

  real(kind=doubleR),external::inverse_collapse_time

! local variables
  integer(kind=singleI)::local_z,global_x,global_y,global_z,fail
  real(kind=doubleR):: Fnew
  real(kind=doubleR):: local_variance,global_variance,local_average,global_average,delta

! allocates tmp plane
  call allocate_tmp_plane(ngrid)

! initialize the variance of the slice
  local_variance = 0.d0
  local_average  = 0.d0

! loop on planes
  do local_z=0,Lgrid_local(ngrid,3)-1
     global_z=local_z+Lgrid_z_start(ngrid)

! reads the differentials for the plane
     call get_component_plane(ngrid,global_z,2,0)

     local_average  = local_average  + sum(plane)
     local_variance = local_variance + sum(plane**2)

! end of loop on planes
  enddo

  call deallocate_tmp_plane()

! Saves the obtained variance
  global_variance = 0.d0
  global_average  = 0.d0
#ifdef MPI
  CALL MPI_REDUCE(local_variance, global_variance, 1, MPI_DOUBLE_PRECISION, MPI_SUM, 0, mpi_comm_world, ierr)
  CALL MPI_REDUCE(local_average , global_average , 1, MPI_DOUBLE_PRECISION, MPI_SUM, 0, mpi_comm_world, ierr)
#else
  global_variance = local_variance
  global_average  = local_average
#endif
  if (proc_id == 0) then
     global_variance = global_variance / dble(npart(ngrid))
     global_average  = global_average  / dble(npart(ngrid))
     True_variances(ismooth) = global_variance
     global_variance = sqrt(global_variance)
     write (*,*) 'average and rms of the smoothed density: ',global_average, global_variance
  endif

  call message ('variance done')

! bye!
  return
end subroutine compute_only_variance
