!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

! This file contains the routines that handles the swapping of matrices
! to and from the tmp files
!
! NB: tmp files are written and read through the plane array;
!     a single plane is contained in each file;
!     exceptions is the double precision vector in the subroutines
!     write_ and read_kspace_vector, who is written and read as a whole (not
!     plane by plane)
! 
! DIRECTORIES
!
! tmp planes relative to igrid<ngrid go to common_directory
! tmp planes of ngrid (differentials, Fmax, Rmax, vmax) go to local_directory
! k-space density vectors go to local_directory
!
!
! SUBROUTINES AND FUNCTIONS IN THIS FILE
!
! - subroutine make_filename(ig,id,ic,ip,filename,unit,scratch_flag)
!     gives filename and unit number, according to the saving stragegy adopted
!     calls: write_string(ip)                         fmax-tmpfiles
! - subroutine open_tmp_files()
!     opens all the necessary tmp scratch files
!     calls: make_filename                            fmax-tmpfiles
! - subroutine close_tmp_files()
!     closes all the tmp scratch files
!     calls: make_filename                            fmax-tmpfiles
! - subroutine save_slice(igrid, derivative, component)
!     saves a real-space slice into tmp files
!     calls: vector_to_plane(iz)                      fmax-tmpfiles
!            write_tmp_plane(filename,unit,flag)      fmax-tmpfiles
! - subroutine vector_to_plane(local_z)
!     extracts a plane from the real-space fftw vector
!     calls: index_rspace(...)                        fmax-fftw
! - subroutine plane_to_vector(local_z)
!     puts a plane into the real-space fftw vector
!     calls: index_rspace(...)                        fmax-fftw
! - subroutine read_tmp_plane(filename,unit,flag)
!     read a real-space plane from a tmp file
! - subroutine write_tmp_plane(filename,unit,flag)
!     write a real-space plane to a tmp file
! - subroutine remove_tmp_plane(filename,unit,flag)
!     over-write a tmp file
! - subroutine read_tmp_integer_plane(filename,unit,flag)
!     read an integer real-space plane from a tmp file
! - subroutine write_tmp_integer_plane(filename,unit,flag)
!     write an integer real-space plane to a tmp file
! - subroutine remove_slice(igrid, derivative, component)
!     removes the tmp files of a slice
!     calls: remove_tmp_plane(filename,unit,flag)     fmax-tmpfiles
! - subroutine get_diff_plane(igrid, iplane, derivative)
!     get the first or second derivatives of a plane
!     calls: get_component_plane(igrid,iplane,derivative,icomp) fmax-tmpfiles
! - subroutine get_component_plane(igrid, iplane, derivative, component)
!     get a plane of a component of a derivative
!     calls: make_filename                            fmax-tmpfiles
! - subroutine get_Fmax_plane(iplane)
!     reads one Fmax plane
!     calls: read_tmp_plane(filename,unit,flag)       fmax-tmpfiles
!            make_filename                            fmax-tmpfiles
! - subroutine save_Fmax_plane(iplane)
!     updates one Fmax plane
!     calls: write_tmp_plane(filename,unit,flag)      fmax-tmpfiles
!            make_filename                            fmax-tmpfiles
! - subroutine get_Rmax_plane(iplane)
!     reads one Rmax plane
!     calls: read_tmp_integer_plane(filename,unit,flag)    fmax-tmpfiles
!            make_filename                            fmax-tmpfiles
! - subroutine save_Rmax_plane(iplane)
!     updates one Rmax plane
!     calls: write_tmp_integer_plane(filename,unit,flag)   fmax-tmpfiles
!            make_filename                            fmax-tmpfiles
! - subroutine get_vmax_plane(iplane)
!     reads one vmax plane
!     calls: read_tmp_plane(filename,unit,flag)       fmax-tmpfiles
!            make_filename                            fmax-tmpfiles
! - subroutine save_vmax_plane(iplane)
!     update one vmax plane
!     calls: write_tmp_plane(filename,unit,flag)      fmax-tmpfiles
!            make_filename                            fmax-tmpfiles
! - subroutine save_kspace_vector(igrid)
!     writes a transformed density slice into a tmp file
!     calls: make_filename                            fmax-tmpfiles
! - subroutine get_kspace_vector(igrid)
!     gets a transformed density slice from a tmp file
!     calls: make_filename                            fmax-tmpfiles


subroutine make_filename(ig,id,ic,ip,filename,unit,scratch_flag)

! gives filename and unit number, according to the saving stragegy adopted
! NB: ig runs from 1 to ngrid
!     id runs from 0 to 2
!     ic runs from 0 to 6
!     ip runs from 0 to Lgrid(ig,3)-1
!
!     if ig<=1 then
!                  ig== 0 -> Fmax plane            6
!                  ig==-1 -> vmax-x plane          7
!                  ig==-2 -> vmax-y plane          8
!                  ig==-3 -> vmax-z plane          9
!                  ig==-4 -> Rmax plane           10
!     if ip<0 -> k-space linear density slice     11-15

  use numbers
  use pinocchio_mpi
  use fmax_grids
  use fmax_run
  use save_directory
  implicit none
  integer(kind=singleI),intent(in) ::ig,id,ic,ip
  character(len=80)    ,intent(out)::filename
  integer(kind=singleI),intent(out)::unit
  logical              ,intent(out)::scratch_flag
  character(len=4)     ,external   ::write_string

! local variables
  character(len=4)::string
  integer(kind=singleI)::maxgrid
  integer, parameter :: unit_off=100
  maxgrid=maxval(Lgrid(:,3))


! consistency check on components
  if (ig>ngrid .or. ig<-4 .or. id<0 .or. id>2 .or. ic<0 .or. ic>6 &
       .or. ip<-nproc .or. ip>maxgrid) then
     print *,ig,id,ic,ip
     call abort_code ('Severe: error in make_filename!!!')
  endif

  if (ig>0) then
! differentials
     if (ip>=0) then
        if (use_scratch_for_derivatives .and. ig==ngrid) then
           unit = ((ig-1)*18+id*6+(ic-1))*maxgrid+ip+unit_off
           filename = ''
           scratch_flag = .true.
        else
           unit=10
           string = write_string(ip)
           write(filename,'(''save_g'',i1,''_d'',i1,''_c'',i1,''_p'')') ig,id,ic
           filename = trim(filename)//string//'.d'
           if (ig==ngrid) then
              filename = local_directory//filename
           else
              filename = common_directory//filename
           endif
           scratch_flag=.false.
        endif
     else
! k-space linear density slice
        if (use_scratch_for_derivatives) then
           unit = ((ig+10)*18+id*6+(ic-1))*maxgrid-ip+unit_off
           filename = ''
           scratch_flag = .true.
        else
           unit=10
           string = write_string(-ip-1)
           write(filename,'(''density_g'',i1,''_proc'')') ig
           filename = trim(filename)//string//'.d'
           filename = local_directory//filename
           scratch_flag=.false.
        endif
     endif
! Fmax, vmax and Rmax files
  else
     if (use_scratch_for_all) then
        unit = ((-ig+6)*18+id*6+(ic-1))*maxgrid+ip+unit_off
        filename = ''
        scratch_flag = .true.
     else
        unit=10
        string = write_string(ip)
        select case (ig)
        case (0)
           filename = 'save_fmax_p'//string//'.d'
        case (-1)
           filename = 'save_vmax_c1_p'//string//'.d'
        case (-2)
           filename = 'save_vmax_c2_p'//string//'.d'
        case (-3)
           filename = 'save_vmax_c3_p'//string//'.d'
        case (-4)
           filename = 'save_rmax_p'//string//'.d'
        end select

        filename = local_directory//filename
        scratch_flag=.false.
     endif
  endif

  return
end subroutine make_filename


subroutine open_tmp_files()

! opens all the necessary tmp scratch files

  use fmax_grids
  use pinocchio_mpi
  use fmax_run
  implicit none
  
! local variables
  integer(kind=singleI)::ig,id,ic,ip,ic1,ic2,iz,unit,io_unit
  character(len=80)::filename
  logical::scratch_flag
  logical,save    ::first_call=.true.
  character(len=80) :: file_info


  io_unit = 10+proc_id
  if (scratch_flag) then
     write (file_info,'(''myunit.'',i4.4)') proc_id
     if(first_call)then
        open(unit=io_unit,file=file_info,status='unknown')
        first_call = .false.
     else
!     open(unit=io_unit,file=file_info,status='old',access='append')
        open(unit=io_unit,file=file_info,status='old',position='append')
     endif
  endif

! nothing to do if scratch files are not required
  if (.not.(use_scratch_for_derivatives .or. use_scratch_for_all)) return

  do ig=1,ngrid
     do id=0,2
        select case (id)
        case (0)
           ic1=1 ; ic2=1
        case (1)
           ic1=1 ; ic2=3
        case (2)
           ic1=0 ; ic2=6
        end select
        do ic=ic1,ic2
           do iz=0,Lgrid_local(ig,3)-1
              ip = iz + Lgrid_z_start(ig)

              call make_filename(ig,id,ic,ip,filename,unit,scratch_flag)
              if (scratch_flag) then
                 open(unit,status='scratch',form='unformatted')
                 write (io_unit,'('' unit = '',i5)') unit

!                 print *,scratch_flag,unit,trim(filename)
              endif

           enddo
        enddo
     enddo
  enddo

  do ig=-4,0
     do iz=0,Lgrid_local(ngrid,3)-1
        ip = iz + Lgrid_z_start(ngrid)

        call make_filename(ig,0,0,ip,filename,unit,scratch_flag)
        if (scratch_flag) then
           open(unit,status='scratch',form='unformatted')
           write (io_unit,'('' unit = '',i5)') unit
!           print *,scratch_flag,unit,trim(filename)
        endif

     enddo
  enddo

  do ig=1,ngrid
     call make_filename(ig,0,0,-(1+proc_id),filename,unit,scratch_flag)
     if (scratch_flag) then
        open(unit,status='scratch',form='unformatted')
        write (io_unit,'('' unit = '',i5)') unit
!        print *,scratch_flag,unit,trim(filename)
     endif
  enddo
  close(io_unit)
  return
end subroutine open_tmp_files


subroutine close_tmp_files()

! opens all the necessary tmp scratch files

  use fmax_grids
  use pinocchio_mpi
  use fmax_run
  implicit none
  
! local variables
  integer(kind=singleI)::ig,id,ic,ip,ic1,ic2,iz,unit
  character(len=80)::filename
  logical::scratch_flag

! nothing to do if scratch files are not required
  if (.not.(use_scratch_for_derivatives .or. use_scratch_for_all)) return

  do ig=1,ngrid
     do id=0,2
        select case (id)
        case (0)
           ic1=1 ; ic2=1
        case (1)
           ic1=1 ; ic2=3
        case (2)
           ic1=0 ; ic2=6
        end select
        do ic=ic1,ic2
           do iz=0,Lgrid_local(ig,3)-1
              ip = iz + Lgrid_z_start(ig)
              call make_filename(ig,id,ic,ip,filename,unit,scratch_flag)
              if (scratch_flag) close(unit)
           enddo
        enddo
     enddo
  enddo

  do ig=-4,0
     do iz=0,Lgrid_local(ngrid,3)-1
        ip = iz + Lgrid_z_start(ngrid)
        call make_filename(ig,0,0,ip,filename,unit,scratch_flag)
        if (scratch_flag) close(unit)
     enddo
  enddo

  do ig=1,ngrid
     call make_filename(ig,0,0,-(1+proc_id),filename,unit,scratch_flag)
     if (scratch_flag) close(unit)
  enddo

  return
end subroutine close_tmp_files


subroutine save_slice(igrid, derivative, component)

! Saves a real-space slice into tmp files

  use fmax_grids
  use save_directory
  use pinocchio_mpi
  implicit none
  integer(kind=singleI),intent(in)::igrid,derivative,component
! local variables
  integer(kind=singleI)::global_z,ix,iy,iz,unit
  integer(kind=doubleI):: index
  character(len=80)::filename
  integer::iproc
  logical::scratch_flag

! loop over local planes
  do iproc=0,nproc-1
     if(iproc .eq. proc_id) then
        do iz=0,Lgrid_local(igrid,3)-1
           global_z = iz + Lgrid_z_start(igrid)
           
! extract the plane from the vector
           call vector_to_plane(iz)
           
! construct filename for saving this plane
           call make_filename(igrid,derivative,component,global_z,filename,unit,scratch_flag)

! writes the plane on a local file
           call write_tmp_plane(filename,unit,scratch_flag)
        enddo
     endif
!!$#ifdef MPI
!!$! MPI barrier
!!$     call mpi_barrier(mpi_comm_world,ierr)
!!$#endif
  enddo
  
  return
end subroutine save_slice


subroutine vector_to_plane(local_z)

! extracts a plane from the real-space fftw vector

  use numbers
  use fftw_variables
  use tmp_plane
  implicit none
  integer(kind=singleI), intent(in) :: local_z
! local variables
  integer(kind=singleI) :: local_x, local_y
  integer(kind=doubleI) :: index
  integer(kind=doubleI), external :: index_rspace

  do local_y=0, ny_fftw-1
     do local_x=0, nx_fftw-1
        index        = index_rspace(local_x, local_y, local_z)
        plane(local_x+1,local_y+1) = vector(index)
     enddo
  enddo

  return
end subroutine vector_to_plane


subroutine plane_to_vector(local_z)

! puts a plane into the real-space fftw vector

  use tmp_plane
  use fftw_variables
  implicit none
  integer(kind=singleI), intent(in) :: local_z
! local variables
  integer(kind=singleI) :: local_x, local_y
  integer(kind=doubleI) :: index
  integer(kind=doubleI), external :: index_rspace

  do local_y=0, ny_fftw-1
     do local_x=0, nx_fftw-1
        index         = index_rspace(local_x,local_y,local_z)
        vector(index) = plane(local_x+1,local_y+1)
     enddo
  enddo

  return
end subroutine plane_to_vector


subroutine read_tmp_plane(filename,unit,scratch_flag)

! read a real-space plane from a tmp file

!  use fftw_variables, ONLY: nx_fftw, ny_fftw
  use fmax_grids, ONLY : ngrid
  use tmp_plane
  implicit none
  integer(kind=singleI),intent(in)::unit
  character(len=*)     ,intent(in)::filename
  logical              ,intent(in)::scratch_flag

! local variables
  integer(kind=singleI)::L1,L2
  character(len=9)::label

  if (scratch_flag) then
     rewind(unit)
  else
     open (unit,file=filename,status='unknown',form='unformatted',err=10)
  endif

  read (unit,err=10)L1,L2
  if (L1/=size(plane,1) .or. L2/=size(plane,2)) then
     call message ('file: '//trim(filename))
     call abort_code ('Inconsistent grid lengths in read_tmp_plane') 
  endif
  read (unit,err=10) plane

  if (.not.scratch_flag) close(unit)

  return
10 write(label,'(i9)')unit
  call abort_code('Error in read_tmp_plane, unit '//label//', filename '//trim(filename))
end subroutine read_tmp_plane


subroutine write_tmp_plane(filename,unit,scratch_flag)

! write a real-space plane to a tmp file

!  use fftw_variables, ONLY: nx_fftw, ny_fftw
  use fmax_grids, ONLY : ngrid
  use tmp_plane
  implicit none
  integer(kind=singleI),intent(in)::unit
  character(len=*)     ,intent(in)::filename
  logical              ,intent(in)::scratch_flag

! local variables
  integer(kind=singleI)::L1,L2
  character(len=19)::label

  L1 = size(plane,1) ; L2 = size(plane,2)

! write single plane to a file
  if (scratch_flag) then
     rewind(unit)
  else
     open (unit,file=filename,status='unknown',form='unformatted',err=10)
  endif

  write(unit,err=10) L1,L2
  write(unit,err=12) plane

  if (.not.scratch_flag) close(unit)

  return
10 write(*,*) 'error writing l1,l2 ',l1,l2
12 write(label,'(i9,1x,i4,1x,i4)')unit,L1,L2
  call abort_code('Error in write_tmp_plane, unit, L1,L2 '//label//', filename '//trim(filename))
end subroutine write_tmp_plane


subroutine remove_tmp_plane(filename,unit,scratch_flag)

! removes a tmp plane by over-writing on the file

  use numbers
  implicit none
  integer(kind=singleI),intent(in)::unit
  character(len=*)     ,intent(in)::filename
  logical              ,intent(in)::scratch_flag

! local variables
  logical::exist_flag


  if (scratch_flag) then
     rewind(unit)
  else
     inquire(file=filename,exist=exist_flag)
     if (exist_flag) then
        open (unit,file=filename,status='unknown',form='unformatted')
     else
        return
     endif
  endif

  write(unit)0

  if (.not.scratch_flag) close(unit)

  return
end subroutine remove_tmp_plane


subroutine read_tmp_integer_plane(filename,unit,scratch_flag)

! read an integer real-space plane from a tmp file

!  use fftw_variables, ONLY: nx_fftw, ny_fftw
  use fmax_grids, ONLY : ngrid
  use tmp_plane
  implicit none
  integer(kind=singleI),intent(in)::unit
  character(len=*)     ,intent(in)::filename
  logical              ,intent(in)::scratch_flag

! local variables
  integer(kind=singleI)::L1,L2

  if (scratch_flag) then
     rewind(unit)
  else
     open (unit,file=filename,status='unknown',form='unformatted')
  endif

  read (unit)L1,L2
  if (L1/=size(int_plane,1) .or. L2/=size(int_plane,2)) then
     call message ('file: '//trim(filename))
     call abort_code ('Inconsistent grid lengths in read_tmp_integer_plane') 
  endif
  read (unit) int_plane

  if (.not.scratch_flag) close(unit)

  return
end subroutine read_tmp_integer_plane


subroutine write_tmp_integer_plane(filename,unit,scratch_flag)

! write an integer real-space plane to a tmp file

!  use fftw_variables, ONLY: nx_fftw, ny_fftw
  use fmax_grids, ONLY : ngrid
  use tmp_plane
  implicit none
  integer(kind=singleI),intent(in)::unit
  character(len=*)     ,intent(in)::filename
  logical              ,intent(in)::scratch_flag

! local variables
  integer(kind=singleI)::L1,L2
  L1 = size(int_plane,1) ; L2 = size(int_plane,2)

! write single plane to a file
  if (scratch_flag) then
     rewind(unit)
  else
     open (unit,file=filename,status='unknown',form='unformatted')
  endif

  write(unit) L1,L2
  write(unit) int_plane

  if (.not.scratch_flag) close(unit)

  return
end subroutine write_tmp_integer_plane


subroutine remove_slice(igrid, derivative, component)

! removes the tmp files of a slice

  use fmax_grids
  use save_directory
  implicit none
  integer(kind=singleI),intent(in)::igrid,derivative,component

! local variables
  integer(kind=singleI)::iplane,iz,unit
  character(len=80)::filename
  logical::exist_flag,scratch_flag

! Procedure remove a slice from the tmp memory

! loop on planes
  do iz=0,Lgrid_local(igrid,3)-1
     iplane = iz + Lgrid_z_start(igrid)

! filename for the plane
     call make_filename(igrid,derivative,component,iplane,filename,unit,scratch_flag)
! removes tmp plane
     call remove_tmp_plane(filename,unit,scratch_flag)

! end of loop on planes
  enddo

  return
end subroutine remove_slice


subroutine get_diff_plane(igrid, iplane, derivative)

! get the first or second derivatives of a plane

  use fmax_grids
  use fmax_one_plane
  use tmp_plane
  implicit none
  
  integer(kind=singleI),intent(in):: igrid, iplane, derivative

! local variables
  integer(kind=singleI)::components, icomp

!  call allocate_tmp_plane()

  select case (derivative)
  case (1)
     components=3
  case (2)
     components=6
  end select
  
  do icomp = 1, components

     call get_component_plane(igrid, iplane, derivative, icomp)

     differentials_plane(icomp,:,:) = plane

! endo of loop on components
  enddo

!  call deallocate_tmp_plane()

  return
end subroutine get_diff_plane


subroutine get_component_plane(igrid, iplane, derivative, component)

! get a plane of a component of a derivative

  use numbers
  use fmax_grids
  use fmax_one_plane
  use save_directory
  implicit none
  
  integer(kind=singleI),intent(in):: igrid, iplane, derivative, component

! local variables
  character(len=80)    ::filename
  logical              ::scratch_flag
  integer(kind=singleI)::unit

  call make_filename(igrid,derivative,component,iplane,filename,unit,scratch_flag)

  call read_tmp_plane(filename,unit,scratch_flag)

  return
end subroutine get_component_plane


subroutine get_Fmax_plane(iplane)

! Reads one Fmax plane
  
  use fmax_one_plane
  use save_directory
  use tmp_plane

  implicit none
  integer(kind=singleI),intent(in)::iplane

! local variables
  integer(kind=singleI)::unit
  character(len=80)::filename
  logical::exist_flag,scratch_flag

!  call allocate_tmp_plane()

  call make_filename(0,0,0,iplane,filename,unit,scratch_flag)

  call read_tmp_plane(filename,unit,scratch_flag)
  Fmax_plane = plane

!  call deallocate_tmp_plane()

  return
end subroutine get_Fmax_plane


subroutine save_Fmax_plane(iplane)

! Updates one Fmax plane

  use fmax_one_plane
  use save_directory
  use tmp_plane

  implicit none
  integer(kind=singleI),intent(in)::iplane

! local variables
  integer(kind=singleI)::unit
  character(len=80)::filename
  logical::scratch_flag

!  call allocate_tmp_plane()

  call make_filename(0,0,0,iplane,filename,unit,scratch_flag)

  plane = Fmax_plane
  call write_tmp_plane(filename,unit,scratch_flag)

!  call deallocate_tmp_plane()

  return
end subroutine save_Fmax_plane


subroutine get_Rmax_plane(iplane)

! Reads one Rmax plane

  use fmax_one_plane
  use save_directory
  use fmax_grids
  use tmp_plane

  implicit none
  integer(kind=singleI),intent(in)::iplane

! local variables
  integer(kind=singleI)::unit
  character(len=80)::filename
  logical::scratch_flag

  call allocate_tmp_integer_plane(ngrid)

  call make_filename(-4,0,0,iplane,filename,unit,scratch_flag)

  call read_tmp_integer_plane(filename,unit,scratch_flag)
  Rmax_plane = int_plane

  call deallocate_tmp_integer_plane()

  return
end subroutine get_Rmax_plane


subroutine save_Rmax_plane(iplane)

! Updates one Rmax plane

  use fmax_one_plane
  use save_directory
  use fmax_grids
  use tmp_plane

  implicit none
  integer(kind=singleI),intent(in)::iplane

! local variables
  integer(kind=singleI)::unit
  character(len=80)::filename
  logical::scratch_flag

  call allocate_tmp_integer_plane(ngrid)

  call make_filename(-4,0,0,iplane,filename,unit,scratch_flag)

  int_plane = Rmax_plane
  call write_tmp_integer_plane(filename,unit,scratch_flag)

  call deallocate_tmp_integer_plane()


  return
end subroutine save_Rmax_plane


subroutine get_vmax_plane(iplane)

! Reads one vmax plane

  use fmax_one_plane
  use save_directory
  use tmp_plane

  implicit none
  integer(kind=singleI),intent(in)::iplane


! local variables
  integer(kind=singleI)::unit,icomp
  character(len=80)::filename
  logical::scratch_flag


!  call allocate_tmp_plane()

  do icomp=1,3

     call make_filename(-icomp,0,0,iplane,filename,unit,scratch_flag)

     call read_tmp_plane(filename,unit,scratch_flag)
     vmax_plane(icomp,:,:) = plane

  enddo

!  call deallocate_tmp_plane()

  return
end subroutine get_vmax_plane


subroutine save_vmax_plane(iplane)

! Update one vmax plane

  use fmax_one_plane
  use save_directory
  use tmp_plane

  implicit none
  integer(kind=singleI),intent(in)::iplane

! local variables
  integer(kind=singleI)::unit,icomp
  character(len=80)::filename
  logical::scratch_flag

!  call allocate_tmp_plane()

  do icomp = 1,3

     call make_filename(-icomp,0,0,iplane,filename,unit,scratch_flag)

     plane = vmax_plane(icomp,:,:)
     call write_tmp_plane(filename,unit,scratch_flag)

  enddo

!  call deallocate_tmp_plane()

  return
end subroutine save_vmax_plane


subroutine save_kspace_vector(igrid)

! Writes a transformed density slice into the tmp file

  use save_directory
  use fftw_variables
  use pinocchio_mpi
  implicit none
  integer(kind=singleI),intent(in)::igrid

! local variables
  integer(kind=singleI)::unit
  character(len=80):: filename
  logical::scratch_flag
  
  call make_filename(igrid,0,0,-(1+proc_id),filename,unit,scratch_flag)

  if (scratch_flag) then
     rewind(unit)
  else
     open (unit,file=filename,status='unknown',form='unformatted')
  endif

  write(unit,err=20)vector

  if (.not.scratch_flag) close(unit)

  return
20 call abort_code('error in writing tmp file'//trim(filename))
  return
end subroutine save_kspace_vector


subroutine get_kspace_vector(igrid)

! Gets a transformed density slice from a tmp file

  use save_directory
  use fftw_variables
  use pinocchio_mpi
  implicit none
  integer(kind=singleI),intent(in)::igrid

! local variables
  integer(kind=singleI)::unit
  character(len=80):: filename
  logical::scratch_flag

  call make_filename(igrid,0,0,-(1+proc_id),filename,unit,scratch_flag)

  if (scratch_flag) then
     rewind(unit)
  else
     open (unit,file=filename,status='unknown',form='unformatted')
  endif

  read (unit,err=10)vector

  if (.not.scratch_flag) close(unit)

  return
10 call abort_code('Error in reading density from file '//trim(filename))
end subroutine get_kspace_vector


