!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

module fundamentals
  use numbers
! Number of master processors
  integer(kind=singleI), parameter :: NMaster = 4
! first master processor that contains group data
  integer(kind=singleI), parameter :: first_inode = 2           
! max number of vertices on master node; for 2 Gb of RAM use 7e6
  integer(kind=singleI), parameter :: MaxVertexOnMaster = 7e6
! max groups on server node; for 1Gb use 15e6
  integer(kind=singleI), parameter :: MaxGroupsOnServer=15e6
! max groups in communication
  integer(kind=singleI), parameter :: MaxGroupsBuffer = 1e6
! write recover files each StepsToSkip steps
  integer(kind=singleI), parameter :: StepsToSkip=3
end module fundamentals

module GridData
  use numbers
  use fundamentals
  integer(kind=singleI), parameter :: GroupsPerVertex=6
  integer(kind=singleI), parameter :: GridVertexTypeSize=56
  type GridVertexType
     sequence
     real(kind=singleR)    :: fmax, velx, vely, velz
     integer(kind=singleI) :: i,j,k
     integer(kind=singleI) :: MyGroupID
     integer(kind=singleI) :: GroupID(GroupsPerVertex)
  end type GridVertexType
  type(GridVertexType), allocatable  :: GridVertex(:), GridVertexBuffer(:)
  integer(kind=singleI) :: n_total_vertex
  integer(kind=singleI), parameter :: VertexPointerTypeSize=56
  type VertexPointerType
     sequence
     integer(kind=doubleI) :: GlobalID
     integer(kind=singleI) :: Neighbour(GroupsPerVertex)
     integer(kind=singleI) :: Filament(GroupsPerVertex)
  end type VertexPointerType
  type(VertexPointerType), allocatable :: VertexPointer(:)
  integer(kind=singleI) :: GridVertexMPIType,GridVertexSize
  type GridOffsetType
     sequence
     integer(kind=singleI) :: i_offset, j_offset, k_offset
  end type GridOffsetType
  type(GridOffsetType)  :: GridOffset(GroupsPerVertex)
  integer(kind=singleI) :: InverseNeighbour(GroupsPerVertex)
end module GridData


module FilamentData
  use numbers
  use GridData, only : GroupsPerVertex
  integer(kind=singleI) :: FilamentDataMPIType, FilamentDataMPISize, n_filaments
  integer(kind=singleI), parameter :: is_filament=-10, FilamentTypeSize=40
  type FilamentType
     sequence
     real(kind=singleR)    :: velx, vely, velz
     integer(kind=singleI) :: MyGroupID
     integer(kind=doubleI) :: GlobalID
     logical               :: alive,d1,d2,d3
  end type FilamentType
  type(FilamentType), allocatable :: MasterFilaments(:)
  type(FilamentType), allocatable :: ServerFilaments(:)
  integer(kind=singleI), parameter :: FilamentPointerTypeSize=48
  type FilamentPointerType
     sequence
     integer(kind=singleI) :: NeighbourVertex(GroupsPerVertex), position(GroupsPerVertex)
  end type FilamentPointerType
  type(FilamentPointerType), allocatable :: FilamentPointer(:)
end module FilamentData


module GroupData
  use numbers
  use fundamentals
  integer(kind=singleI), parameter :: GroupDataTypeSize=64
  type GroupDataType
     sequence
     real(kind=singleR)    :: CoMPos(3), CoMVel(3), z_merge, z_appear
     integer(kind=singleI) :: Mass, masterID
     integer(kind=doubleI) :: Peak
     logical               :: modified, alive, d1, d2
  end type GroupDataType
  integer(kind=singleI) :: GroupDataMPIType, GroupDataMPISize,NewGroupName=0
  integer(kind=singleI), allocatable :: GroupIndex(:), GroupName(:)

  integer(kind=singleI) :: n_master_groups,n_allocated_groups,ngroups_on_server
  type(GroupDataType), allocatable :: ServerGroups(:), MasterGroups(:)

  type(GroupDataType)    :: GroupBuffer(MaxGroupsBuffer)

  integer(kind=singleI) :: n_listed_groups,group_list_size
    integer(kind=singleI), allocatable :: group_list(:,:)

end module GroupData


module fragment_mpi
  use numbers
  use fundamentals
  include 'mpif.h'
  integer(kind=singleI) :: MyPE, NumPEs, NServer, GroupPE
  integer(kind=singleI), parameter :: master=1, server=0
  integer(kind=singleI)  :: my_type, server_comm, master_comm, inter_comm
  integer(kind=singleI)  :: mpi_status(mpi_status_size)
  integer(kind=singleI), parameter :: mpi_buff_length=1000000
  real(kind=singleR) :: comm_buffer(mpi_buff_length)
#ifdef RECOVER
  integer, parameter :: recover_unit=9
#endif
end module fragment_mpi


module grid
  use numbers

! myserver(1:Lgrid_global) is GroupID of server node on which plane lives.
! plane_start(1:Nsever) is global_id of first plane on given server node (starts from 0)
! nplanes(1:NServer) is number of planes on a given server node

  integer(kind=singleI):: Lgrid(3),Lgrid_global(3),Lstart(3),Npart,local_z_start
  integer(kind=singleI), allocatable :: plane_start(:), nplanes(:), myserver(:)
!  integer(kind=singleI):: ngrpmax
  real(kind=singleR)   :: BoxSize_global(3)
  logical,dimension(3) :: pbc_flag

! offs-sets for writing different variables
  integer(kind=singleI) :: fmax_offset, index_offset, Velx_offset, Vely_offset, Velz_offset

! grid variables (only allocated by server nodes
  real(kind=singleR), allocatable :: Fmax(:,:,:), plane(:,:), plane1d(:), recvplane1D(:)
  integer(kind=singleI), allocatable :: Index(:,:), indexx(:)
  integer(kind=singleI), allocatable :: n_my_planes(:)
  integer(kind=singleI), allocatable :: server_GroupID(:,:,:)
#ifdef STORE_FMAX
  real(kind=singleR), allocatable :: stored_fmax(:,:)
  integer(kind=singleI), allocatable :: stored_index(:,:)
#endif
#ifdef STORE_VEL
  real(kind=singleR), allocatable :: stored_vel(:,:,:)
#endif

! grid variables allocate by master nodes only
  integer(kind=singleI), allocatable :: num_this_proc(:)
  integer(kind=singleI), allocatable :: plane_done(:)
  
end module grid


module runtime_parameters
  use numbers
  real(kind=singleR) :: current_fmax, final_fmax
  integer(kind=singleI), parameter :: string_length=80
  character(len=string_length) :: InputDir, OutputDir, RunString
  character(len=string_length) :: Fmax_file, Velx_file, Vely_file, Velz_file
  integer(kind=singleI)::stop_flag,out_flag,param_flag
  real(kind=singleR)::Flim,zfin
  integer(kind=singleI)::nout
  real(kind=singleR),allocatable,dimension(:)::fmax_out,z_out
  integer(kind=singleI),parameter::MassMin=10,merger_unit=11
  integer(kind=singleI)::varnumber=0
end module runtime_parameters


module virial_params
  use numbers
  real(kind=singleR)::f_m,f_rm,f_a,f_s,f_ra,espo
end module virial_params


module pinocchio_common
! This module contains the basic properties of the (highest-resolution) box, 
! the cosmological parameters, the run name and other parameters needed by
! cosmo-sp-2.2
  use numbers
! Cosmological parameters
  real(kind=singleR):: Omega0  
  real(kind=singleR):: OmegaLambda
  real(kind=singleR):: Hubble100
  real(kind=singleR):: Sigma8
  real(kind=singleR):: PowerSpectrum  ! flag for the transfer function
  real(kind=singleR):: csp            ! normalization constant for P(k)
  real(kind=singleR):: delta_c        ! delta_c Press & Schechter parameter
  integer(kind=singleI):: filter      ! flag for window filter
  integer(kind=singleI):: flag_ps     ! flag for the analytical mass function
! Basic properties of the highest resolution box
  real(kind=singleR):: InterPartDist  ! comoving inter-particle distance
  real(kind=singleR):: BoxSize(3)     ! 3D comoving size
  real(kind=singleR):: Mpart          ! particle mass
!
  character(len=6)  :: run, kernel    ! name of the run

  real(kind=singleR),parameter::Omegab=0.0456   ! baryonic Omega
end module pinocchio_common


module pinocchio_constants
! CONSTANTS
  use numbers
  real(kind=singleR), parameter :: pi   =3.1415926535897931
  real(kind=singleR), parameter :: twopi=2.*pi
end module pinocchio_constants


module borders
  use numbers
  integer(kind=singleI) :: n_border_max, n_border_left, n_border_right, n_border_from_left, n_border_from_right
  integer(kind=singleI),allocatable,dimension(:) :: border_left, border_right, border_from_left, border_from_right
end module borders


module fast_fmax_module
  use numbers

! log(F) is sampled in a grid of nbin bins running from Flargest to Fsmallest
! This settings should be OK for every run to z=0
  integer(kind=singleI),parameter :: nbin=1001
  real(kind=singleR),parameter :: binsize=0.005
  real(kind=singleR),parameter :: Flargest=binsize*(nbin-1), Fsmallest=0.
  integer(kind=singleI),dimension(nbin) :: fast_fmax_plane=0
  integer(kind=singleI),allocatable,dimension(:,:) :: fast_fmax
  integer(kind=singleI) :: n_current_fmax, n_next_fmax
  integer(kind=singleI) :: ibin
  real(kind=singleI) :: fbin, rbin, largest_fmax=0., next_largest
end module fast_fmax_module


module pinocchio_hdf5
! numbers required to handle hdf5
  use numbers
  integer(kind=singleI) :: file_handle, group_handle, attribute_handle
end module pinocchio_hdf5


!!!module merger_list_module
!!!  use numbers
!!!  integer(kind=singleI) :: n_listed_groups, merger_list_size
!!!  integer(kind=singleI), allocatable, dimension(:,:) :: merger_list
!!!end module merger_list_module
