!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

program merger_histories

! Creates the merger histories of a run of fragment-parallel

  implicit none
  integer::ierr

#ifdef HDF5_OUT
  write (*,*) 'Sorry, this code presently does not read catalogues in HDF5'
  write (*,*) 'run the fragment-parallel code again with UNFORM_OUT or ASCII_OUT'
  stop
#endif

  call initialization()

  call allocate_vectors()

  call build_histories()

  call finish('done')
end program merger_histories


subroutine initialization()

! Initializes the run

  use pinocchio_common
  use merger_common
  implicit none

! local variables
  integer(kind=singleI)::Lgrid1,i
  real(kind=singleR)::Mtot,zfin
  real(kind=singleR),external::grow
  logical::cubic=.false.,success=.false.

! reads the run name from fragment.info
  open(info_unit,file='fragment.info',status='old',err=22)
  read(info_unit,'(a6)',err=22,end=22)run
  close(info_unit)
  write(*,*)'run: ',run

! Reads the $run.info file
  do while (.not.success)
     open(run_unit,file=run//'.info',status='old',err=56)
     read(run_unit,*)BoxSize(1)
     read(run_unit,*)Omega0
     read(run_unit,*)OmegaLambda
     read(run_unit,*)PowerSpectrum
     read(run_unit,*)Sigma8
     read(run_unit,*)Hubble100
     if (.not.cubic) then
! first: tries to read the three grid lengths (non-cubic case)
        read(run_unit,*,err=56) Lgrid
        success  = .true.
     else
! second: reads only one grid length and assumes a cubic box
        read(run_unit,*,err=56) Lgrid1
        success      = .true.
        Lgrid = Lgrid1
     endif
50   close(run_unit)
     if(success) then
        InterPartDist=BoxSize(1)/float(Lgrid(1))
        BoxSize = InterPartDist*Lgrid
     endif
     cubic = .true.
  enddo

! setting cosmology
  call normalization()
  filter=2
  delta_c=1.686d0
  flag_ps=1
  call grow_start()

! Opens fragment.info again and reads the kernel file name
  open(info_unit,file='fragment.info',status='old')
  read(info_unit,*)
  read(info_unit,'(a6)',end=22)kernel
  write(*,*)'kernel for output files: '//kernel

! final redshift and inverse growing mode
  read(info_unit,*,end=22) zfin

! intermediate outputs
  read(info_unit,*,end=22)nout
  allocate (z_out(nout+1)) ; allocate (fmax_out(nout+1))
  read(info_unit,*,end=22)z_out(1:nout)
  z_out(nout+1)=zfin
! sorts the output redshifts in decreasing order
  call sort(nout+1,z_out)
  write(*,*)'Total numer of redshifts: ',nout+1
  write(*,'(99(1x,f5.2))')z_out

  do i=1,nout+1
     fmax_out(i)=1./grow(z_out(i))
  enddo
!  write(*,'(99(1x,f5.2))')fmax_out

! computes the particle mass
  Mtot   = 2.778e11*Hubble100**2*Omega0*product(BoxSize)
  Mpart  = Mtot/float(product(Lgrid))

! endo of reading of fragment.info
  close(info_unit)

! information file
  open(group_unit,file='mergers.info',status='old')
  read(group_unit,*)firstproc
  read(group_unit,*)lastproc
  read(group_unit,*)number_of_groups
  close(group_unit)

  return
! error messages
22 call abort_code ('Error in fragment.info')

56 call abort_code ('Error in '//run//'.info')

end subroutine initialization


subroutine allocate_vectors()

! Allocates the vectors to store the merger histories

  use pinocchio_common
  use merger_common
  use vectors
  implicit none
  
  integer(kind=singleI)::ierr

  allocate (belongs_to(number_of_groups),stat=ierr)
  if (ierr/=0) call abort_code('Error in allocating vector belongs_to')
  belongs_to=0

  allocate (linking_list(number_of_groups),stat=ierr)
  if (ierr/=0) call abort_code ('Error in allocating vector linking_list')
  linking_list=0

  allocate (merged_with(number_of_groups),stat=ierr)
  if (ierr/=0) call abort_code ('Error in allocating vector merged_with')
  merged_with=0

  allocate (mass(number_of_groups),stat=ierr)
  if (ierr/=0) call abort_code ('Error in allocating vector mass')
  mass=0

  allocate (mass_main_at_merger(number_of_groups),stat=ierr)
  if (ierr/=0) call abort_code ('Error in allocating vector mass_main_at_merger')
  mass_main_at_merger=0

  allocate (z_appear(number_of_groups),stat=ierr)
  if (ierr/=0) call abort_code ('Error in allocating vector z_appear')
  z_appear=-1.

  allocate (z_merge(number_of_groups),stat=ierr)
  if (ierr/=0) call abort_code ('Error in allocating vector z_merge')
  z_merge=-1.

  print *,'Allocations done'

  return
end subroutine allocate_vectors


subroutine build_histories()

  use pinocchio_common
  use merger_common
  use vectors
  implicit none

  real   (kind=singleR), external :: inverse_grow

  integer(kind=singleI) :: Group1, Group2, Mass1, Mass2, iout, old_i
  integer(kind=doubleI) :: Peak1, Peak2
  real   (kind=singleR) :: fmax, z_appear1, z_appear2, z_now
  character(len=string_length) :: filename
  integer(kind=singleI) :: count1=0, count2=0, ccc, iii
  
!open the file with the list of mergers
  filename = 'list_of_mergers.dat'
  open(merger_unit,file=filename,status='old',form='unformatted')
  iout=1

! loop on the mergers
  do while (.true.)

! read a single line
     read(merger_unit,end=1)fmax, Group1, Peak1, Mass1, z_appear1, Group2, Peak2, Mass2, z_appear2

     if (fmax<fmax_out(iout)) then
        call write_merger_histories(iout)
        iout=iout+1
     endif

     if (Group2==0) then
! initializes an appearing group
        z_now                = inverse_grow(1./fmax)
        z_appear(Group1)     = z_now
        belongs_to(Group1)   = Group1
        linking_list(Group1) = Group1
        mass(Group1)         = Mass1
        z_merge(Group1)      = -1.
        count1=count1+1
     else

! merger
        z_now                  = inverse_grow(1./fmax)
        if (z_appear(Group1)==-1.) z_appear(Group1) = inverse_grow(1./z_appear1)
        if (z_appear(Group2)==-1.) z_appear(Group2) = inverse_grow(1./z_appear2)
        z_merge(Group2)        = z_now
        mass(Group1)           = Mass1
        mass(Group2)           = Mass2
        mass_main_at_merger(Group2) = Mass1
        merged_with(Group2)    = Group1

        if (linking_list(Group1)==0) linking_list(Group1)=Group1
        if (belongs_to(Group1)==0) belongs_to(Group1)=Group1
        count2=count2+1

! check on the merger history of Group2
        ccc=1
        iii=Group2
        do while (linking_list(iii)/=Group2)
           if (linking_list(iii)==0) then
              print *,'INCONSISTENCY IN THE LINKING LIST!',Group2,ccc
              exit
           endif
           iii=linking_list(iii)
           ccc=ccc+1
           if (ccc>1e6) then
              print *,'LOOPING OF THE LINKING LIST!',Group2,ccc,iii,linking_list(iii)
              exit
           endif
        enddo

! Updating of the linking lists

! Case 1: two groups with no substructure
        if (linking_list(Group1)==Group1 .and. linking_list(Group2)==Group2) THEN 
           linking_list(Group1)=Group2
           linking_list(Group2)=Group1

! Case 2: group 1 has substructure, group 2 has not
        else if (linking_list(Group1)/=Group1 .and. linking_list(Group2)==Group2) then
! old procedure
!        linking_list(Group2)=linking_list(Group1)
!        linking_list(Group1)=Group2
! new procedure
           linking_list(Group2)=Group1
           old_i=Group1
           do while (linking_list(old_i)/=Group1)
              old_i=linking_list(old_i)
           enddo
           linking_list(old_i)=Group2

! Case 3: group 2 has substructure, group 1 has not
        else if (linking_list(Group1)==Group1 .and. linking_list(Group2)/=Group2) then
           old_i=Group2
           do while (linking_list(old_i)/=Group2)
              old_i=linking_list(old_i)
              belongs_to(old_i)=Group1
           enddo
           belongs_to(Group2)=Group1
           linking_list(Group1)=linking_list(Group2)
           linking_list(Group2)=Group1

! Case 4: the two groups have substructure
        else
           old_i=Group2
           do while (linking_list(OLD_I)/=Group2)
              old_i=linking_list(old_i)
              belongs_to(old_i)=Group1
           enddo
! old procedure
!        INDX=linking_list(Group1)
!        linking_list(Group1)=linking_list(Group2)
!        linking_list(Group2)=INDX
! new procedure
           old_i=Group1
           do while (linking_list(old_i)/=Group1)
              old_i=linking_list(old_i)
           enddo
           linking_list(old_i)=linking_list(Group2)
           linking_list(Group2)=Group1

        endif

        belongs_to(Group2)=Group1

     endif
  enddo

1 call write_merger_histories(iout)
  if (iout/=nout+1) write(*,*)'WARNING: Inconsistency in the number of outputs'

  print *,'number of appears: ',count1
  print *,'number of mergers: ',count2

  return
end subroutine build_histories


subroutine write_merger_histories(iout)

  use pinocchio_common
  use merger_common
  use vectors
  implicit none

  integer(kind=singleI),intent(in)::iout
  real(kind=singleR),external::inverse_grow

! local variables
  integer  (kind=singleI)      :: iproc,Group,NewMass,i
  real     (kind=singleR)      :: aa(10)
  character(len=string_length) :: filename
  character(len=4)             :: redshift
  character(len=2)             :: label

  if (z_out(iout) < 10) then
     write(redshift,'(f4.2)')z_out(iout)
  else
     write(redshift,'(f4.1)')z_out(iout)
  endif

! Read the grp files to update the masses of existent   
  write(*,*)'Updating masses from grp file'
  do iproc=firstproc,lastproc
     write(label,'(i2)')iproc
     if (iproc<10)label(1:1)='0'
     filename='frag.'//redshift//'.'//kernel//'.grp.out.'//trim(adjustl(label))
     write(*,*)'Opening file '//trim(filename)
#ifdef ASCII_OUT
     open(cat_unit,file=trim(filename),status='old')
#endif
#ifdef UNFORM_OUT
     open(cat_unit,file=trim(filename),status='old',form='unformatted')
#endif
     do while (.true.)
#ifdef ASCII_OUT
        read(cat_unit,*,end=1,err=1)Group, aa, NewMass
#endif
#ifdef UNFORM_OUT
        read(cat_unit,end=1,err=1)Group, aa, NewMass
#endif

        if (NewMass> MassMin .and. belongs_to(Group)/=Group) then
           print *,'ERRORE!!!',Group,NewMass,belongs_to(Group),z_appear(Group)
        endif

        if (NewMass > MassMin .and. NewMass < Mass(Group)) then
           print *,'INCONSISTENZA!!!',Group,NewMass,Mass(Group)
        endif


        if (NewMass>MassMin) Mass(Group)=NewMass
     enddo
1    close(cat_unit)
     write(*,*)'Updated masses from file '//trim(filename)
  enddo

  filename='frag.'//redshift//'.'//kernel//'.halo.par.out'
  write(*,*)'Writing file '//trim(filename)
  open(out_unit,file=filename,status='unknown')
  do i=1,number_of_groups
     if (mass(i)>massmin) THEN
        write(out_unit,'(7(1x,i9),3(1X,F9.4),3x,i12)') &
                i,belongs_to(i),linking_list(i),merged_with(i) &
                ,mass(i),mass_main_at_merger(i),0,z_merge(i),0.0,z_appear(i)
     endif
  enddo
  close(out_unit)
  print *,'Done writing file'

  return
end subroutine write_merger_histories


subroutine abort_code(messg)
  implicit none
  character(*) :: messg
  write (*,'(a)') messg
  stop
end subroutine abort_code

subroutine finish(messg)
  implicit none
  character(*) :: messg
  write (*,'(a)') messg
  stop
end subroutine finish


