!*****************************************************************
!*                       PINOCCHI0  V2.2                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2008
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

! 
! Routines written by Pierluigi Monaco
!   adapted from Numerical Recipes in Fortran,
!   W.H. Press, S.A. Teukolsky, W.T. Vettering, B.P. Flannery,
!   1992, Cambridge University Press 
!

subroutine index_sort(n,vector,index)

! performs indexed sorting of a vector
! using the quicksort algorithm

  use numbers
  implicit none
  integer(kind=singleI),intent(in)::n
  real(kind=singleR),intent(in)::vector(n)
  integer(kind=singleI),intent(out)::index(n)

  integer(kind=singleI),parameter::m=7,nsave=100

  integer(kind=singleI)::save(nsave),isave
  integer(kind=singleI)::i,ii,j,k,l,indx,tmp
  real(kind=singleR)::c
! init

  do i=1,n
     index(i)=i
  enddo
  isave=0
  k=1
  ii=n
! main loop
  do while (.true.)
     if (ii-k.lt.m) then
! straight insertion
        do j=k+1,ii
           indx=index(j)
           c=vector(indx)
           do i=j-1,1,-1
              if (vector(index(i)).le.c) goto 11
              index(i+1)=index(i)
           enddo
           i=0
11         index(i+1)=indx
        enddo
! finished?
        if (isave.eq.0) return
! update
        ii=save(isave)
        k=save(isave-1)
        isave=isave-2

     else
! partition exchange
        l=(k+ii)/2
        tmp=index(l)
        index(l)=index(k+1)
        index(k+1)=tmp

        if (vector(index(k+1)).gt.vector(index(ii))) then
           tmp=index(k+1)
           index(k+1)=index(ii)
           index(ii)=tmp
        endif

        if (vector(index(k)).gt.vector(index(ii))) then
           tmp=index(k)
           index(k)=index(ii)
           index(ii)=tmp
        endif

        if (vector(index(k+1)).gt.vector(index(k))) then
           tmp=index(k+1)
           index(k+1)=index(k)
           index(k)=tmp
        endif

        i=k+2
        j=ii-1
        indx=index(k)
        c=vector(indx)

        do while (.true.)
           do while (vector(index(i)).lt.c)
              i=i+1
           enddo
           do while (vector(index(j)).gt.c) 
              j=j-1
           enddo
           if (j.lt.i) goto 12
           tmp=index(i)
           index(i)=index(j)
           index(j)=tmp
           i=i+1
           j=j-1

        enddo

12      index(k)=index(j)
        index(j)=indx
        isave=isave+2

        if (isave.gt.nsave) stop 'Increase nsave'

        if (ii-i+1.ge.j-k) then
           save(isave)=ii
           save(isave-1)=i
           ii=j-1
        else
           save(isave)=j-1
           save(isave-1)=k
           k=i
        endif

     endif
  enddo
end subroutine index_sort


subroutine integrate(f,x1,x2,result)

! Romberg's integration routine

  use numbers
  implicit none
  real(kind=singleR),intent(in) :: x1,x2
  real(kind=singleR),intent(out) :: result
  real(kind=singleR),external ::f

  integer(kind=singleI),parameter :: nmax=30,nextr=5
  real(kind=singleR),parameter :: small=1.e-6
  integer(kind=singleI) :: i
  real(kind=singleR) :: p(nmax+1),r(nmax+1),err

  p(1)=1.
  r(1)=0.
  call trapezoidal (1,f,x1,x2,r(1))

  do i=2,nmax
     p(i)=0.25*p(i-1)
     r(i)=r(i-1)
     call trapezoidal(i,f,x1,x2,r(i))

     if (i.ge.nextr) then
        call polinomial(p(i-nextr+1),r(i-nextr+1),nextr,0.,result,err)
        if (abs(err).le.abs(result)*small) return
     endif

  enddo
  print '(''Warning: integration did not converge '',1x,e12.6,1x,e12.6)',result,err
  return

end subroutine integrate


subroutine trapezoidal(n,f,x1,x2,result)

! Integral evaluation with trapezoidal rule of order n

  use numbers
  implicit none
  integer(kind=singleI),intent(in) :: n
  real(kind=singleR),intent(in) :: x1,x2
  real(kind=singleR),intent(out) :: result
  real(kind=singleR),external :: f

  integer(kind=singleI) :: ng,i
  real(kind=singleR) :: deltax,dx,x,step

  deltax=x2-x1
  if (n.eq.1) then
     result=deltax*0.5*(f(x1)+f(x2))
  else
     ng=2**(n-2)
     dx=deltax/float(ng)
     step=0.
     do i=1,ng
        x=x1+(-0.5+i)*dx
        step=step+f(x)
     enddo
     result=0.5*(result+dx*step)
  endif

  return
end subroutine trapezoidal


subroutine polinomial(x,y,n,xp,yp,err)

! Polynomial interpolation with Neville's algorithm

  use numbers
  implicit none
  integer(kind=singleI),intent(in) :: n
  real(kind=singleR),intent(in) :: x(n),y(n),xp
  real(kind=singleR),intent(out) :: yp,err

  integer(kind=singleI),parameter :: nmax=20
  integer(kind=singleI) :: i,m,nearest
  real(kind=singleR) :: C(nmax),D(nmax),a,b,e
! initializes C and D coefficients
  do i=1,n
     C(i)=y(i)
     D(i)=y(i)
  enddo
! finds nearest grid point
  a=100.*abs(x(n)-x(1))
  do i=1,n
     if (abs(xp-x(i)).lt.a) then
        a=abs(xp-x(i))
        nearest=i
     endif
  enddo
! first guess
  yp=y(nearest)
! iteration
  do m=1,n-1
     do i=1,n-m
        e=(C(i+1)-D(i))/(x(i)-x(i+m))
        a=(x(i)  -xp)*e
        b=(x(i+m)-xp)*e
        C(i)=a
        D(i)=b
     enddo
! further guess
     if (2*(nearest-1).lt.n-m) then
        err=C(nearest)
     else
        nearest=nearest-1
        err=D(nearest)
     endif
     yp=yp+err
  enddo
        
  return
end subroutine polinomial


subroutine interpola(xv,yv,nv,x,y)

! Quadratic interpolation from x- and y-vectors

  use numbers
  implicit none
  integer(kind=singleI),intent(in) :: nv
  real(kind=singleR),intent(in) :: xv(nv),yv(nv),x
  real(kind=singleR),intent(out) :: y

  integer(kind=singleI) ::i,iv
  real(kind=singleR) :: dy,xp(3),yp(3),xp4(4),yp4(4)

  if (x.le.xv(1)) then
     y=yv(1)
  else if (x.ge.xv(nv)) then
     y=yv(nv)
  else if (x.le.xv(2)) then
     do i=1,3
        xp(i)=xv(i)
        yp(i)=yv(i)
     enddo
     call polinomial(xp,yp,3,x,y,dy)
  else if (x.ge.xv(nv-1)) then
     do i=1,3
        xp(i)=xv(nv-3+i)
        yp(i)=yv(nv-3+i)
     enddo
     call polinomial(xp,yp,3,x,y,dy)
  else
         
!         iv=int((x-xv(1))/(xv(2)-xv(1)))+2

!         if (iv<1 .or. iv>nv .or. .not.(x.ge.xv(iv-1) .and. x.le.xv(iv))) then

     do iv=1,nv
        if (x.le.xv(iv)) exit
     enddo
            
!         endif

     if (iv.eq.nv+1) then
        print *,'Error in INTERPOLA ',x,xv(1),xv(nv),iv
        stop
     endif

     do i=1,4
        xp4(i)=xv(iv-3+i)
        yp4(i)=yv(iv-3+i)
     enddo
     call polinomial(xp4,yp4,4,x,y,dy)
  endif
  return
end subroutine interpola


subroutine linint(xv,yv,nv,x,y)

! Linear interpolation

  use numbers
  implicit none
  integer(kind=singleI) :: nv
  real(kind=singleR) :: xv(nv),yv(nv),x,y,dy
  integer(kind=singleI) :: iv

  if (x.le.xv(1)) then
     y=yv(1)
  else if (x.ge.xv(nv)) then
     y=yv(nv)
  else
     do iv=1,nv
        if (x.le.xv(iv)) exit
     enddo
     if (x.eq.xv(iv-1)) then
        y=yv(iv-1)
     else if (x.eq.xv(iv)) then
        y=yv(iv)
     else
        y=(yv(iv)*(x-xv(iv-1))+yv(iv-1)*(xv(iv)-x))/(xv(iv)-xv(iv-1))
     endif
  endif
  return
end subroutine linint
